// HP80DIS - Everett Kaser - January 2015
/*********************  Header Files  *********************/

#ifdef __linux__
#include <stdlib.h>
#include <strings.h>
#include <unistd.h>
#define BOOL unsigned int
#define BYTE unsigned char
#define WORD unsigned short
#define DWORD unsigned int
#define FALSE 0
#define _lseek(a, b, c) lseek(a, b, c)
#define max(a,b) (((a) > (b)) ? (a) : (b))
#define min(a,b) (((a) < (b)) ? (a) : (b))
#define _O_BINARY 0
#define _O_CREAT O_CREAT
#define _O_RDONLY O_RDONLY
#define _O_RDWR O_RDWR
#define _O_TRUNC O_TRUNC
#define _S_IREAD S_IREAD
#define _S_IWRITE S_IWRITE
#define stricmp(a, b) strcasecmp(a, b)
#define strnicmp(a, b, c) strncasecmp(a, b, c)
#define SYSTEMTIME struct tm *
#define TRUE 1
#else
#include <windows.h>
#include <io.h>
#endif
#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <string.h>
//#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>

// DEVTIME sets default command line values when debugging in the development environment.
// Set it to 0 before release.
#define	DEVTIME	0




/// NOTE: If you have problems with DAD/EQU labels at the start of the .LST file getting
/// overwritten/changed to other labels, usually with values of 0, then the problem MAY
/// be caused by .declaring GLOBAL FILE DADs as .fda rather than .fgl. Usually, the
/// offending label is the one that the first DAD or EQU in the file gets changed TO.

/// OR... you're .f'ing a label that is NOT defined!

/********************* Global Variables *******************/
char			*TitleRevStr = "CAPDIS 4.4 (19 Nov 2023)";

/* REVISIONS:

NEXT REVISION: SUPPORT HEX ADDRESSES IN GLOBALS FILE.
	4.4 19 NOV 2023
		FIXED: A .BSZ would work the first time, but then get changed to a .BYT the next time.
	4.3 10 NOV 2023
		FIXED: When disassembling system ROMs 1, 2, & 3 (non-bank-switched), the "bytes free" calculation at the end of the ROM was crashing (when there was no starting .LST file).
	4.2 8 NOV 2023
		ADDED: For BPGMs that were LOADBIN'd and then STOREBIN'd, the address tables show the absolute addresses at which they'd been loaded.
			CAPDIS now goes through and turns those addresses back into relative addresses, so the disassembly looks like a 'new' BPGM with
			offset addresses from 0 (the beginning of the BPGM's PCB).
	4.1 28 OCT 2023
		FIXED: ".fda label" was being ignored sometimes (after ROMJSB), because .def (later) overwrote the REC flags.
	4.0 27 OCT 2023
		CHANGED: Made <address> optional on many DOT-commands. To do so, you MUST place the DOT-command(s) on lines
			IMMEDIATELY before the line they apply to, so that CAPDIS can search for the next line with an address
			and use it. On some DOT-commands, the address can be simply omitted, on others (with following arguments)
			a '-' character must be used in place of the address.
	3.30 3 OCT 2023
		ADDED: -S command to export source with NO line numbers.
	3.20 30 SEP 2023
		ADDED: Changes from Felix/Philippe Bergheaud to add .feq to name single bytes, typically ROM or ERROR numbers after ROMJSB/ERROR.
		FIXED: WARNING: line len > 80 didn't have a CR/LF on the end
	3.19 10 Mar 2022
		CHANGED: FormatAddress() FindGlobal call to fix problem with BPGMs and DAD values < length of the BPGM (from Felix)
	3.18 14 Mar 2020
		FIXED: HP-87 started doing operations (like LDM, SBM, ADM) of addresses and EQUs on 3-byte register groups, and the disassembler wasn't properly handling them.
		FIXED: When disassembling system ROMs at 00000, 20000, 40000, the code to auto-format ROMJSB, ERROR, ERROR+ calls wasn't being invoked.
	3.17 24 Dec 2016
		FIXED: Some messages printed as a result of command-line syntax errors didn't have CR/LF on the end.
		ADDED: -Bl options (and .opt -Bl:n) to control auto-backup of .lst files, where 'l' is from 0 to 9 and 'n' is <'l'.
	3.16 21 Dec 2016
		FIXED: Full-line comments immediately preceding code at address 0 often became duplicated over and over.
		ADDED: options: -g (use GLOBALS file from CURRENT directory, -G (use GLOBALS file from CAPDIS.EXE directory (default))
		ADDED: options: -k (Keeps comments from .dot commands, otherwise .com and .eol are ignored and ONLY comments in the disassembled source are kept, making it easier to DELETE comments)
		FIXED: With a .lst file containing hex addresses, disassembling code at addresses starting with 'A'-'F' crashed and burned, one way or another.
	3.15 19 Dec 2016
		CHANGED: Used to be that the first WHITESPACE (SPACE or TAB) after a comment '!' char was tossed.
		     Now, only if the first char after the '!' is a SPACE does it get tossed, the TAB is retained.
	3.14 15 Dec 2016
		FIXED: When disassembling a ROM, the -c -q .opt values weren't 'sticking'.
		FIXED: 'empty' full-line comments were being removed.
		FIXED: in hex mode, addresses with lowercase a-f didn't work.
		ADDED: .val address EQU_label for equating a single byte with a label (and vice versa)
		CHANGED: if formatting a literal number as .dec or .bcd and the number is < 8, then the trailing 'D' or 'C' is omitted (to avoid "0C" for example).
		ADDED: -C and -Q (uppercase) command line options to override .opt settings
	3.1 12 Dec 2016
		Changed program name to CAPDIS (CAPricorn DISassembler).
		Changed command line options.
		Added .OPT DOT COMMAND that appears as FIRST line in .LST file.
	3.0 9 Dec 2016
		Added -75 support for HP-75 LEX file disassembly (but only RAM LEX files, not ROM).
		Added globals75.txt file from Martin Hepperle.
		Added automatic ".dec <addr>" DOT COMMAND for ERROR# following JSB=ERROR or JSB=ERROR+.
	2.4 14 Aug 2016
		Quick-and-dirty -iLLLLLL=IIIIII:EEEEEE option to copy comments, labels, formatting from one file to another at different addresses (CP/M -> TERM/80, for example)
	2.3 20 July 2016
		Fixed: single-byte immediate operands paid no attention to .feq, .fgl
		Added .dec and .bcd for forcing format of single-byte =immediate numeric values as well as in BYT and BSZ
	2.2  7 Jul 2016
		Fixed input file length problems when dealing with 87 BPGMs.
	2.1 27 Feb 2015
		Added line numbers to -s (source only) option
		Changed output file to .SRC when -s (source only) option
		Shortened len of comments on table of addresses
		Changed register references from 'r' to 'R' (Assembler ROM doesn't like 'r'!)
		Removed leading 0's from =values (Assembler ROM doesn't like them!)
		Added WARNING when formatting out 16-bit literal numbers (Assembler ROM doesn't like them!)
		Added WARNING when formatting out 16-bit relative jump numeric address rather than a label
		Added check for duplicate labels.
		Allow COMMENT lines in GLOBAL file to start with '!' as well as '*'
		Added WARNING when -s output and line length is exactly 80, and when line len is >80
		Added -d option for DELETING comments during -s output (to make smaller files for HP-85 import).
		Moved EOL comments 4 characters to right to line up with HP-87 comment column
		When -s, don't output ":ROM#" on DAD statements.
*/

#define	MACH_85	0
#define	MACH_87 1
#define	MACH_75 2

// .LST file options (that are overridden by command-line options
long			LstMach=-1, mach=-1;	// default to HP-85 (0) unless -87 (1) or -75 (2) option is specified
long			LstIsROM=-1, IsROM=-1;	// 0-7=8K ROM base address (0=0, 1=8K, 2=16K, etc), default to 8=BPGM
long			LstForceComment=-1, ForceComment=-1;	// if add a '!' on lines that don't have it, default to 0 (OFF)
long			LstDoWarnings=-1, DoWarnings=-1;		// if warn about literal addresses that the Assembler ROM won't like, default to 1 (ON)
long			LstLocalGlobal=-1, LocalGlobal=-1;	// if use local (current dir) globals file instead of where CAPDIS.EXE is located
long			LstBackupNum=-1, BackupNum=-1;		// .bak# for backing up .LST file
long			LstBackupMax=-1, BackupMax=-1;		// maximum # of backup files before recycling

BOOL			HexOut=FALSE;	// defaults to OCTAL unless 'x' included after model#
BOOL			HexIn=FALSE;	// dafaults to OCTAL unless HEX format is detected in input .LST file
long			OnlySrc=0;		// defaults to 0 unless -s or -S option is specified
BOOL			DoLineLenWarning=FALSE;	// if warn about line length >80
BOOL			DelComments=FALSE;	// if delete ALL comments from -s .SRC output
BOOL			IncDotCOM=FALSE;	// if include .COM and .EOL comments.  Defaults OFF, so only comments from the .lst "source" code are included.  This allows easier DELETION of previously entered comments.

BOOL			BailOut = FALSE;	// in case an unknown error occurs processing the inputs

#define	PATHLEN	512
char			BinFileName[PATHLEN]="";
char			ListFileName[PATHLEN], ImportFileName[PATHLEN], GlobalFileName[PATHLEN];

WORD			TableBase;
int				OutLineNum=1;	// for writing OnlySrc

unsigned char	*InBuf;		// ptr to malloc'd input ROM or BPGM bytes
long			InBufLen;

char			*LstBuf;	// ptr to malloc'd input LST file
char			*pLST;		// working ptr into LstBuf
long			LstBufLen;

DWORD			LstStart, ImpStart=0, ImpEnd=0, ImpDelta=0;

int				OutFile;	// file handle to which all output is written (temp .LST file)

char			TmpFileName[]="_tmp.tmp.LST";

char			eolseq[]="\r\n";
int				eollen=2;

char			ScratBuf[8192];

char			*pLstLines[65536];	// ptrs to each line of the input .LST file
									// up to 8 pointers for EVERY byte of the ROM.  Should be enough.
int				LstLinesCnt;
int				CurLineNum;

//**********************
// Disassembly stuff
//

char *globfile[]={
	"globals85.txt",
	"globals87.txt",
	"globals75.txt"
};

#define	GLO_DAD		0
#define	GLO_EQU		1

struct gloval {
	char	type;
	char	label[11];// really only use 9 max, the others keep the structure a multiple of 4 in len
	DWORD	address;
};
typedef struct gloval GLOBAL;

GLOBAL	Globals[8192];
int		GlobalsCnt;

long	BaseAddress, EndAddress;
BOOL	IsSystemROM=FALSE;

#define	REC_COD	0
#define	REC_BYT	1
#define	REC_ASC	2
#define	REC_ASP	3
#define	REC_BSZ	4
#define	REC_DEF	5
#define	REC_VAL	6
#define	REC_RAM	7
#define	REC_I_O	8
#define	REC_DRP	9
#define	REC_REM	10
#define	REC_UND	11
#define	REC_DAD	12
#define	REC_EQU 13

#define	FLG_ROMJSB_DEF	0x0001
#define	FLG_FDA			0x0002
#define	FLG_DAD_DEFINED	0x0004
#define	FLG_FEQ			0x0008
#define	FLG_EQU_DEFINED 0x0010
#define	FLG_GLO			0x0020
#define	FLG_GTO			0x0040
#define FLG_DEC			0x0080
#define FLG_BCD			0x0100
#define	FLG_SYSJSB_DEF	0x0200
#define FLG_LABEL_ADDR	0x0400
#define	FLG_IGNORE_DB	0x0800	// when a .DEC or .BCD is seen BEFORE an address, then ignore .DEC/.BCD in DOT-commands section AND don't clear FLG_DEC/FLG_BCD when .BYT is seen in DOT-commands section

struct codrec {
	WORD			type;
	WORD			flags;
	char			label[9];	// line-label
	char 			reflabel[9];// argument EQU label reference
	WORD			datover;	// over ride of how many data bytes there are for disassembly
	DWORD			address;
		// Note: the 'address' field contains the actual HP-85 address shifted over 16 bits.
		// If the lower 16 bits is 0xFFFF, then this record applies to real bytes in the HP-85.
		// Otherwise (lower 16 bits == 0x0000 to 0xFFFE), this record merely contains a full-line
		// comment that precedes the actual HP-85 bytes.  It's done this way for ease of
		// sorting/finding/positioning full-line comments and HP-85 addresses relative to
		// each other.  This allows up to 65535 lines of comments before any one line of HP-85
		// code.  Should be enough.
	char			*Comment;
	struct codrec	*pNext;
	struct codrec	*pPrev;
};

typedef struct codrec DISASMREC;

#define	FindRecNext(f)	(f->pNext)
#define	FindRecPrev(f)	(f->pPrev)
#define	HP85ADDR(n)		(((n)>>16)&0x0000FFFF)
#define	SkipWhite(p)	while( *(p) && (*(p)==' ' || *(p)=='\t') ) ++(p)

#define	MAXROMS		32

// ptr to blocks of DISASMRECs for code we're disassembling
// For each of the pDis entries, there are two values, the first ([0]) is the pointer to the
// FIRST record, and the second ([1]) is the pointer to the LAST record.
DISASMREC	*pDis[2];

// ROMnums[] contains the ROM numbers of the supported option-ROMs.
// numROMS, of course, keeps track of how many option-ROMs (including ROM0) have been loaded.
int	ROMnums[MAXROMS] = {
	0010,	// Service
	0050, 	// Assembler
	0250,	//
	0260,	//
	0300,	// I/O
	0317,	//
	0320,	// Mass Storage
	0340,	//
	0350,	//
	0360	//
	};
int			numROMS=10, curROM=0;
BYTE		DisArp=0xFF, DisDrp=0xFF, DisTyp=REC_COD, LastDisTyp;
WORD		DisDataOverride;
//DWORD		ScrollIndex[2048];
//#define	NUMSI	(sizeof(ScrollIndex)/sizeof(DWORD))

DWORD		NxtByt, NxtRom, saveNxtByt;	// "pointer" to next byte to disassemble
DISASMREC	*NextRec, *pDefLabel=NULL, ScratchDISASM;

char	*opc[256] = {
	// 000
	"ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ",
	"ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ",
	"ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ",
	"ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ",
	"ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ",
	"ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ",
	"ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ",
	"ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ",
	// 100
	"DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ",
	"DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ",
	"DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ",
	"DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ",
	"DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ",
	"DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ",
	"DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ",
	"DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ",
	// 200
	"ELB ","ELM ","ERB ","ERM ","LLB ","LLM ","LRB ","LRM ",
	"ICB ","ICM ","DCB ","DCM ","TCB ","TCM ","NCB ","NCM ",
	"TSB ","TSM ","CLB ","CLM ","ORB ","ORM ","XRB ","XRM ",
	"BIN ","BCD ","SAD ","DCE ","ICE ","CLE ","RTN ","PAD ",
	"LDB ","LDM ","STB ","STM ","LDBD","LDMD","STBD","STMD",
	"LDB ","LDM ","STB ","STM ","LDBI","LDMI","STBI","STMI",
	"LDBD","LDMD","STBD","STMD","LDBD","LDMD","STBD","STMD",
	"LDBI","LDMI","STBI","STMI","LDBI","LDMI","STBI","STMI",
	// 300
	"CMB ","CMM ","ADB ","ADM ","SBB ","SBM ","JSB ","ANM ",
	"CMB ","CMM ","ADB ","ADM ","SBB ","SBM ","JSB ","ANM ",
	"CMBD","CMMD","ADBD","ADMD","SBBD","SBMD","NOP3","ANMD",
// The above NOP3 (opcode 326) is not a real opcode and was
// never used (so far as I know) by any Series 80 code.  It is
// an officially unsupported opcode in the HP-85 CPU.
// The hardware engineers investigated it for me in the early 1980's
// and found that, the way the CPU was designed, it resulted in a
// 3-byte "no operation".  What it does is fetch the following two
// bytes from memory, but does nothing with them.  So, this code:
//	BYT	326
//	BYT skip1
//	BYT skip2
//	BYT nextopcode
// will skip the skip1 and skip2 bytes and then resume execution
// with the nextopcode byte, with no modifications to the CPU registers,
// status, or memory, other than the incrementing of the PC (reg 4).
// I include it here just for the sake of completeness.
	"CMBD","CMMD","ADBD","ADMD","SBBD","SBMD","NOP1","ANMD",
// The above NOP1 (opcode 336) is not a real opcode and was
// never used (so far as I know) by any Series 80 code.  It is
// an officially unsupported opcode in the HP-85 CPU.
// The hardware engineers investigated it for me in the early 1980's
// and found that, the way the CPU was designed, it resulted in a
// 1-byte "no operation".  So, this code:
//	BYT	336
//	BYT nextopcode
// fetches the 326 opcode, but does nothing, and then resumes execution
// with the nextopcode byte, with no modifications to the CPU registers,
// status, or memory, other than the incrementing of the PC (reg 4).
	"POBD","POMD","POBD","POMD","PUBD","PUMD","PUBD","PUMD",
	"POBI","POMI","POBI","POMI","PUBI","PUMI","PUBI","PUMI",
	"JMP ","JNO ","JOD ","JEV ","JNG ","JPS ","JNZ ","JZR ",
	"JEN ","JEZ ","JNC ","JCY ","JLZ ","JLN ","JRZ ","JRN "
};

//******************************************
void HOaddr(char *buf, unsigned int adr) {

	if( HexOut ) sprintf(buf, "%4.4X", adr);
	else sprintf(buf, "%6.6o", adr);
}

/***********************************************/
int CountLabels(char *label) {

	DISASMREC	*psrch;
	int			cnt;

	cnt = 0;
	psrch = pDis[0];
	for(; psrch; psrch = psrch->pNext) if( psrch->label[0] && !stricmp((char*)psrch->label, label)) ++cnt;
	return cnt;
}

/***********************************************/
DISASMREC	*FindEQUlabel(char *label) {

	DISASMREC	*psrch;

	psrch = pDis[0];
	for(; psrch && (psrch->type!=REC_EQU || stricmp((char*)psrch->label, label)); psrch = psrch->pNext) ;
	return psrch;
}
/***********************************************/
DISASMREC	*FindNextEQUaddress(DISASMREC *psrch, DWORD val) {

	if( psrch==NULL ) psrch = pDis[0];
	else psrch = psrch->pNext;
	for(; psrch && (psrch->type!=REC_EQU || HP85ADDR(psrch->address)!=val); psrch = psrch->pNext) ;
	return psrch;
}
/***********************************************/
DISASMREC	*FindDADlabel(char *label) {

	DISASMREC	*psrch;

	psrch = pDis[0];
	for(; psrch && (psrch->type!=REC_DAD || stricmp((char *)psrch->label, label)); psrch = psrch->pNext) ;
	return psrch;
}
/***********************************************/
DISASMREC	*FindNextDADaddress(DISASMREC *psrch, DWORD addr) {

	if( psrch==NULL ) psrch = pDis[0];
	else psrch = psrch->pNext;
	for(; psrch && (psrch->type!=REC_DAD || HP85ADDR(psrch->address)!=addr); psrch = psrch->pNext) ;
	return psrch;
}

//************************************************
DISASMREC *AddRec(WORD typ, WORD flags, char *lbl, DWORD addr, char *rem, WORD datcnt) {

	DISASMREC	*ptmp, *psrch;

	if( NULL!=(ptmp=(DISASMREC *)malloc(sizeof(DISASMREC))) ) {
		if( pDis[0]==NULL ) {
			pDis[0] = pDis[1] = ptmp;
			ptmp->pNext = ptmp->pPrev = NULL;
		} else {
			for(psrch=pDis[0]; psrch!=NULL && psrch->address <= addr; psrch = psrch->pNext) ;
			if( psrch==NULL ) {
				pDis[1]->pNext = ptmp;
				ptmp->pPrev = pDis[1];
				ptmp->pNext = NULL;
				pDis[1] = ptmp;
			} else {
				ptmp->pNext = psrch;
				if( psrch->pPrev==NULL ) {
					pDis[0] = ptmp;
					ptmp->pPrev = NULL;
				} else {
					ptmp->pPrev = psrch->pPrev;
					psrch->pPrev->pNext = ptmp;
				}
				psrch->pPrev = ptmp;
			}
		}
		ptmp->type = typ;
		ptmp->flags = flags;
		strcpy((char *)ptmp->label, lbl);
		ptmp->reflabel[0] = 0;
		ptmp->address = addr;
		if( (!OnlySrc || !DelComments) && rem != NULL && rem[0]!=0 ) {
			ptmp->Comment = (char *)malloc(strlen(rem)+1);
			if( ptmp->Comment ) strcpy(ptmp->Comment, rem);
		} else ptmp->Comment = NULL;
		ptmp->datover = datcnt;
		if( ptmp->label[0] && CountLabels(ptmp->label) > 1 ) printf("ERROR: duplicate label %s\r\n", ptmp->label);
	}
	return ptmp;
}

/**********************************************************/
DISASMREC	*FindRecAddress(DWORD addr) {

	DISASMREC	*psrch;
//	DWORD		addr16;

//	addr16 = HP85ADDR(addr);
	psrch = pDis[0];
	for(; psrch && psrch->address < addr; psrch = psrch->pNext) ;
	if( psrch && psrch->address != addr ) {
		psrch = NULL;
	}
	return psrch;
}

/**********************************************************/
DISASMREC	*FindRecEQU(char *pLabel) {

	DISASMREC	*psrch;
//	DWORD		addr16;

//	addr16 = HP85ADDR(addr);
	psrch = pDis[0];
	for(; psrch && (psrch->type!=REC_EQU || stricmp(psrch->label, pLabel)); psrch = psrch->pNext) ;
	return psrch;
}

/**********************************************************/
DISASMREC	*FindRecDAD(char *pLabel) {

	DISASMREC	*psrch;
//	DWORD		addr16;

//	addr16 = HP85ADDR(addr);
	psrch = pDis[0];
	for(; psrch && (psrch->type!=REC_DAD || stricmp(psrch->label, pLabel)); psrch = psrch->pNext) {

	}
	return psrch;
}

/*******************************************************/
GLOBAL	*FindGlobal(DWORD addr, BOOL romjsb) {

	int		i;

	if( romjsb || addr<BaseAddress || addr>=EndAddress || IsSystemROM ) {	// if address is outside 60000-77777 range OR we're disassembling the System ROM (option ROM 0), search the GLOBAL table
		for(i=0; i<GlobalsCnt; i++) {
			if( Globals[i].type==GLO_DAD && Globals[i].address==addr ) {
				return &Globals[i];
			}
		}
	}
	return NULL;
}

/*******************************************************/
GLOBAL	*FindGlobalLabel(char *label) {

	int		i;

	for(i=0; i<GlobalsCnt; i++) {
		if( !stricmp(Globals[i].label, label) ) {
			return &Globals[i];
		}
	}
	return NULL;
}

/*******************************************************/
void FormatReg(char *pBuf, BYTE r) {

//	char	tmpbuf[16];

	if( r==0xFF ) strcpy(pBuf, "R#");
	else if( r==0x01 ) strcpy(pBuf, "R*");
	else {
		sprintf(pBuf, "R%o", r);
//		strcpy(pBuf, tmpbuf+1);	// throw away leading '0' (octal is assumed)
	}
}

/*********************************************************/
void FormatAddress(char *pBuf, WORD addr, DISASMREC *pLineRec, BOOL multi) {

	DISASMREC	*pRec;
	GLOBAL		*pGlob;
	WORD		flags;
	char		haddr[8], laddr[8];

	flags = (pLineRec==NULL) ? 0 : pLineRec->flags;

	HOaddr(haddr, addr);
	HOaddr(laddr, (unsigned int)HP85ADDR(NxtByt));
	if( mach==MACH_75 && (flags & FLG_SYSJSB_DEF) ) {
		if( NULL!=(pGlob=FindGlobal(addr, TRUE)) ) strcpy(pBuf, pGlob->label);
		else if( NULL==(pRec=FindRecAddress((addr<<16) | 0x0000FFFF)) || pRec->label[0]==0 ) sprintf(pBuf, "%s", haddr);
		else strcpy(pBuf, pGlob->label);
	} else if( flags & FLG_ROMJSB_DEF ) {	// it could be in a System ROM (InBuf[NxtRom+2]==0), this ROM (InBuf[NxtRom+2]==curROM), another ROM (InBuf[NxtRom+2]!=curROM)
		if( InBuf[NxtRom+2]==0 && NULL!=(pGlob=FindGlobal(addr, TRUE)) ) strcpy(pBuf, pGlob->label);
		else if( (BaseAddress>=060000 && InBuf[NxtRom+2]!=curROM) || (BaseAddress<060000) ) {
			do {
				pRec = FindNextDADaddress(NULL, addr);
			} while( pRec!=NULL && pRec->datover!=InBuf[NxtRom+2] );
			if( pRec!=NULL ) strcpy(pBuf, pRec->label);
			else {
				if( DoWarnings ) {
					if( multi ) printf("WARNING at %s: multi-byte literal number %s\r\n", laddr, haddr);
					else printf("WARNING at %s: relative jump label needed\r\n", laddr);
				}
				sprintf(pBuf, "%s", haddr);
			}
		} else if( NULL==(pRec=FindRecAddress((addr<<16) | 0x0000FFFF)) || pRec->label[0]==0 ) sprintf(pBuf, "%s", haddr);
		else strcpy(pBuf, pGlob->label);
	} else if( flags & FLG_FEQ ) {	// we're forced to search EQUs for
		if( pLineRec->reflabel[0] ) {
			pRec = FindRecEQU(pLineRec->reflabel);
			if( pRec && !(pRec->flags & FLG_EQU_DEFINED) ) {
				pRec->flags |= FLG_EQU_DEFINED;
				pRec->address = addr;
			}
			if( pRec && HP85ADDR(pRec->address)!= addr ) printf("ERROR: .feq mismatch at %s code=%s EQU=%lX\r\n", laddr, haddr, HP85ADDR(pRec->address));
			strcpy(pBuf, pLineRec->reflabel);
		} else {
			if( DoWarnings ) {
				if( multi ) printf("WARNING at %s: multi-byte literal number %s\r\n", laddr, haddr);
				else printf("WARNING at %s: relative jump label needed\r\n", laddr);
			}
			sprintf(pBuf, "%s", haddr);
		}
	} else if( flags & FLG_FDA ) {	// we're forced to search DADs for
		if( pLineRec->reflabel[0] ) {
			pRec = FindRecDAD(pLineRec->reflabel);
			if( pRec && !(pRec->flags & FLG_DAD_DEFINED) ) {
				pRec->flags |= FLG_DAD_DEFINED;
				pRec->address = addr;
			}
			strcpy(pBuf, pLineRec->reflabel);
		} else {
			if( DoWarnings ) {
				if( multi ) printf("WARNING at %s: multi-byte literal number %s\r\n", laddr, haddr);
				else printf("WARNING at %s: relative jump label needed\r\n", laddr);
			}
			sprintf(pBuf, "%s", haddr);
		}
	} else if( flags & FLG_GLO ) {	// we're forced to search GLOs for
		if( pLineRec->reflabel[0] ) {
			pGlob = FindGlobalLabel(pLineRec->reflabel);
			if( pGlob && pGlob->address==addr ) strcpy(pBuf, pLineRec->reflabel);
			else printf("ERROR: GLOBAL not found or address mismatch: .fgl %s %s\r\n", haddr, pLineRec->reflabel);
		} else {
			if( DoWarnings ) {
				if( multi ) printf("WARNING at %s: multi-byte literal number %s\r\n", laddr, haddr);
				else printf("WARNING at %s: relative jump label needed\r\n", laddr);
			}
			sprintf(pBuf, "%s", haddr);
		}
	} else {
		if( NULL!=(pRec=FindRecAddress((addr<<16) | 0x0000FFFF)) && pRec->label[0]!=0 ) strcpy(pBuf, (char *)pRec->label);
		else{
			if( NULL!=(pGlob=FindGlobal(addr, multi && !BaseAddress)) ) strcpy(pBuf, pGlob->label);
			else {
				if( DoWarnings ) {
					if( multi ) printf("WARNING at %s: multi-byte literal number %s\r\n", laddr, haddr);
					else printf("WARNING at %s: relative jump label needed\r\n", laddr);
				}
				sprintf(pBuf, "%s", haddr);
			}
		}
	}
}

/*********************************************************/
void FormatSingleByte(char *pBuf, WORD addr, DISASMREC *pLineRec) {

	DISASMREC	*pRec;
	GLOBAL		*pGlob;
	WORD		flags;
	char		laddr[8], haddr[8];

	flags = (pLineRec==NULL) ? 0 : pLineRec->flags;

	HOaddr(haddr, addr);

	HOaddr(laddr, (unsigned int)HP85ADDR(NxtByt));
	if( flags & FLG_FEQ ) {	// we're forced to search EQUs for
		if( pLineRec->reflabel[0] ) {
			pRec = FindRecEQU(pLineRec->reflabel);
			if( pRec && !(pRec->flags & FLG_EQU_DEFINED) ) {
//				pRec->flags |= FLG_EQU_DEFINED;
//				pRec->address = addr;
				if( DoWarnings ) printf("WARNING at %s: .FEQ not defined: %s\r\n", laddr, pLineRec->reflabel);
			} else if( pRec && HP85ADDR(pRec->address)!= addr ) printf("ERROR: .feq mismatch at %s code=%s EQU=%lX\r\n", laddr, haddr, HP85ADDR(pRec->address));
			strcpy(pBuf, pLineRec->reflabel);
		} else {
			if( DoWarnings && addr>255 ) printf("WARNING at %s: single-byte literal number %o > 255\r\n", laddr, addr);
			if( addr>7 && (flags & FLG_DEC) ) sprintf(pBuf, "%uD", addr);
			else if( addr>7 && (flags & FLG_BCD) ) sprintf(pBuf, "%xC", addr);
			else sprintf(pBuf, "%o", addr);
		}
	} else if( flags & FLG_GLO ) {	// we're forced to search GLOs for
		if( pLineRec->reflabel[0] ) {
			pGlob = FindGlobalLabel(pLineRec->reflabel);
			if( pGlob && pGlob->address==addr ) {
				if( DoWarnings && addr>255 ) printf("WARNING at %s: single-byte literal number %o > 255\r\n", laddr, addr);
				strcpy(pBuf, pLineRec->reflabel);
			} else printf("ERROR: GLOBAL not found or address mismatch: .fgl %s %s\r\n", laddr, pLineRec->reflabel);
		} else {
			if( DoWarnings && addr>255 ) printf("WARNING at %s: single-byte literal number %o > 255\r\n", laddr, addr);
			if( addr>7 && (flags & FLG_DEC) ) sprintf(pBuf, "%uD", addr);
			else if( addr>7 && (flags & FLG_BCD) ) sprintf(pBuf, "%xC", addr);
			else sprintf(pBuf, "%o", addr);
		}
	} else {
		if( DoWarnings && addr>255 ) printf("WARNING at %s: single-byte literal number %o > 255\r\n", laddr, addr);
		if( addr>7 && (flags & FLG_DEC) ) sprintf(pBuf, "%uD", addr);
		else if( addr>7 && (flags & FLG_BCD) ) sprintf(pBuf, "%xC", addr);
		else sprintf(pBuf, "%o", addr);
	}
}

/********************************************************************
 * NxtByt is offset into ROM[] of what to disassemble
 *	(actually, it's ((offset into ROM[])<<16) | c) where 'c' is
 *	0x0000FFFF if we're pointing to actual ROM code, and some
 *	value from 0x00000000 to 0x0000FFFE if we're pointing to a
 *	comment DISASMREC that PRECEDES acutal ROM code.
 * pBuf is pointer to output buffer
 ********************************************************************/
void DisAsm() {

	BYTE	cod1, cod2, cod3, ARPDRP1, ARPDRP2, AD1, AD2;
	char	arpstr[6], drpstr[6], addstr[16], *pForm, *pBuf;
	DWORD	i, k, s, m, w, b;
	WORD	addr;
	DISASMREC *ThisRec;
	DISASMREC *pRec;
	BOOL	haveComment;

/*
if( HP85ADDR(NxtByt)>=0132323 ) {
	i = 1;
}
*/
donextline:
	pBuf = ScratBuf;

//	TargetAddress = 0xFFFFFFFF;

	if( HexOut ) pForm = "%4.4X  %2.2X";
	else pForm = "%6.6o  %3.3o";

	k = HP85ADDR(NxtByt) - BaseAddress;
	sprintf(pBuf, pForm, BaseAddress + k, InBuf[k]);
	k = strlen(pBuf);
	pBuf += k;
	while( k++ < 24 ) *pBuf++ = ' ';
	s = NxtRom = HP85ADDR(NxtByt) - BaseAddress;	// just 0-0xFFFF address in ROM (disregarding full-line comments)
	// try to find DISASMREC for this "line" (NxtByt address) of code
	if( NextRec==NULL ) NextRec = pDis[0];
	while( NextRec != NULL  &&  (NextRec->address < NxtByt || NextRec->type==REC_DAD || NextRec->type==REC_EQU) ) {
		if( NextRec->type < REC_UND ) DisTyp = NextRec->type;
		NextRec = NextRec->pNext;
	}

	if( NextRec!=NULL							// if found a record
	 && NextRec->address > NxtByt				// but it's HIGHER than the one we're looking for
	 && (NxtByt & 0x0000FFFF)!=0x0000FFFF ) {	// and the one we're looking for is NOT code
		if( NextRec->address <= (NxtByt | 0x0000FFFF) ) {	// if record is for next full-line comment OR NxtByt's "real" code
			NxtByt = NextRec->address;							// then jump to there
		} else NxtByt |= 0x0000FFFF;						// else jump NxtByt to "real" code
	}
	saveNxtByt = NxtByt;

	// check for one byte in between tables of DEFs, and force COD type so next table comes out
	// (this is rare, but when it occurs, it can be ugly in the disassembly)
	if( NextRec!=NULL								// if found a DISASMREC after NxtByt
	 && (NextRec->address & 0x0000FFFF)==0x0000FFFF	// and it points to "real" code
	 && (NxtByt & 0x0000FFFF)==0x0000FFFF			// and NxtByt points to "real" code
	 && NextRec->address==(NxtByt+0x00010000)		// and the following DISASMREC "real" code is 1 byte after the NxtByt "real" code
	 && DisTyp==REC_DEF								// and the previous code was a DEF address
	 && NextRec->type!=REC_DEF )					// and the next code is a DEF address
		DisTyp = REC_COD;					// then use REC_COD for this one byte, not REC_DEF, to keep the following table aligned

	if( DisTyp==REC_REM ) DisTyp = LastDisTyp;
	if( NextRec!=NULL && NextRec->address==NxtByt ) {
		ThisRec = NextRec;
// we got a DISASMREC for this line of code
		if( (NxtByt & 0x0000FFFF)!=0x0000FFFF ) {	// not "real" code, just a full-line comment
			++NxtByt;		// so we move to next line next time
			// erase the address/object code that was output above
			for(pBuf=ScratBuf, k=0; k<24; k++) *pBuf++ = ' ';

			*pBuf++ = '!';	// output full-line comment
			*pBuf++ = ' ';
			if( ThisRec->Comment ) strcpy(pBuf, ThisRec->Comment);
			else *pBuf = 0;
			if( DelComments ) goto donextline;
			return;			// all done with this line
		}
		strcpy(pBuf, (char *)ThisRec->label);
		for(cod1=0; cod1<9; cod1++, pBuf++) if( *pBuf==0 ) break;
		for(; cod1<9; cod1++) *pBuf++ = ' ';
		*pBuf = 0;
		if( ThisRec->type != REC_UND ) DisTyp = ThisRec->type;
		// if label (not blanks), set DRP/ARP to undefined
		if( ThisRec->label[0]!=0 ) DisDrp = DisArp = 0xFF;
		DisDataOverride = ThisRec->datover;
	} else {
		ThisRec = NULL;
// no DISASMREC found, output blanks for label
		strcpy(pBuf, "         ");
		pBuf += 9;
		DisDataOverride = 0;
	}
	// output "code", based upon disassembly type
	LastDisTyp = DisTyp;
	switch( DisTyp ) {
	 case REC_ASC:	// ASCII string of "datover" len (max 8 chars)
		if( (i=DisDataOverride)==0 ) i = 1;
		strcpy(pBuf, "ASC  \"");
		pBuf += strlen(pBuf);
		while( i-- && pBuf<(ScratBuf+256-3) ) *pBuf++ = InBuf[NxtRom++];
		*pBuf++ = '"';
		*pBuf = 0;
		break;
	 case REC_ASP:	// ASCII string with parity (most-significant) bit set on last character
		strcpy(pBuf, "ASP  \"");
		pBuf += strlen(pBuf);
		while( !(InBuf[NxtRom] & 128) && pBuf<(ScratBuf+256-3) ) {
			*pBuf++ = InBuf[NxtRom++];
		}
		if( (InBuf[NxtRom] & 0x7F)<32 ) *pBuf++ = ' ';
		else *pBuf++ = InBuf[NxtRom] & 0x7F;
		++NxtRom;
		*pBuf++ = '"';
		*pBuf = 0;
		break;
	 case REC_DEF:	// DEF statement to define an address (as in a table)
	 	addr = (WORD)(((WORD)InBuf[NxtRom]) + (((WORD)InBuf[NxtRom+1])<<8));
	 	FormatAddress(addstr, addr, ThisRec, TRUE);
		sprintf(pBuf, "DEF  %s", addstr);
		++NxtRom;
		++NxtRom;
		break;
	 case REC_DRP:	// stand-alone DRP or ARP
		cod1 = InBuf[NxtRom];
		AD1 = cod1 & 0xC0;
		ARPDRP1 = cod1 & 0x3F;
		goto disarpdrp;
	 case REC_COD:	// "normal" assembly code (CPU instructions)
		cod1 = InBuf[NxtRom];
		AD1 = cod1 & 0xC0;
		ARPDRP1 = cod1 & 0x3F;
		cod2 = InBuf[NxtRom+1];
		AD2 = cod2 & 0xC0;
		ARPDRP2 = cod2 & 0x3F;
		cod3 = InBuf[NxtRom+2];
		if( AD1 & 0x80 ) {	// no ARP or DRP
			if( cod1==0xC6 ) goto jsbinx;
			else if( cod1==0xCE ) {	// JSB =
				FormatAddress(addstr, (WORD)((WORD)cod2+(((WORD)cod3)<<8)), ThisRec, TRUE);
				sprintf(pBuf, "%s =%s", opc[cod1], addstr);
				++NxtRom;
				++NxtRom;
				++NxtRom;
				DisArp = DisDrp = 0xFF;
			} else if( ((cod1 & 0xFC)==0x94)
			 || ((cod1 & 0xF8)==0xA0)
			 || ((cod1 & 0xFC)==0xAC)
			 || ((cod1 & 0xFC)==0xB4)
			 || ((cod1 & 0xFC)==0xBC)
			 || ((cod1 & 0xF8)==0xC0)
			 || ((cod1 & 0xF8)==0xD8)
			 || ((cod1 & 0xF0)==0xE0) ) goto regreg;
			else if( ((cod1 & 0xF8)==0x80)
				  || ((cod1 & 0xFC)==0x88)
				  || ((cod1 & 0xFC)==0x8C)
				  || ((cod1 & 0xFC)==0x90) ) goto reg;
			else if( (cod1 & 0xF0)==0xF0 ) {	// jmp opcodes (JNZ, JNC, etc)
				FormatAddress(addstr, (WORD)(BaseAddress + NxtRom+2+(WORD)((int)((char)cod2))), ThisRec, FALSE);
				sprintf(pBuf, "%s %s", opc[cod1], addstr);
				++NxtRom;
				++NxtRom;
			} else if( (cod1 & 0xF8)==0x98 ) {	// single byte opcodes (BIN,BCD,SAD,DCE,ICE,CLE,RTN,PAD)
				sprintf(pBuf, "%s", opc[cod1]);
				++NxtRom;
				if( cod1==0x9F ) DisArp = DisDrp = 0xFF;	// if PAD, ARP/DRP are unknown
			} else if( ((cod1 & 0xFC)==0xA8) || ((cod1 & 0xF8)==0xC8) ) goto regimm;
			else if( ((cod1 & 0xFC)==0xB0)
				  || ((cod1 & 0xFC)==0xB8)
				  || ((cod1 & 0xF8)==0xD0) ) goto regdir;
			else goto unknown;
		} else if( !(AD1 & 0x80) && !(AD2 & 0x80)	// first 2 bytes are ARP or DRP
		 && ((AD1==0x40 && AD2==0x00) || (AD1==0x00 && AD2==0x40))
		 && (((cod3 & 0xFC)==0x94)
		  || ((cod3 & 0xF8)==0xA0)
		  || ((cod3 & 0xFC)==0xAC)
		  || ((cod3 & 0xFC)==0xB4)
		  || ((cod3 & 0xFC)==0xBC)
		  || ((cod3 & 0xF8)==0xC0)
		  || ((cod3 & 0xF8)==0xD8)
		  || ((cod3 & 0xF0)==0xE0)) ) {
			if( AD1==0x40 ) {
				DisDrp = ARPDRP1;
				DisArp = ARPDRP2;
			} else {
				DisDrp = ARPDRP2;
				DisArp = ARPDRP1;
			}
			++NxtRom;
			++NxtRom;
			cod1 = cod3;
regreg:
			strcpy(pBuf, opc[cod1]);
			FormatReg(drpstr, DisDrp);
			FormatReg(arpstr, DisArp);
			++NxtRom;
			if( (cod1 & 0xF0)==0xE0 ) {
				if( cod1 & 0x02 ) sprintf(pBuf, "%s %s,-%s", opc[cod1], drpstr, arpstr);
				else sprintf(pBuf, "%s %s,+%s", opc[cod1], drpstr, arpstr);
			} else if( (cod1 & 0xFC)==0xB4 || (cod1 & 0xFC)==0xBC ) {
				arpstr[0] = 'X';
				FormatAddress(addstr, (WORD)((WORD)InBuf[NxtRom]+(((WORD)InBuf[NxtRom+1])<<8)), ThisRec, TRUE);
				++NxtRom;
				++NxtRom;
				sprintf(pBuf, "%s %s,%s,%s", opc[cod1], drpstr, arpstr, addstr);
			} else sprintf(pBuf, "%s %s,%s", opc[cod1], drpstr, arpstr);
		} else if( (!(AD1 & 0x80) && !(AD2 & 0x80)) || (AD1==0x40 && cod2==0xCE) ) {	// stand-alone ARP or DRP
disarpdrp:
			if( AD1==0x40 ) DisDrp = ARPDRP1;
			else DisArp = ARPDRP1;
			FormatReg(drpstr, ARPDRP1);
			sprintf(pBuf, "%s %s", opc[cod1], drpstr);
			++NxtRom;
		} else if( AD1==0x00 && cod2==0xC6 ) {	// jsb xREG,ADDR
			DisArp = ARPDRP1;
			cod1 = cod2;
			cod2 = cod3;
			++NxtRom;
jsbinx:
			++NxtRom;
			++NxtRom;
			FormatReg(arpstr, DisArp);
			arpstr[0] = 'X';
			FormatAddress(addstr, (WORD)(((WORD)cod2+(((WORD)InBuf[NxtRom])<<8))), ThisRec, TRUE);
			sprintf(pBuf, "%s %s,%s", opc[cod1], arpstr, addstr);
			++NxtRom;
			DisArp = DisDrp = 0xFF;
		} else if( AD1==0x00
		 && (((cod2 & 0xFC)==0x94)
		  || ((cod2 & 0xF8)==0xA0)
		  || ((cod2 & 0xFC)==0xAC)
		  || ((cod2 & 0xFC)==0xB4)
		  || ((cod2 & 0xFC)==0xBC)
		  || ((cod2 & 0xF8)==0xC0)
		  || ((cod2 & 0xF8)==0xD8)
		  || ((cod2 & 0xF0)==0xE0)) ) {
			DisArp = ARPDRP1;
			++NxtRom;
			cod1 = cod2;
			goto regreg;
		} else if( AD1==0x40
		 && (((cod2 & 0xFC)==0x94)
		  || ((cod2 & 0xF8)==0xA0)
		  || ((cod2 & 0xFC)==0xAC)
		  || ((cod2 & 0xFC)==0xB4)
		  || ((cod2 & 0xFC)==0xBC)
		  || ((cod2 & 0xF8)==0xC0)
		  || ((cod2 & 0xF8)==0xD8)
		  || ((cod2 & 0xF0)==0xE0)) ) {
			DisDrp = ARPDRP1;
			++NxtRom;
			cod1 = cod2;
			goto regreg;
		} else if( AD1==0x40
		 && (((cod2 & 0xF8)==0x80)
		  || ((cod2 & 0xFC)==0x88)
		  || ((cod2 & 0xFC)==0x8C)
		  || ((cod2 & 0xFC)==0x90)) ) {		// single register opcodes (ORB, TSB, etc)
			DisDrp = ARPDRP1;
			++NxtRom;
			cod1 = cod2;
reg:
			FormatReg(drpstr, DisDrp);
			sprintf(pBuf, "%s %s", opc[cod1], drpstr);
			++NxtRom;
		} else if( AD1==0x40
		 && (((cod2 & 0xFC)==0xA8)
		  || ((cod2 & 0xF8)==0xC8)) ) {	// literal immediate (LDB=, CMB=, etc)
			DisDrp = ARPDRP1;
			++NxtRom;
			cod1 = cod2;
			cod2 = cod3;
regimm:
			++NxtRom;
			FormatReg(drpstr, DisDrp);
			if( cod1==0251 && (DisDrp==4 || (ThisRec && (ThisRec->flags & FLG_GTO))) ) {	// LDM R4,=   Change it to GTO xxxx
				++NxtRom;	// skip first byte of address which is in cod2
				FormatAddress(addstr, 1 + (WORD)((WORD)cod2+(((WORD)InBuf[NxtRom])<<8)), ThisRec, TRUE);
				sprintf(pBuf, "GTO %s", addstr);
				++NxtRom;	// skip second byte of address
				DisDrp = DisArp = 0xFF; // DRP & ARP are undefined after a GTO
			} else if( (cod1 & 1) && ((DisDataOverride>1) || (DisDrp<32 && !(DisDrp & 1)) || (DisDrp>=32 && (DisDrp & 7)<7)) ) {	// multibyte
				if( (DisDataOverride==0 && (DisDrp<32 || (DisDrp>=32 && (DisDrp & 7)==6)))
				 || (DisDataOverride==2 && ThisRec && (ThisRec->flags & FLG_FEQ))
				 || (DisDataOverride==2 && ThisRec && (ThisRec->flags & FLG_GLO))
				 || (DisDataOverride==2 && ThisRec && (ThisRec->flags & FLG_FDA))
/* Added 3.19 to implement ".fco address label" to force an immediate multi-byte load on an unknown drp to be a 2-byte LABEL */
				 || (DisDataOverride==2 && ThisRec && (ThisRec->flags & FLG_LABEL_ADDR))

/* Added 2.3 to fix LDM R#,=address showing up as LDM R#,=#,#
				 || (DisDataOverride==2 && DisDrp==0xFF)
   taken back out, reverted to 2.2 July 10, 2016.
*/
					) {
					FormatAddress(addstr, (WORD)((WORD)cod2+(((WORD)InBuf[NxtRom+1])<<8)), ThisRec, TRUE);
					sprintf(pBuf, "%s %s,=%s", opc[cod1], drpstr, addstr);
					++NxtRom;
					++NxtRom;
// The following line tests for LDM R55,=SOMEADDR, and outputs that, ignoring the fact that there's not enough arguments,
// so the user will just have to manually make the extra argument(s) into a BYT argument_list, ugliness. This only appears on 86/87 code.
// Added 3.18.
				} else if( DisDataOverride==0 && (DisDrp>=32 && (DisDrp & 7)<6) && ThisRec && (ThisRec->flags & (FLG_FEQ | FLG_GLO | FLG_FDA)) ) {
					FormatAddress(addstr, (WORD)((WORD)cod2+(((WORD)InBuf[NxtRom+1])<<8)), ThisRec, TRUE);
					sprintf(pBuf, "%s %s,=%s", opc[cod1], drpstr, addstr);
					++NxtRom;
					++NxtRom;
				} else {
					sprintf(pBuf, "%s %s,=", opc[cod1], drpstr);
					if( DisDataOverride==0 ) cod2 = DisDrp & 7;
					else cod2 = 8-DisDataOverride;


					pForm = "%o,";


					while( cod2<7 ) {
						pBuf += strlen(pBuf);
						FormatSingleByte(pBuf, (WORD)InBuf[NxtRom], ThisRec);
						strcat(pBuf, ",");
						++NxtRom;
						++cod2;
					}
					pBuf += strlen(pBuf);
					FormatSingleByte(pBuf, (WORD)InBuf[NxtRom], ThisRec);
					++NxtRom;
				}
			} else {			// single byte
				FormatSingleByte(addstr, cod2, ThisRec);
				pForm = "%s %s,=%s";
				sprintf(pBuf, pForm, opc[cod1], drpstr, addstr);
				++NxtRom;
			}
		} else if( AD1==0x40
		 && (((cod2 & 0xFC)==0xB0)
		  || ((cod2 & 0xFC)==0xB8)
		  || ((cod2 & 0xF8)==0xD0)) ) {
			DisDrp = ARPDRP1;
			cod1 = cod2;
			cod2 = cod3;
			++NxtRom;
regdir:
			++NxtRom;
			++NxtRom;
			FormatReg(drpstr, DisDrp);
			FormatAddress(addstr, (WORD)((WORD)cod2+(((WORD)InBuf[NxtRom])<<8)), ThisRec, TRUE);
			sprintf(pBuf, "%s %s,=%s", opc[cod1], drpstr, addstr);
			++NxtRom;
		} else {
unknown:
			if( !(cod1 & 0x80) ) goto disarpdrp;
			pForm = "BYT  %o";
			sprintf(pBuf, pForm, cod1);
			++NxtRom;
		}
		break;
	 case REC_RAM:
		pForm = "RAM  %o";
		i = (DisDataOverride==0)?1:DisDataOverride;
		sprintf(pBuf, pForm, i);
		NxtRom += i;
		break;
	 case REC_I_O:
		pForm = "I/O  %o";
		i = (DisDataOverride==0)?1:DisDataOverride;
		sprintf(pBuf, pForm, i);
		NxtRom += i;
		break;
	 case REC_BSZ:
	 	if( ThisRec && (ThisRec->flags & FLG_DEC) ) pForm = "BSZ  %uD";
		else if( ThisRec && (ThisRec->flags & FLG_BCD) ) pForm = "BSZ  %xC";
	 	else pForm = "BSZ  %o";
		i = (DisDataOverride==0)?1:DisDataOverride;
		sprintf(pBuf, pForm, i);
		NxtRom += i;
		break;
	 case REC_VAL:
	 	addr = (WORD)InBuf[NxtRom];
	 	FormatAddress(addstr, addr, ThisRec, FALSE);
		sprintf(pBuf, "VAL  %s", addstr);
		++NxtRom;
		break;
	 case REC_BYT:
	 default:
		if( ThisRec && (ThisRec->flags & FLG_FEQ) && ThisRec->reflabel[0] ) {
			pRec = FindRecEQU(ThisRec->reflabel);
			if( pRec ) {
//				strcpy(pBuf, ThisRec->reflabel);
				sprintf(pBuf, "BYT  %s", ThisRec->reflabel);
				pBuf += strlen(pBuf);
				++NxtRom;
				break;
			}
		}
	 	if( ThisRec && (ThisRec->flags & FLG_DEC) ) pForm = "%uD";
		else if( ThisRec && (ThisRec->flags & FLG_BCD) ) pForm = "%xC";
	 	else pForm = "%o";
		if( (i=DisDataOverride)==0 ) i = 1;
		strcpy(pBuf, "BYT  ");
		pBuf += strlen(pBuf);
		while( i-- && pBuf<(ScratBuf+256-3) ) {
			sprintf(pBuf, pForm, InBuf[NxtRom++]);
			pBuf += strlen(pBuf);
			if( i ) *pBuf++ = ',';
		}
		*pBuf = 0;
		break;
	}
	NxtByt = (NxtRom+BaseAddress)<<16;	// leave lower word == 0 in case of full-line comments in front of next "real" code
	// if DISASMREC and comment, output the comment
	haveComment = ThisRec!=NULL && ThisRec->address==saveNxtByt && ThisRec->Comment!=NULL;
	if( (ForceComment>0 || haveComment) && !DelComments ) {
		k = strlen(ScratBuf);
		pBuf = ScratBuf + k;
		do {
			*pBuf++ = ' ';
		} while( k++ < 54 );
		*pBuf++ = '!';
		*pBuf++ = ' ';
		if( haveComment ) strcpy(pBuf, ThisRec->Comment);
		else *pBuf = 0;
	}

	if( HexOut ) {
		b = 6;
		m = 6;
		w = 3;
		pForm = "%2.2X";
	} else {
		b = 8;
		m = 4;
		w = 4;
		pForm = "%3.3o";
	}
	for(k=1; k<min(m, NxtRom-s); k++) {
		sprintf(ScratBuf+b+k*w, pForm, InBuf[s+k]);
		ScratBuf[b+k*w+w-1] = ' ';
	}
	if( isdigit(ScratBuf[0]) || (HexOut && ScratBuf[0]>='A' && ScratBuf[0]<='F') ) ScratBuf[23] = ':';
//	ScratBuf[maxlen] = 0;
}

/***********************************************
 * Format current timestamp into a char buffer *
 ***********************************************/
char* TimeString() {

	static char		timbuf[24];
	char		ampm[4];
	struct tm	*now=NULL;
	int			hour=0;
	time_t		timval=0;

	timval = time(NULL);
	now = localtime(&timval);
	hour = now->tm_hour;
	if( hour>12 ) {
		hour -= 12;
		strcpy(ampm, "PM");
	} else {
		strcpy(ampm, "AM");
	}
	sprintf(timbuf, "%2.2d/%2.2d/%2.2d %2.2d:%2.2d:%2.2d %s", now->tm_mon+1, now->tm_mday, now->tm_year+1900, hour, now->tm_min, now->tm_sec, ampm);

	return timbuf;
}

/****************************************
 * Write a line to the output LST file	*
 ****************************************/
void WriteLine(char *buf) {

	char	*b, *e;
	int		i, t, c;

	if( !OnlySrc || *buf!='.' ) {	// if OnlySrc, discard all DOT-COMMANDS from output
		b = ScratBuf+4096;
		if( OnlySrc ) {
			sprintf(b, "%6.6d ", OutLineNum++);
			for(i=0; i<7 && b[i]=='0'; i++) b[i] = ' ';
			strcat(b, buf+24);
		} else strcpy(b, buf);
		// trim any trailing spaces
		for(e=b + strlen(b); e>b && *(e-1)==' '; --e) ;
		*e = 0;
		// spit warning if line length is EXACTLY 32, 64, 80, or >80

		i = strlen(b);
		if( OnlySrc==1 ) {
			if( DoWarnings && (/*i==32 || i==64 ||*/ i==80) ) {
				strcpy(ScratBuf+6000, b);
				ScratBuf[6032] = 0;
				printf("WARNING: output line len=%d -- %s\r\n", i, ScratBuf+6000);
				strcat(b, " !");
				i += 2;
			}
			if( DoLineLenWarning && i>79 ) printf("WARNING: line %d len > 80\r\n", OutLineNum-1);
		}
		if( OnlySrc==2 ) {
			e = b+7;
			if( *e=='!' ) write(OutFile, e, i-7);	// full line comment
			else if( *e!=' ' ) {						// label
				if( !strnicmp(e+9, "DAD", 3) || !strnicmp(e+9, "EQU", 3) ) {	// DAD/EQU label
					write(OutFile, e, i-7);
				} else {	// non-DAD/EQU label
					for(i=0; i<8; i++) if( e[i]==' ' ) break;	// find space after label
					write(OutFile, e, i);
					write(OutFile, eolseq, eollen);
					e += i;
dumpcode:
					// convert spaces to tabs
					// place EOL-comment at fixed tab column
					t = i = 0;	// t = output column (takes tab-chars into account), i = buffer output (treats tab-chars as just 1 char)
					b[i++] = '\t';
					t = 4;
					while( *e==' ' ) ++e;	// find start of code
					if( strlen(e)>=22 && e[22]=='!' ) {	// have EOL-comment in 'standard' column
						for(c=21; c>0 && e[c-1]==' '; c--) ;// find end of 'code', searching back from EOL-comment
						strncpy(b+i, e, c);
						i += c;
						t += c;
						// now space EOL-comment over to 'standard' column or beyond
						do {
							b[i++] = '\t';
							t = ((t+4)&(~3));
						} while( t<24 );
						strcpy(b+i, e+22);	// copy EOL-comment
					} else {
						strcpy(b+i, e);	// copy whole line, leave EOL-comment alone
/*				write(OutFile, "*****\r\n",7);
				write(OutFile, b, strlen(b));
				write(OutFile, eolseq, eollen);
				write(OutFile, e, strlen(e));
				write(OutFile, eolseq, eollen);
				write(OutFile, "*****\r\n",7);
*/
					}
					write(OutFile, b, strlen(b));
				}
			} else {
				goto dumpcode;
			}
		} else {
			write(OutFile, b, i);
		}
		write(OutFile, eolseq, eollen);
	}
}

/***************************************
 * get a number (octal or dec) or return 0
 ******************************************/
DWORD	GetNum(char **p) {

	char	*pc, *pe;
	DWORD	num;

	num = 0;
	pc = *p;
	SkipWhite(pc);
	for(pe=pc; isdigit(*pe); pe++);
	if( toupper(*pe)=='D' ) {	// decimal
		num = atol(pc);
		pc = pe+1;
	} else {					// octal
		while( *pc>='0' && *pc<='7' ) num = (num<<3) + (*pc++ - '0');
	}
	*p = pc;
	return num;
}

/***************************************
 * get a 16-bit address (octal or hex) or return 0
 * skip:
 *	0 nothing to skip, proceed to get address
 *	1 'p' has been skipped ahead 5, if no address, skip to end-of-line (may be other arguments there, like a label), then proceed to skip to next address, then restore 'p' and return address
 *	5 skip 'p' ahead 5, then either address or EOL expected, scan ahead for a non-DOT line to get the next start-of-line address and return it, leaving 'p' where it is.
 ******************************************/
DWORD	GetAddr(char **p, long skip) {

	char	*pc, *lap, c;
	long	ln;
	DWORD	num;

	num = 0;
	pc = *p;
	if( skip==5 ) pc += 5;
	SkipWhite(pc);
	if( !isdigit(*pc) ) {	// no address there, scan ahead for the next address line and get the address
		if( *pc=='-' ) {
			++pc;	// allow '-' to replace address, causing us to search for the next line with an address, but still allowing successive arguments on the line.
			*p = pc;	// advance the calling ptr past the '-', too
		}
		for(ln=CurLineNum+1; ln<LstLinesCnt; ln++) {
			lap = pLstLines[ln];
			if( isdigit(*lap) ) {	// line starts with number (address)
				if( HexIn ) {
					do {
						c = toupper(*lap)-'0';
						if( c<0 ) break;
						else if( c>9 ) {
							if( c>16 && c<23 ) c -= 7;
							else break;
						}
						num = (num<<4) + c;
						++lap;
					} while( *lap );
				} else {
					while( *lap>='0' && *lap<='7' ) num = (num<<3) + (*lap++ - '0');
				}
				// leave 'p' where it was pointing
				return num;
			}
		}
	}
	SkipWhite(pc);
	if( HexIn ) {
		do {
			c = toupper(*pc)-'0';
			if( c<0 ) break;
			else if( c>9 ) {
				if( c>16 && c<23 ) c -= 7;
				else break;
			}
			num = (num<<4) + c;
			++pc;
		} while( *pc );
	} else {
		while( *pc>='0' && *pc<='7' ) num = (num<<3) + (*pc++ - '0');
	}
	*p = pc;	// skip 'p' ahead to after number
	return num;
}

/*****************************************************
 * Read LstBuf and convert GLOBAL entries to GLOBALS *
 *****************************************************/
void ProcessGLOBALS() {

	int		i, remlen, linenum;//, lastType, dataType, dataCnt;
	char	*bp, label[10];

	// build array of line ptrs into GLOBALS file and replace \r and \n with NUL
	pLST = LstBuf;
	LstLinesCnt = 0;
	while( (remlen = LstBufLen - (pLST-LstBuf)) > 0 ) {
		// save ptr to start of line
		pLstLines[LstLinesCnt++] = pLST;
		// skip to end of line
		while( remlen && *pLST!='\r' && *pLST!='\n' ) {
			++pLST;
			--remlen;
		}
		// skip past end of line sequence and replace it with NUL chars
		if( remlen && *pLST=='\r' ) {
			*pLST++ = 0;
			--remlen;
		}
		if( remlen && *pLST=='\n' ) {
			*pLST++ = 0;
			--remlen;
		}
	}
	// now process the lines, building DISASMRECs for DOT-COMMANDS, LABELS, and COMMENTS
	for(linenum=0; linenum<LstLinesCnt; linenum++) {
		bp = pLstLines[linenum];
		SkipWhite(bp);
		if( *bp && *bp!='*' && *bp!='!' ) {	// lines starting with '*' are full-line comments, so ignore them
			// Get up to 8 chars of label
			for(i=0; i<8 && *bp && *bp!=' ' && *bp!='\t'; ) label[i++] = *bp++;
			label[i] = 0;
			if( i==(mach?8:6) && *bp!=' ' && *bp!='\t' ) printf("ERROR: GLOBAL label too long: %s", label);
			// skip WhiteSpace
			SkipWhite(bp);
			if( !strnicmp(bp, "EQU", 3) ) Globals[GlobalsCnt].type = GLO_EQU;
			else if( !strnicmp(bp, "DAD", 3) ) Globals[GlobalsCnt].type = GLO_DAD;
			else {
				printf("ERROR: GLOBAL unknown statement: %s", pLstLines[linenum]);
				continue;
			}
			if( *bp ) ++bp;	// skip DAD or EQU
			if( *bp ) ++bp;	// skip DAD or EQU
			if( *bp ) ++bp;	// skip DAD or EQU
			SkipWhite(bp);
			Globals[GlobalsCnt].address = GetNum(&bp);	// get address or value
			strcpy(Globals[GlobalsCnt].label, label);
			++GlobalsCnt;
		}
	}
}

/****************************************************
 * Read LstBuf and convert some stuff to DISASMRECs *
 ****************************************************/
void ProcessLST(BOOL IsImport) {

	int		i, remlen, cnt, dadcnt, equcnt, repcnt;
	char	*bp, label[10], comment[256];
	DWORD	dw, lastAddress;
	DISASMREC	*pRec;

	//======================================================================
	// build array of line ptrs into .LST file and replace \r and \n with NUL
	pLST = LstBuf;
	LstLinesCnt = 0;
	while( (remlen = LstBufLen - (pLST-LstBuf)) > 0 ) {
		// save ptr to start of line
		pLstLines[LstLinesCnt++] = pLST;
		// skip to end of line
		while( remlen && *pLST!='\r' && *pLST!='\n' ) {
			++pLST;
			--remlen;
		}
		// skip past end of line sequence and replace it with NUL chars
		if( remlen && *pLST=='\r' ) {
			*pLST++ = 0;
			--remlen;
		}
		if( remlen && *pLST=='\n' ) {
			*pLST++ = 0;
			--remlen;
		}
	}
	// figure out if the .LST file is HEX or OCTAL format
	for(CurLineNum=0; CurLineNum<LstLinesCnt; CurLineNum++) {
		bp = pLST = pLstLines[CurLineNum];
		SkipWhite(bp);
		if( *bp!='.' && *bp!='!' && *bp!=':' ) { // if it's a code line
			if( *(bp+4)==' ' ) HexIn = TRUE;
			else HexIn = FALSE;
			break;
		}
	}
	// now process the lines, building DISASMRECs for DOT-COMMANDS, LABELS, and COMMENTS
	lastAddress = -1;	// in case file starts with full-line comment
	if( !IsImport ) dadcnt = equcnt = 0;
	for(CurLineNum=0; CurLineNum<LstLinesCnt; CurLineNum++) {
		bp = pLST = pLstLines[CurLineNum];
		SkipWhite(bp);
		// process line
		if( *bp=='.' ) { // if it's a command
			if( !strnicmp(bp, ".opt", 4) ) {	// already processed, skip

			} else if( !strnicmp(bp, ".dis", 4) ) {	// discard

			} else if( !strnicmp(bp, ".byt", 4) ) {	// add a REC_BYT record at the specified address
				dw = GetAddr(&bp, 5);	// get BYT address
				if( IsImport ) {
					if( dw>=ImpStart && dw<=ImpEnd ) dw += ImpDelta;
					else continue;
				}
				dw = (dw<<16) | 0x0000FFFF;
				cnt = GetNum(&bp);	// get optional number of bytes (or default to 0)
				if( !cnt ) cnt = 1;
				repcnt = GetNum(&bp);	// get optional REPEAT count
				if( !repcnt ) repcnt = 1;
				while( repcnt-- ) {
					if( NULL==(pRec=FindRecAddress(dw)) ) AddRec(REC_BYT, 0, "", dw, NULL, cnt);
					else if( pRec->type==REC_UND ) {
// change 12-12-16: don't overwrite earlier with later DOT-COMMAND, let earliest one RULE!
						pRec->type = REC_BYT;
						pRec->datover = cnt;
					}
					dw += ((DWORD)cnt << 16);	// bump address for repeating .byt
				}
			} else if( !strnicmp(bp, ".arp", 4) || !strnicmp(bp, ".drp", 4) ) {	// add a REC_DRP record at the specified address
				dw = GetAddr(&bp, 5);	// get DRP address
				if( IsImport ) {
					if( dw>=ImpStart && dw<=ImpEnd ) dw += ImpDelta;
					else continue;
				}
				dw = (dw<<16) | 0x0000FFFF;
				if( NULL==(pRec=FindRecAddress(dw)) ) AddRec(REC_DRP, 0, "", dw, NULL, 0);
// change 12-12-16: don't overwrite earlier with later DOT-COMMAND, let earliest one RULE!
				else if( pRec->type==REC_UND ) pRec->type = REC_DRP;
			} else if( !strnicmp(bp, ".def", 4) ) {	// add a REC_DEF record at the specified address
				dw = GetAddr(&bp, 5);	// get DEF address
				if( IsImport ) {
					if( dw>=ImpStart && dw<=ImpEnd ) dw += ImpDelta;
					else continue;
				}
				dw = (dw<<16) | 0x0000FFFF;
				cnt = GetNum(&bp);	// get possible FLG_ROMJSB_DEF flag
				if( NULL==(pRec=FindRecAddress(dw)) ) {
						AddRec(REC_DEF, cnt, "", dw, NULL, 0);
				} else if( pRec->type==REC_UND ) {
// change 12-12-16: don't overwrite earlier with later DOT-COMMAND, let earliest one RULE!
					pRec->type = REC_DEF;
					pRec->flags |= (cnt ? FLG_ROMJSB_DEF : 0);
				}
			} else if( !strnicmp(bp, ".val", 4) ) {	// add a REC_VAL record at the specified address
				bp += 5;
				dw = GetAddr(&bp, 0);	// get VAL address
				if( IsImport ) {
					if( dw>=ImpStart && dw<=ImpEnd ) dw += ImpDelta;
					else continue;
				}
				SkipWhite(bp);
				for(i=0; i<8 && *bp && *bp!=' ' && *bp!='\t'; ) label[i++] = *bp++;
				label[i] = 0;
				dw = (dw<<16) | 0x0000FFFF;
				if( NULL==(pRec=FindRecAddress(dw)) ) pRec = AddRec(REC_VAL, FLG_GLO, "", dw, NULL, 0);	// find or add a REC_ for this address and set it's FLG_EQU
				else if( pRec->type==REC_UND ) {
					pRec->type = REC_VAL;
					pRec->flags |= FLG_GLO;
				}
				if( pRec!=NULL ) strcpy(pRec->reflabel, label);	// if REC_ found or added, copy 'label' to the 'reflabel'
			} else if( !strnicmp(bp, ".cod", 4) ) {	// add a REC_COD record at the specified address
				dw = GetAddr(&bp, 5);	// get COD address
				if( IsImport ) {
					if( dw>=ImpStart && dw<=ImpEnd ) dw += ImpDelta;
					else continue;
				}
				dw = (dw<<16) | 0x0000FFFF;
				cnt = GetNum(&bp);	// get optional flag (0=normal .cod, 1=force 1-byte stand-alone DRP/ARP instruction)
				if( NULL==(pRec=FindRecAddress(dw)) ) AddRec(REC_COD, 0, "", dw, NULL, cnt);
				else if( pRec->type==REC_UND ) {
// change 12-12-16: don't overwrite earlier with later DOT-COMMAND, let earliest one RULE!
					pRec->type = REC_COD;
					pRec->datover = cnt;
				}
			} else if( !strnicmp(bp, ".asc", 4) ) {	// add a REC_ASC record at the specified address
				bp += 5;
				dw = GetAddr(&bp, 0);	// get ASC address
				if( IsImport ) {
					if( dw>=ImpStart && dw<=ImpEnd ) dw += ImpDelta;
					else continue;
				}
				dw = (dw<<16) | 0x0000FFFF;
				cnt = GetNum(&bp);	// get len of ASC string
				if( NULL==(pRec=FindRecAddress(dw)) ) AddRec(REC_ASC, 0, "", dw, NULL, cnt);
				else if( pRec->type==REC_UND ) {
// change 12-12-16: don't overwrite earlier with later DOT-COMMAND, let earliest one RULE!
					pRec->type = REC_ASC;
					pRec->datover = cnt;
				}
			} else if( !strnicmp(bp, ".asp", 4) ) {	// add a REC_ASP record at the specified address
				dw = GetAddr(&bp, 5);	// get ASP address
				if( IsImport ) {
					if( dw>=ImpStart && dw<=ImpEnd ) dw += ImpDelta;
					else continue;
				}
				dw = (dw<<16) | 0x0000FFFF;
				if( NULL==(pRec=FindRecAddress(dw)) ) AddRec(REC_ASP, 0, "", dw, NULL, 0);
				else if( pRec->type==REC_UND ) pRec->type = REC_ASP;
			} else if( !strnicmp(bp, ".bsz", 4) ) {	// add a REC_BSZ record at the specified address
				bp += 5;
				dw = GetAddr(&bp, 0);	// get BSZ address
				if( IsImport ) {
					if( dw>=ImpStart && dw<=ImpEnd ) dw += ImpDelta;
					else continue;
				}
				dw = (dw<<16) | 0x0000FFFF;
				cnt = GetNum(&bp);	// get len of BSZ block
				if( !cnt ) cnt = 1;
				repcnt = GetNum(&bp);	// get optional REPEAT count
				if( !repcnt ) repcnt = 1;
				while( repcnt-- ) {
					if( NULL==(pRec=FindRecAddress(dw)) ) AddRec(REC_BSZ, 0, "", dw, NULL, cnt);
					else if( pRec->type==REC_UND ) {
						pRec->type = REC_BSZ;
						pRec->datover = cnt;
					}
					dw += ((DWORD)cnt << 16);	// bump address for repeating .byt
				}
			} else if( !strnicmp(bp, ".lab", 4) ) {	// make sure a REC exists with the LABEL at the specified address
				bp += 5;
				dw = GetAddr(&bp, 0);	// get label address
				if( IsImport ) {
					if( dw>=ImpStart && dw<=ImpEnd ) dw += ImpDelta;
					else continue;
				}
				dw = (dw<<16) | 0x0000FFFF;
				SkipWhite(bp);
				for(i=0; i<8 && *bp && *bp!=' ' && *bp!='\t'; ) label[i++] = *bp++;
				label[i] = 0;
				if( NULL==(pRec=FindRecAddress(dw)) ) AddRec(REC_UND, 0, label, dw, NULL, 0);
				else if( pRec->label[0]==0 ) {
					strcpy((char*)pRec->label, label);
					if( CountLabels(pRec->label) > 1 ) printf("ERROR: duplicate label %s\r\n", pRec->label);
				}
			} else if( !strnicmp(bp, ".eol", 4) ) {	// make sure a REC exists with the EOL COMMENT at the specified address
				if( IsImport || IncDotCOM ) {
					bp += 5;
					dw = GetAddr(&bp, 0);	// get comment address
					if( IsImport ) {
						if( dw>=ImpStart && dw<=ImpEnd ) dw += ImpDelta;
						else continue;
					}
					dw = (dw<<16) | 0x0000FFFF;
					if( *bp==' ' || *bp=='\t' ) ++bp;  // skip ONE whitespace, to allow 'indenting' of eol comments
					for(i=0; i<255 && *bp; ) comment[i++] = *bp++;
					if( i==0 ) comment[i++] = ' ';	// make sure comment is not empty
					comment[i] = 0;
					if( NULL==(pRec=FindRecAddress(dw)) ) AddRec(REC_UND, 0, "", dw, comment, 0);
					else if( pRec->Comment==NULL ) {
						pRec->Comment = (char *)malloc(strlen(comment)+1);
						if( pRec->Comment ) strcpy(pRec->Comment, comment);
					} // ignore the DOT-COMMAND comment if there's already a comment stored for this line
				}
			} else if( !strnicmp(bp, ".com", 4) ) {	// make sure a REC exists with the FULL-LINE COMMENT at the specified address
				if( IsImport || IncDotCOM ) {
					bp += 5;
					dw = GetAddr(&bp, 0);	// get comment address
					if( IsImport ) {
						if( dw>=ImpStart && dw<=ImpEnd ) dw += ImpDelta;
						else continue;
					}
					dw = (dw<<16) & 0xFFFF0000;
					SkipWhite(bp);
					cnt = GetNum(&bp); // sequence# of this full-line comment at this address
					if( *bp==' ' || *bp=='\t' ) ++bp;  // skip ONE whitespace, to allow 'indenting' of comments
					for(i=0; i<255 && *bp; ) comment[i++] = *bp++;
					if( i==0 ) comment[i++] = ' ';	// make sure comment is not empty
					comment[i] = 0;
					if( NULL==(pRec=FindRecAddress(dw)) ) {
						AddRec(REC_REM, 0, "", dw, comment, 0);
					} else {
						do {
							if( pRec->datover==cnt ) {
								break; // there's already a comment with this index
							}
							if( pRec->pNext==NULL || ((pRec->pNext)->address & 0xFFFF0000)!=dw || ((pRec->pNext)->address & 0x0000FFFF)>=0x00008000 ) {
								dw = pRec->address + 1;	// put this as the next full-line comment
								if( (pRec->address & 0x00008000)==0 ) {
									AddRec(REC_REM, 0, "", dw, comment, dw & 0x00007FFF); // add the comment as long as there are less than 32K of them.  Heh-heh.
								}
								break;
							}
							pRec = pRec->pNext;
						} while( pRec != NULL );
					}
				}
			} else if( !strnicmp(bp, ".feq", 4) ) {	// make sure a REC_EQU exists with the reflabel
				bp += 5;
				dw = GetAddr(&bp, 1);	// get ref address
				if( IsImport ) {
					if( dw>=ImpStart && dw<=ImpEnd ) dw += ImpDelta;
					else continue;
				}
				SkipWhite(bp);
				for(i=0; i<8 && *bp && *bp!=' ' && *bp!='\t'; ) label[i++] = *bp++;
				label[i] = 0;
				dw = (dw<<16) | 0x0000FFFF;
				if( NULL==(pRec=FindRecEQU(label)) ) pRec = AddRec(REC_EQU, 0, label, 0, NULL, 0);	// create the Rec but FLG_EQU_DEFINED not set yet
				if( NULL==(pRec=FindRecAddress(dw)) ) pRec = AddRec(REC_UND, FLG_FEQ, "", dw, NULL, 0);	// find or add a REC_ for this address and set it's FLG_FEQ
				else pRec->flags |= FLG_FEQ;
				if( pRec!=NULL ) strcpy(pRec->reflabel, label);	// if REC_ found or added, copy 'label' to the 'reflabel'
			} else if( !strnicmp(bp, ".fda", 4) ) {	// make sure a REC_DAD exists with the reflabel
				bp += 5;
				dw = GetAddr(&bp, 1);	// get ref address
				if( IsImport ) {
					if( dw>=ImpStart && dw<=ImpEnd ) dw += ImpDelta;
					else continue;
				}
				SkipWhite(bp);
				for(i=0; i<8 && *bp && *bp!=' ' && *bp!='\t'; ) label[i++] = *bp++;
				label[i] = 0;
				dw = (dw<<16) | 0x0000FFFF;
				if( NULL==(pRec=FindRecDAD(label)) ) {
					pRec = AddRec(REC_DAD, 0, label, 0, NULL, 0);	// create the Rec but FLG_DAD_DEFINED not set yet
				}
				if( NULL==(pRec=FindRecAddress(dw)) ) {
					pRec = AddRec(REC_UND, FLG_FDA, "", dw, NULL, 0);	// find or add a REC_ for this address and set it's FLG_FDA
				} else {
					pRec->flags |= FLG_FDA;
				}
				if( pRec!=NULL ) strcpy(pRec->reflabel, label);	// if REC_ found or added, copy 'label' to the 'reflabel'
			} else if( !strnicmp(bp, ".fgl", 4) ) {	// make sure a REC_GLO exists with the reflabel
				bp += 5;
				dw = GetAddr(&bp, 1);	// get ref address
				if( IsImport ) {
					if( dw>=ImpStart && dw<=ImpEnd ) dw += ImpDelta;
					else continue;
				}
				SkipWhite(bp);
				for(i=0; i<8 && *bp && *bp!=' ' && *bp!='\t'; ) label[i++] = *bp++;
				label[i] = 0;
				dw = (dw<<16) | 0x0000FFFF;
				if( NULL==(pRec=FindRecAddress(dw)) ) pRec = AddRec(REC_UND, FLG_GLO, "", dw, NULL, 0);	// find or add a REC_ for this address and set it's FLG_FDA
				else pRec->flags |= FLG_GLO;
				if( pRec!=NULL ) strcpy(pRec->reflabel, label);	// if REC_ found or added, copy 'label' to the 'reflabel'
/* Added 3.19 to implement ".fco address label" to force an immediate multi-byte load on an unknown drp to be a 2-byte LABEL */
			} else if( !strnicmp(bp, ".fco", 4) ) {	// make sure a REC_COD exists and set the FLG_LABEL_ADDR flag
				bp += 5;
				dw = GetAddr(&bp, 1);	// get ref address
				if( IsImport ) {
					if( dw>=ImpStart && dw<=ImpEnd ) dw += ImpDelta;
					else continue;
				}
				SkipWhite(bp);
				for(i=0; i<8 && *bp && *bp!=' ' && *bp!='\t'; ) label[i++] = *bp++;
				label[i] = 0;
				dw = (dw<<16) | 0x0000FFFF;
				if( NULL==(pRec=FindRecAddress(dw)) ) pRec = AddRec(REC_UND, FLG_LABEL_ADDR, "", dw, NULL, 0);	// find or add a REC_ for this address and set it's FLG_LABEL_ADDR

				if( pRec!=NULL ) {
					pRec->flags |= FLG_LABEL_ADDR;
					strcpy(pRec->reflabel, label);	// if REC_ found or added, copy 'label' to the 'reflabel'
				}
			} else if( !strnicmp(bp, ".gto", 4) ) {	// make sure a REC_COD exists with FLG_GTO set
				bp += 5;
				dw = GetAddr(&bp, 0);	// get address at which the GTO exists
				if( IsImport ) {
					if( dw>=ImpStart && dw<=ImpEnd ) dw += ImpDelta;
					else continue;
				}
				dw = (dw<<16) | 0x0000FFFF;
				if( NULL==(pRec=FindRecAddress(dw)) ) pRec = AddRec(REC_COD, FLG_GTO, "", dw, NULL, 0);	// find or add a REC_ for this address and set it's FLG_GTO
				else pRec->flags |= FLG_GTO;
			} else if( !strnicmp(bp, ".dec", 4) ) {	// make sure a REC_COD or REC_BYT exists with FLG_DEC set
				dw = GetAddr(&bp, 5);	// get address
				if( IsImport ) {
					if( dw>=ImpStart && dw<=ImpEnd ) dw += ImpDelta;
					else continue;
				}
				dw = (dw<<16) | 0x0000FFFF;
				if( NULL==(pRec=FindRecAddress(dw)) ) {
					pRec = AddRec(REC_UND, FLG_DEC | FLG_IGNORE_DB, "", dw, NULL, 0);	// find or add a REC_ for this address and set it's FLG_DEC
				} else if( !(pRec->flags & FLG_IGNORE_DB) ) pRec->flags |= FLG_DEC;
			} else if( !strnicmp(bp, ".bcd", 4) ) {	// make sure a REC_COD or REC_BYT exists with FLG_BCD set
				dw = GetAddr(&bp, 5);	// get address
				if( IsImport ) {
					if( dw>=ImpStart && dw<=ImpEnd ) dw += ImpDelta;
					else continue;
				}
				dw = (dw<<16) | 0x0000FFFF;
				if( NULL==(pRec=FindRecAddress(dw)) ) {
					pRec = AddRec(REC_UND, FLG_BCD | FLG_IGNORE_DB, "", dw, NULL, 0);	// find or add a REC_ for this address and set it's FLG_BCD
				} else if( !(pRec->flags & FLG_IGNORE_DB) ) pRec->flags |= FLG_BCD;
			}
		} else if( !IsImport ) {	// else normal code line in LST file (only do DOT-commands in IMP file
			if( *bp==':' ) {	// DAD or EQU
				++bp;	// skip colon
				SkipWhite(bp);
				if( *bp && *bp!=' ' && *bp!='\t' ) {	// if LABEL exists
					for(i=0; i<8 && *bp && *bp!=' ' && *bp!='\t'; ) label[i++] = *bp++;	// get label
					label[i] = 0;
					SkipWhite(bp);
					if( !strnicmp(bp, "DAD", 3) ) {
						bp += 3;
						SkipWhite(bp);
						dw = (GetNum(&bp)<<16) | (0x00008000 + dadcnt++);
						SkipWhite(bp);
						if( *bp!=':' ) {
							printf("ERROR: DAD option ROM# missing on line %d", CurLineNum);
							cnt = 0;
						} else {
							++bp;
							cnt = GetNum(&bp);
						}
						SkipWhite(bp);
						if( *bp=='!' ) {
							++bp;	// skip the !
							if( *bp==' ' || *bp=='\t' ) ++bp;  // skip ONE whitespace, to allow 'indenting' of eol comments
							for(i=0; i<255 && *bp; ) comment[i++] = *bp++;
							if( i==0 ) comment[i++] = ' ';	// make sure comment is not empty
							comment[i] = 0;
						} else comment[0] = 0;
						if( DelComments ) comment[0] = 0;
						AddRec(REC_DAD, 0, label, dw, comment[0]==0?NULL:comment, cnt);
					} else if( !strnicmp(bp, "EQU", 3) ) {
						bp += 3;
						SkipWhite(bp);
						dw = (GetNum(&bp)<<16) | (0x0000C000 + equcnt++);
						SkipWhite(bp);
						if( *bp=='!' ) {
							++bp;	// skip the !
							if( *bp==' ' || *bp=='\t' ) ++bp;  // skip ONE whitespace, to allow 'indenting' of eol comments
							for(i=0; i<255 && *bp; ) comment[i++] = *bp++;
							if( i==0 ) comment[i++] = ' ';	// make sure comment is not empty
							comment[i] = 0;
						} else comment[0] = 0;
						pRec = AddRec(REC_EQU, FLG_EQU_DEFINED, label, dw, comment[0]==0?NULL:comment, 0);
					} else printf("ERROR on line %d: something bogus", CurLineNum+1);
				} // if colon but no label, bogus, ignore
			} else if( *bp=='!' ) {	// full-line comment
				++bp;
				if( *bp==' ' ) {
					++bp;	// skip possible first SPACE after '!'
				}
				if( *bp==0 ) {
					strcpy(comment, " ");	// blank line, make it a comment with a space
				} else {
					for(i=0; i<sizeof(comment)-1 && *bp; ) comment[i++] = *bp++;
					comment[i] = 0;
				}
				// create a record with this full-line comment
				if( lastAddress==-1 ) {
					lastAddress = (BaseAddress<<16);
				} else if( (lastAddress & 0xFFFF)==0xFFFF ) {	// if first full-line comment after a line of real code
					for(i=CurLineNum+1; i<LstLinesCnt && !(isdigit(pLstLines[i][0]) || (HexIn && toupper(pLstLines[i][0])>='A' && toupper(pLstLines[i][0])<='F')); i++) ;	// find next line with real code
					if( i<LstLinesCnt ) {
						bp = pLstLines[i];
						dw = GetAddr(&bp, 0);
						lastAddress = (dw<<16);
					} else lastAddress = (lastAddress & 0xFFFF0000)+0x00010000;
				} else ++lastAddress;	// Not first full-line comment in this block, just bump the line# up
				AddRec(REC_REM, 0, "", lastAddress, comment, lastAddress & 0x00007FFF);
			} else {	// must be line with address at start (code line)
				dw = GetAddr(&bp, 0);	// get address
				lastAddress = (dw<<16) | 0x0000FFFF;	// combined address with 0xFFFF in lower WORD meaning REAL code, not COMMENT

				while( *bp && *bp++!=':' ) ;	// skip past object codes and (hopefully) ':'
				if( *bp && *bp!=' ' && *bp!='\t' ) {	// if LABEL exists
					for(i=0; i<8 && *bp && *bp!=' ' && *bp!='\t'; ) label[i++] = *bp++;	// get label
					label[i] = 0;
				} else label[0] = 0;
				SkipWhite(bp);	// skip spacing between label and opcode

				// check for and get end-of-line comment
				comment[0] = 0;
				while( *bp ) {
					while( *bp && *bp!='!' ) ++bp;
					if( *bp=='!' ) {
						if( *(bp-1)==' ' || *(bp-1)=='\t' ) {	// found eol comment, save it
							++bp;
							if( *bp==' ' ) ++bp;
							for(i=0; i<sizeof(comment)-1 && *bp; ) comment[i++] = *bp++;
							comment[i] = 0;
						} else ++bp;
					}
				}
				if( DelComments ) comment[0] = 0;
				if( NULL==(pRec=FindRecAddress(lastAddress)) ) {
					if( label[0] || comment[0] ) {
						AddRec(REC_UND, 0, label, lastAddress, comment, 0);
					}
				} else {	// rec already created, must have been from a DOT-COMMAND, just add a few things to it
					if( label[0] ) {
						strcpy((char *)pRec->label, label);
						if( CountLabels(pRec->label) > 1 ) printf("ERROR: duplicate label %s\r\n", pRec->label);
					}
					if( comment[0] ) {
						if( pRec->Comment ) free(pRec->Comment);	// if DOT-COMMAND gave this line a comment, let the (possibly) edited comment overwrite it
						pRec->Comment = (char *)malloc(strlen(comment)+1);
						if( pRec->Comment ) strcpy(pRec->Comment, comment);
					}
				}
			}
		}
	}
}

DWORD GetOctal(char *pA) {

	DWORD	retval;

	for(retval=0; *pA>='0' && *pA<'8'; pA++) retval = (retval<<3) + (*pA-'0');
	return retval;
}

//*****************************************************************
// Adds DOT-COMMAND lines to LinBuf when creating initial .LST file
void AddLst(char **ppL, char *str) {

	int		len;
	char	*pL;

	pL = *ppL;
	len = strlen(str);
	if( (pL-LstBuf)+len < LstBufLen ) {
		strcpy(pL, str);
		pL += len;
		*pL++ = '\r';
		*pL++ = '\n';
		*ppL = pL;
	}
}

/********************************
 * SYNTAX: HP80DIS <input_name>	*
 ********************************/
/**
This is a horrible block of code that grew and grew, like a cancer,
as more and more features were added.  It works, though, for the most part.
Here is an overview of what the main() function does and in what order:

1) Build path for default GLOBALS files (same as .exe folder).
2) Chew up command line arguments, spit SYNTAX if errors.
3) Build the .LST file name.
4) Read and process the .opt line from the .LST file.
5) Set any operating variables that .opt SET but which command line did NOT.
6) Set any non-initialized operating variables to default values.
7) Load the INPUT file (ROM, BPGM, LEX, etc.)
8) Do preliminary setup/processing of INPUT file, trashing headers,
   adjusting length, setting base address, etc.
9) Load and process the GLOBALS file (either local or from .exe folder).
10) If not .LST file, generate preliminary one in memory based upon INPUT,
    setting up default .DOT commands for most obvious things.
    Else, IF .LST file, load it and make appropriate BACKUP copy (-B option).
11) Process the .LST file, setting up DISASMRECs for everything.
12) If IMPORTING .DOT commands from another .LST file, load and process it.
13) Create temporary working OUTPUT file, output .opt and .dis header lines.
14) Dump all the DADs and EQUs up front.
15) Output ABS or NAM line if 85/87.
16) Diassembler INPUT file (loop on DisAsm()).
17) Output FIN line
18) Dump all .DOT commands.
19) Delete OLD .LST or .SRC file (if there was one) and rename the temporary
    working OUTPUT file to .LST or .SRC name.
20) Take a nap.

The Devil, of course, is in the details.  And the details are largely in
ProcessLst() and DisAsm().
**/
int main(int argc, char **argv) {

	int     	i, j, k, g, fullLineCnt=0;
	int     	fp;
	DISASMREC	*pRec;
	DWORD		dw, rp, tp, rep, pp, lastFullLineAddr;
	WORD		addrROMJSB, addrERROR, addrERRORplus, addrSYSJSB, addrMELJSB, addrWCKJSB;
	char		*pL, label[10], *bp, *pA, haddr[8];

#if DEVTIME
if( argc==1 ) {
	strcpy(BinFileName, "romsys1B");
//	strcpy(BinFileName, "rom050");
	strcpy(ListFileName, BinFileName);
	strcat(ListFileName, ".lst");
	printf("%s\r\n", TitleRevStr);
	IsROM = 0;
	IsSystemROM = TRUE;
	mach = MACH_85;
	BaseAddress = 8192*IsROM;	// 0, 20000, 40000, or 60000 octal
	EndAddress = BaseAddress+8192;
	OnlySrc = 0;
	DelComments = FALSE;
	ForceComment = 1;
	LocalGlobal = 1;
} else {
#endif

#ifdef __linux__
	GlobalFileName[0] = 0;
#else
	// construct PATH to globals file, ending in '\' or ':', ready to have filename appended later
	i = GetModuleFileName(NULL, GlobalFileName, PATHLEN-1);

	while( i && GlobalFileName[i]!='\\' && GlobalFileName[i]!=':' ) --i;
	if( GlobalFileName[i]=='\\' || GlobalFileName[i]==':' ) ++i;
	if( !i || GlobalFileName[i-1]!='\\' ) GlobalFileName[i++] = '\\';
	GlobalFileName[i] = 0;
#endif

	// check input arguments and optionally print syntax
	for(i=1; i<argc; i++) {
		if( argv[i][0]=='-' || argv[i][0]=='/' ) {
			switch( argv[i][1] ) {
			 case '7':	// 75 hopefully
			 	if( mach<0 ) {
					if( argv[i][2]!='5' ) goto defopt;
					mach = MACH_75;
					goto hout;
			 	} else {
multmodel:
			 		printf("ERROR: multiple MODEL options %s\r\n", argv[i]);
			 	}
			 	break;
			 case '8':	// 85 or 87, hopefully
			 	if( mach<0 ) {
					if( argv[i][2]=='5' ) mach = MACH_85;
					else if( argv[i][2]=='7' ) mach = MACH_87;
					else goto defopt;
hout:
					if( toupper(argv[i][3])=='X' ) HexOut = TRUE;
			 	} else goto multmodel;
				break;
			 case 'k':	// include .COM full-line comments
			 	IncDotCOM = TRUE;
			 	break;
			 case 'g':	// local global file
			 	LocalGlobal = 1;
			 	break;
			 case 'G':	// global file where CAPDIS.EXE is located
				LocalGlobal = 0;
				break;
			 case 'r':	// force ROM (default)
			 	if( IsROM<0 ) {
					IsROM = atoi(argv[i]+2);
					if( IsROM<0 || IsROM>7 ) goto defopt;
			 	} else printf("ERROR: multiple R/B options %s\r\n", argv[i]);
				break;
			 case 'b':	// force BPGM
			 	if( IsROM<0 ) {
					IsROM = 8;
			 	} else printf("ERROR: multiple R/B options %s\r\n", argv[i]);
				break;
			 case 'B':	// set auto-backup limit (0==OFF, 1-9 maximum number of backups before reuse)
			 	BackupMax = argv[i][2]-'0';
			 	if( BackupMax<0 || BackupMax>9 ) {
					printf("ERROR: unexpected -B value, expect 0 to 9.  Using default (5)\r\n");
					BackupMax = 5;
			 	}
				break;
			 case 's':	// output SOURCE ONLY (85/87 Assembler ROM format)
			 	OnlySrc = 1;
				break;
			 case 'S':	// output SOURCE ONLY (ASM85 Assembler format)
			 	OnlySrc = 2;
				break;
			 case 'c':	// force end-of-line COMMENT on every line (ie, add a '!')
			 	ForceComment = 1;
				break;
			 case 'C':	// force NO end-of-line COMMENT on every line
			 	ForceComment = 0;
				break;
			 case 'q':	// QUIET - no warnings
			 	DoWarnings = 0;
				break;
			 case 'Q':	// NOT QUIET - YES warnings
			 	DoWarnings = 1;
				break;
			 case 'l':	// -s line length>80 warnings
			 	DoLineLenWarning = 1;
				break;
			 case 'd':	// strip ALL comments from -s source code output
			 	DelComments = 1;
			 	break;
			 case 'i':	// IMPORT labels and comments from a section of code in the .IMP file
				LstStart = GetOctal(argv[i]+2);
				for(pA=argv[i]; *pA && *pA!='='; pA++);
				if( *pA==0 ) {
ierr:
					printf("\r\n-i bad format: -iLLLLLL=IIIIII:EEEEEE\r\n");
					exit(5);
				}
				++pA;
				ImpStart = GetOctal(pA);
				for(; *pA && *pA!=':'; pA++);
				if( *pA==0 ) goto ierr;
				++pA;
				ImpEnd = GetOctal(pA);
				ImpDelta = LstStart - ImpStart;
				break;
			 default:
defopt:
				printf("ERROR: unknown or invalid option: %s\r\n", argv[i]);
defnam:
		printf("\r\n\
%s - HP Capricorn CPU - Disassemble, comment, label, repeat\r\n\
  CAPDIS [options] <machine_code_filename> [options]\r\n\
    [options] =\r\n\
       -85[x] | -87[x] | -75[x] Select the machine and GLOBALS file.\r\n\
                     Optional 'x' suffix forces HEX output, OCTAL is default.\r\n\
       -b | -r<0-7>  Force BPGM/LEX or ROM disassembly.\r\n\
                     NOTE: -r MUST be followed by a single digit from 0-7\r\n\
                     which is equal to (ROM_base_address / 8192).\r\n\
       -B<0-9>       Set auto-backup limit (0=OFF, else cycles 0 to limit-1)\r\n\
       -s            Output ONLY source code to .SRC in ASM ROM format\r\n\
       -S            Output ONLY source code to .SRC in ASM85 format\r\n\
       -c            Add '!' to end of uncommented lines\r\n\
       -C            Override -c in .LST file's .opt line, turns auto-'!'s OFF\r\n\
       -d            Delete ALL comments from -s source\r\n\
                     NOTE: -d will still add a '!' to the end of lines that are\r\n\
                     exactly 32, 64, or 80 chars long unless -q is specified.\r\n\
       -q            Quiet - don't show WARNINGS (except -l)\r\n\
       -Q            Override -q in .LST file's .opt line, turns warnings ON\r\n\
       -l            During -s warn about line lengths > 79\r\n\
       -g            Use GLOBALS file from current directory\r\n\
       -G            Use GLOBALS file from location of CAPDIS.EXE (default)\r\n\
       -k            Keep comments from .DOT commands, else just from source\r\n\
    NOTE: defaults are \"-85 -b\" with all other options OFF.  Also, if\r\n\
    an option is not specified on the command line, it may be specified on\r\n\
    the first line of the .LST file using the \".OPT\" DOT-COMMAND.  Any options\r\n\
    specified on the command line override the same .OPT option in the .LST.\r\n\
", TitleRevStr);
// undocumented feature:
// to copy (with address adjustments) comments and labels from one file to another (similar source codes)
// rename the IMPORT .lst file to match the current .lst file name but with a .imp file extension.
// Then run HP80DIS with -iLLLLLL=IIIIII:EEEEEE where:
//	LLLLLL is the octal starting address in the current .LST file that...
//  IIIIII is the octal starting address in the .IMP file matching LLLLLL and...
//	EEEEEE is the octal ending address in the .IMP file that matches code in the current .LST file.
				exit(1);
			}
		} else if( BinFileName[0]==0 ) {
			strcpy(BinFileName, argv[i]);
			strcpy(ListFileName, argv[i]);
			strcpy(ImportFileName, argv[i]);
		} else {
			printf("ERROR: too many filenames: %s", argv[i]);
			exit(1);
		}
	}
	// build the .LST file name
	for(i=strlen(ListFileName)-1; i && ListFileName[i]!='.'; i--) ;
	if( i>0 && ListFileName[i]=='.' ) strcpy(ListFileName+i, ".lst");
	else strcat(ListFileName, ".lst");

	// try to read the first line of the .LST file to get .OPTions
	if( -1!=(fp=open(ListFileName, _O_BINARY | _O_RDONLY)) ) {
		i = read(fp, ScratBuf, 256);
		ScratBuf[256] = 0;
		close(fp);

		pL = ScratBuf;
		while( *pL==' ' || *pL=='\t' ) ++pL;

		if( !strnicmp(pL, ".opt", 4) ) {	// OPTIONS
			pL += 4;
			do {
				while( *pL==' ' || *pL=='\t' ) ++pL;
				if( *pL!='-' ) break;
				++pL;
				switch( *pL ) {
				 case '7':	// 75 hopefully
				 	++pL;
					if( *pL=='5' ) {
						LstMach = MACH_75;
Lhout:
						++pL;
						if( toupper(*pL)=='X' ) {
							HexOut = TRUE;
							++pL;
						}
					} else goto Lopt;
					break;
				 case '8':	// 85 or 87, hopefully
				 	++pL;
					if( *pL=='5' ) {
						LstMach = MACH_85;
						goto Lhout;
					} else if( *pL=='7' ) {
						LstMach = MACH_87;
						goto Lhout;
					} else goto Lopt;
					break;
				 case 'r':	// force ROM (default)
				 	++pL;
					LstIsROM = atoi(pL);
					++pL;
					if( LstIsROM<0 || LstIsROM>7 ) goto Lopt;
					break;
				 case 'b':	// force BPGM
				 	++pL;
					LstIsROM = 8;
					break;
				 case 'B':	// set auto-backup limit (0==OFF, 1-9 maximum number of backups before reuse)
				 	++pL;
					LstBackupMax = (*pL++)-'0';
					if( LstBackupMax<0 || LstBackupMax>9 ) LstBackupMax = 5;
					if( *pL!=':' ) printf("ERROR (.opt): Expected '-Bl:n' where 'l' is Limit and 'n' is Number of backup\r\n");
					else {
						++pL;
						LstBackupNum = (*pL++)-'0';
						if( LstBackupNum<0 || LstBackupNum>=LstBackupMax ) LstBackupNum = 0;
					}
					break;
				 case 'g':	// force LOCAL GLOBALS
				 	++pL;
					LstLocalGlobal = 1;
					break;
				 case 'G':	// force GLOBALS from CAPDIS.EXE location
				 	++pL;
					LstLocalGlobal = 0;
					break;
				 case 'c':	// force end-of-line COMMENT on every line (ie, add a '!')
				 	++pL;
					LstForceComment = 1;
					break;
				 case 'C':
				 	++pL;
					LstForceComment = 0;
					break;
				 case 'q':	// QUIET - no warnings
				 	++pL;
					LstDoWarnings = 0;
					break;
				 case 'Q':
				 	++pL;
					LstDoWarnings = 1;
					break;
				 default:
Lopt:
					printf("ERROR: Universe-ending mistake on .OPT line");
					*pL = 0;
					break;
				}
			} while( TRUE );

		}
		if( mach<0 ) mach = LstMach;
		if( IsROM<0 ) IsROM = LstIsROM;
//printf("1 ForceComment=%d LstForceComment=%d\r\n", ForceComment, LstForceComment);
		if( ForceComment<0 ) ForceComment = LstForceComment;
		if( DoWarnings<0 ) DoWarnings = LstDoWarnings;
		if( LocalGlobal<0 ) LocalGlobal = LstLocalGlobal;
		if( BackupMax<0 ) BackupMax = LstBackupMax;
		if( BackupNum<0 ) BackupNum = LstBackupNum;
	}
#if DEVTIME
}
#endif
	if( BinFileName[0]==0 ) goto defnam;

	if( IsROM<0 ) IsROM = 8;	// default to BPGM
	// DoWarnings is ON unless specifically set to 0, so -1 or 1 gives warnings
	if( DoWarnings<0 ) DoWarnings = 1;
	// ForceComment is OFF unless specifically set to 1, so -1 or 0 does NOT add comments
	if( ForceComment<0 ) ForceComment = 0;
	if( mach<0 ) mach = MACH_85;
	if( BackupMax<0 ) {
		BackupMax = 5;	// default
		BackupNum = 0;
	} else if( BackupNum<0 || BackupNum>=BackupMax ) BackupNum = 0;

	//======================================================================
	// LOAD ROM/BINARY PROGRAM
	// open ROM/BPGM file for reading, allocate buffer, read the file in
	if( -1==(fp=open(BinFileName, _O_BINARY | _O_RDONLY)) ) {
		printf("Error reading %s\r\n", BinFileName);
		exit(5);
	}
	InBufLen = lseek(fp, 0, SEEK_END);
	InBuf = (unsigned char *)malloc(InBufLen);
	if( InBuf == NULL ) {
		printf("Error allocating input buffer\r\n");
		exit(6);
	}
	lseek(fp, 0, SEEK_SET);
	read(fp, (void*)InBuf, InBufLen);
	close(fp);

	//======================================================================
	// ROM/BINARY PROGRAM SETUP
	if( IsROM<8 ) {
		curROM = (int)(*InBuf);
		if( IsROM<3 || (IsROM==3 && curROM==0) ) IsSystemROM = TRUE; // if system "plug-in" rom 0
		BaseAddress = 8192*IsROM;	// 0, 20000, 40000, or 60000, etc. (octal)
		EndAddress = BaseAddress+8192;
	} else {
//GEEKDO: handle ABS binary programs, and any diffs between 85/87
		// see if has LIFUTIL header on it, if so, remove it
		if( *((WORD*)InBuf)==0x0080 && !strncmp((char*)InBuf+2, "HFSLIF", 6) ) {	// LIF disk
			if( mach!=MACH_75 ) InBufLen = 512 + *((WORD*)(InBuf+256+28));	// clip end for actual length of file
			InBufLen -= 512;
			memcpy(InBuf, InBuf+512, InBufLen);

/*
if( -1!=(OutFile=open("GeekBPGM", _O_BINARY | _O_RDWR | _O_CREAT | _O_TRUNC, _S_IREAD | _S_IWRITE)) ) {
	char	bubba[64];

	sprintf(bubba, "InBufLen=%ld", InBufLen);
	MessageBox(NULL, bubba, "DEBUG", MB_OK);
	write(OutFile, InBuf, InBufLen);
	close(OutFile);
}
*/
		}
		// see if has 32-byte 75-LIF header, and remove it
		if( mach==MACH_75 && *((WORD*)(InBuf+10))==0x89E0 && *((WORD*)(InBuf+26))==0x0180 ) {
			memcpy(InBuf, InBuf+32, InBufLen-32);
			InBufLen -= 32;
//			InBufLen = 18 + *((WORD*)(InBuf+2));	// clip end for actual length of file
		}
		if( mach==MACH_87 ) {	// HP87 BPGM, check if PROGRAM CONTROL BLOCK says BPGM length is less than InBufLen
			if( *((WORD*)(InBuf+4)) < InBufLen ) InBufLen = *((WORD*)(InBuf+4));
		}
		if( mach!=MACH_75 ) {	// HP85/HP87 BPGM, check if NOT absolute BPGM and make address tables RELATIVE
			if( mach==MACH_85 || *((WORD*)(InBuf+022))==0 ) {	// if HP85 BPGM OR (HP-87 BPGM and NOT absolute)
				WORD	oldbase=*((WORD*)(InBuf+030));

				i = 0;
				do {
					*((WORD*)(InBuf+030+i)) -= oldbase;
					i += 2;
				} while( *((WORD*)(InBuf+030+i))!=0177777 );
			}
		}
		if( mach==MACH_75 ) {	// HP75 LEX file, check if buffer is longer than length specified in LEX directory
			if( *((WORD*)(InBuf+2)) < InBufLen-18 ) InBufLen = *((WORD*)(InBuf+2))+18;
/*
!
! The HP-75 DIRECTORY HEADER is 18 (decimal) bytes long, 0x12.
! It's contents are as follows:
!
1200: EB 87 					ADDRESS (UNUSED)
1202: AD 04						FILE SIZE (04AD) (MINUS THIS 18-BYTE HEADER)
1204: 8D						ACCESS BITS (1000.1101)
1205: 4C						TYPE NAME ("L")
1206: 74 37 41 9C				DATE/TIME
120A: 4D 43 4F 50 59 20	20 20	LEX NAME ("MCOPY   ")
!
! All addresses in the LEX file are offsets relative
! to the following byte, the beginning of the LEXID.
!
1212: 01 50 			DEF LEXID (0x5001)
1214: 0C 00				DEF RUNTIM
1216: 18 00				DEF KEYWRDS
1218: 10 00 			DEF 1222
121A: 16 00 			DEF ERRMSG
121C: 20 00 			DEF INTCPT
121E: 4B 00 	RUNTIM	DEF SOMERTN
1220: 4D 00				DEF MCOPY.
1222: 4B 00		PARSE	DEF SOMERTN
1224: 21 00				DEF MCOPYPRS
1226: FF FF				BYT 377,377		! end of relocation marker
*/
			if( InBuf[5]!='L' ) {	// expect file type 'L' (LEX)
				printf("Error: doesn't look like a LEX file!\r\n");
				exit(6);
			}
			// throw away DIRECTORY HEADER so BaseAddress is 0
			memcpy(InBuf, InBuf+18, InBufLen-18);
			InBufLen -= 18;
			BaseAddress = 0; // last address BPGM was loaded at
			EndAddress = BaseAddress + InBufLen;
		} else {
			if( InBuf[6]!=2 ) {	// if not BPGM file type in Program Control Block (PCB)
				printf("Error: doesn't look like a binary program!\r\n");
				exit(6);
			}
			for(i=7, j=0; i<24; i++) j += (int)InBuf[i];	// rest of PCB 0s?
			if( mach==MACH_85 && j!=0 ) {
				printf("Error: doesn't look like an HP-85 binary program!\r\n");
				exit(6);
			}
			if( mach==MACH_87 && j==0 ) {
				printf("Error: doesn't look like an HP-87 binary program!\r\n");
				exit(6);
			}
			BaseAddress = *((WORD*)(InBuf+24)); // last address BPGM was loaded at
			if( BaseAddress==0 ) BaseAddress = *((WORD*)(InBuf+18));	// if no last loaded address, use ABS BASE ADDRESS
			EndAddress = BaseAddress + InBufLen;
		}
	}

	//======================================================================
	// LOAD and PROCESS GLOBALS FILE
	// open GLOBALS.TXT file (if it exists) for reading, allocate buffer, read the file in
	if( LocalGlobal>0 ) strcpy(GlobalFileName, globfile[mach]);
	else strcat(GlobalFileName, globfile[mach]);

	if( -1!=(fp=open(GlobalFileName, _O_BINARY | _O_RDONLY)) ) {
		LstBufLen = _lseek(fp, 0, SEEK_END);
		// we allocate 1 more than the size of the file so that there's room for
		// a NUL byte at the end in case the idiot user has edited the file and
		// removed the eol seqquence from the end of the last line
		LstBuf = (char *)malloc(LstBufLen+1);
		if( LstBuf == NULL ) {
			printf("Error allocating GLOBALS buffer\r\n");
			exit(8);
		}
		lseek(fp, 0, SEEK_SET);
		read(fp, (void*)LstBuf, LstBufLen);
		close(fp);

// process GLOBALS file, create GLOBAL entries for the DADs and EQUs
		ProcessGLOBALS();
		free(LstBuf);
		LstBuf = NULL;
	}
	LstBufLen = 0;

	//======================================================================
	// LOAD and PROCESS PRE-EXISTING .LST and .IMP FILES
	// modify input file name to create .LST output file name

	// first look in GLOBALSxx.TXT for the correct address values for the routines (85/87)
	addrROMJSB = addrERROR = addrERRORplus = addrSYSJSB = addrMELJSB = addrWCKJSB = 0xFFFF;
	for(i=0; i<GlobalsCnt; i++) {
		if( Globals[i].type==GLO_DAD ) {
			if( !stricmp(Globals[i].label, "SYSJSB") ) addrSYSJSB = Globals[i].address;
			if( !stricmp(Globals[i].label, "MELJSB") ) addrSYSJSB = Globals[i].address;
			if( !stricmp(Globals[i].label, "WCKJSB") ) addrSYSJSB = Globals[i].address;
			if( !stricmp(Globals[i].label, "ROMJSB") ) addrROMJSB = Globals[i].address;
			if( !stricmp(Globals[i].label, "ERROR") ) addrERROR = Globals[i].address;
			if( !stricmp(Globals[i].label, "ERROR+") ) addrERRORplus = Globals[i].address;
		}
	}
	// open LST file (if it exists) for reading, allocate buffer, read the file in
	if( -1==(fp=open(ListFileName, _O_BINARY | _O_RDONLY)) ) {	// .LST does NOT exist
		LstBufLen = 2*65536;	// 128K should be WILDLY enough space for preliminary DOT-COMMANDS
		LstBuf = (char *)malloc(LstBufLen);
		if( LstBuf == NULL ) {
			printf("Error allocating LST buffer\r\n");
			exit(7);
		}
		if( mach==MACH_75 && BaseAddress>=060000 ) {
			pL = LstBuf;

			sprintf(ScratBuf, ".def %o", (unsigned int)BaseAddress);
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".fgl %o RMHEAD", (unsigned int)BaseAddress);
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".eol %o ROM marker: 1CE3", (unsigned int)BaseAddress);
			AddLst(&pL, ScratBuf);

			sprintf(ScratBuf, ".def %o", (unsigned int)(BaseAddress+2));
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".eol %o RMDR.LOC - location of file", (unsigned int)(BaseAddress+2));
			AddLst(&pL, ScratBuf);

			sprintf(ScratBuf, ".byt %o 2", (unsigned int)(BaseAddress+4));
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".eol %o RMDR.SIZ - size of file", (unsigned int)(BaseAddress+4));
			AddLst(&pL, ScratBuf);

			sprintf(ScratBuf, ".val %o TYTOK?", (unsigned int)(BaseAddress+6));
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".eol %o RMDR.TYP - type of file", (unsigned int)(BaseAddress+6));
			AddLst(&pL, ScratBuf);

			sprintf(ScratBuf, ".val %o TYNSYS", (unsigned int)(BaseAddress+7));
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".eol %o RMDR.TNM - name of type", (unsigned int)(BaseAddress+7));
			AddLst(&pL, ScratBuf);

			sprintf(ScratBuf, ".byt %o 4", (unsigned int)(BaseAddress+8));
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".eol %o RMDR.DAT - Date of creation", (unsigned int)(BaseAddress+8));
			AddLst(&pL, ScratBuf);

			sprintf(ScratBuf, ".asc %o 10", (unsigned int)(BaseAddress+12));
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".eol %o RMDR.NAM - name of file", (unsigned int)(BaseAddress+12));
			AddLst(&pL, ScratBuf);

			sprintf(ScratBuf, ".def %o", (unsigned int)(BaseAddress+20));
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".eol %o RMDR.ESZ - size of file (0 for ROMs)", (unsigned int)(BaseAddress+20));
			AddLst(&pL, ScratBuf);

			sprintf(ScratBuf, ".def %o", (unsigned int)(BaseAddress+22));
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".eol %o RM.ID - ROM id#", (unsigned int)(BaseAddress+22));
			AddLst(&pL, ScratBuf);

			sprintf(ScratBuf, ".def %o", (unsigned int)(BaseAddress+24));
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".eol %o RM.RUN - pointer to RUNTIM table", (unsigned int)(BaseAddress+24));
			AddLst(&pL, ScratBuf);
//			dw = *((WORD*)(InBuf+24));
//			sprintf(ScratBuf, ".lab %o RUNTIM", (unsigned int)dw);
//			AddLst(&pL, ScratBuf);

			sprintf(ScratBuf, ".def %o", (unsigned int)(BaseAddress+26));
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".eol %o RM.ASC - pointer to ASCII TOKEN table", (unsigned int)(BaseAddress+26));
			AddLst(&pL, ScratBuf);
//			dw = *((WORD*)(InBuf+24));
//			sprintf(ScratBuf, ".lab %o RUNTIM", (unsigned int)dw);
//			AddLst(&pL, ScratBuf);

			sprintf(ScratBuf, ".def %o", (unsigned int)(BaseAddress+28));
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".eol %o RM.PAR - pointer to PARSE table", (unsigned int)(BaseAddress+28));
			AddLst(&pL, ScratBuf);

			sprintf(ScratBuf, ".def %o", (unsigned int)(BaseAddress+30));
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".eol %o RM.ERR - pointer to ERRMSG table", (unsigned int)(BaseAddress+30));
			AddLst(&pL, ScratBuf);

			sprintf(ScratBuf, ".def %o", (unsigned int)(BaseAddress+32));
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".eol %o RM.INI - pointer to INIT routine", (unsigned int)(BaseAddress+32));
			AddLst(&pL, ScratBuf);

			sprintf(ScratBuf, ".cod %o 0", (unsigned int)(BaseAddress+34));
			AddLst(&pL, ScratBuf);

			LstBufLen = pL-LstBuf;
		} else if( BaseAddress<=060000 ) {	// if ROM @ 60000 and no .LST file, do some automatic DOT-COMMANDS
			pL = LstBuf;
			if( mach==MACH_75 ) {

			} else {
				if( BaseAddress==060000 ) {	// handle ROM tables
					AddLst(&pL, ".byt 60000 2");
					AddLst(&pL, ".eol 60000 ROM# and complement");
					AddLst(&pL, ".def 60002");
					AddLst(&pL, ".eol 60002 ptr to table of runtime addresses");
					dw = *((WORD*)(InBuf+2));
					sprintf(ScratBuf, ".lab %o RUNTIM", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
					sprintf(ScratBuf, ".def %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
					AddLst(&pL, ".eol 60004 ptr to table of TOKENS (ASCII keywords)");
					dw = *((WORD*)(InBuf+4));
					sprintf(ScratBuf, ".lab %o TOKENS", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
					sprintf(ScratBuf, ".asp %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
					dw = *((WORD*)(InBuf+6));
					sprintf(ScratBuf, ".lab %o PARSE", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
					sprintf(ScratBuf, ".def %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
					AddLst(&pL, ".eol 60006 ptr to table of parse addresses");
					dw = *((WORD*)(InBuf+8));
					sprintf(ScratBuf, ".lab %o ERMSGS", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
					AddLst(&pL, ".eol 60010 ptr to table of warning and error messages");
					dw = *((WORD*)(InBuf+10));
					sprintf(ScratBuf, ".lab %o INIT", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
					sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
					AddLst(&pL, ".eol 60012 Address of ROM initialization routine");
				}

				// second look for known code/non-code sequences like calls to ROMJSB, ERROR, ERROR+
				for(i=0; i<InBufLen-3; i++) {
					if( InBuf[i]==0316 && (*((WORD*)(InBuf+i+1))==addrSYSJSB || *((WORD*)(InBuf+i+1))==addrMELJSB || *((WORD*)(InBuf+i+1))==addrWCKJSB) )  {	// if JSB =SYSJSB
						dw = BaseAddress+i;
						sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
						AddLst(&pL, ScratBuf);
						dw += 3;
						g = FLG_SYSJSB_DEF;	// if target ROM is 0, tell .def to use GLOBALS file
						sprintf(ScratBuf, ".def %o %o", (unsigned int)dw, g);
						AddLst(&pL, ScratBuf);
						dw += 2;
						sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
						AddLst(&pL, ScratBuf);
					}
					if( InBuf[i]==0316 && *((WORD*)(InBuf+i+1))==addrROMJSB ) {	// if JSB =ROMJSB
						dw = BaseAddress+i;
						sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
						AddLst(&pL, ScratBuf);
						dw += 3;
						g = (InBuf[i+5]==0)?FLG_ROMJSB_DEF:0;	// if target ROM is 0, tell .def to use GLOBALS file
						sprintf(ScratBuf, ".def %o %d", (unsigned int)dw, g);
						AddLst(&pL, ScratBuf);
						dw += 2;
						sprintf(ScratBuf, ".byt %o", (unsigned int)dw);
						AddLst(&pL, ScratBuf);
						dw += 1;
						sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
						AddLst(&pL, ScratBuf);
					}
					if( InBuf[i]==0316 && (*((WORD*)(InBuf+i+1))==addrERROR || *((WORD*)(InBuf+i+1))==addrERRORplus) ) {	// if JSB =ERROR or JSB =ERROR+
						dw = BaseAddress+i;
						sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
						AddLst(&pL, ScratBuf);
						dw += 3;
						sprintf(ScratBuf, ".byt %o", (unsigned int)dw);
						AddLst(&pL, ScratBuf);
						dw += 1;
						sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
						AddLst(&pL, ScratBuf);
					}
				}
				// third, handle all RUNTIME routine entry points and primary attribute in front of them
				// Keeping track of processed addresses, since sometimes the same routine will be used for
				// different tokens, and we don't want to duplicate DOT-COMMANDs.
				// Use first 1024 bytes of ScratBuf as bit-flags for 8K possible addresses.
				if( BaseAddress==060000 ) {

					memset(ScratBuf, 0, 1024);	// set address flags to "not done"
					rp = *((WORD*)(InBuf+2))+2-BaseAddress;	// point to RUNTIM table, skip token 0 which is never used
					tp = (*((WORD*)(InBuf+4)))-BaseAddress;	// point to TOKENS table
					pp = *((WORD*)(InBuf+6))+2;	// point to PARSE table, skip token 0 which is never used
					while( InBuf[tp] != 0377 ) {	// while not end of table
						sprintf(ScratBuf+1024, ".eol %o ", (unsigned int)(rp+BaseAddress));
						j = 1024+strlen(ScratBuf+1024);
						if( (InBuf[tp] & 0x80) && (InBuf[tp] & 0x7F)<' ' ) {	// 87 ASSEMBLER ROM, at least, has some '200' tokens that are bogus
							++tp;	// skip it
							ScratBuf[1024] = 0;	// set buffer to empty
							j = 1024;	// and buffer length
						} else {
							do ScratBuf[j++] = InBuf[tp] & 0x7F;	// copy TOKEN name to RUNTIM entry comment
							while( !(InBuf[tp++] & 0x080) );
							// reached end of this keyword
							ScratBuf[j] = 0;
							AddLst(&pL, ScratBuf+1024);
						}
						// now add DOT-COMMANDs for RUNTIME routine entry point
						dw = *((WORD*)(InBuf+rp));	// get RUNTIME routine entry point address
						if( dw>BaseAddress && dw<BaseAddress+InBufLen ) {
							rp += 2;	// move RUNTIME table ptr to next entry for next loop
							rep = dw-BaseAddress;
							k = 1;	// for primary attribute
							if( (InBuf[rep-1] & 0x0C0)==0 ) k += 1 + ((InBuf[rep-2]>>4)-2+3)/4;	// it's a FUNCTION, include number of SECONDARY ATTRIBUTES!
							else { // statement, add comment to parse table entry
								sprintf(ScratBuf+1024, ".eol %o", (unsigned int)pp);
								ScratBuf[1024+10] = ' ';	// got overwritten by the terminating NUL of the sprintf above
								AddLst(&pL, ScratBuf+1024);
							}
							pp += 2;
							if( !(ScratBuf[rep/8] & (1<<(rep&7))) ) {	// if not processed this one before
								sprintf(ScratBuf+1024, ".cod %o", (unsigned int)dw);	// the entry point has to be code
								AddLst(&pL, ScratBuf+1024);
								sprintf(ScratBuf+1024, ".byt %o %o", (unsigned int)dw-k, k);	// the byte before RUNTIM entry has to be BYT (keyword attribute)
								AddLst(&pL, ScratBuf+1024);
								ScratBuf[rep/8] |= (1<<(rep&7));	// set flag that we've processed this address
							}
						} else {
							rp += 2;
							pp += 2;
						}
					}
				}
				for(i=8192-5; InBuf[i]==0; i--) ;	// find how many empty bytes at end of ROM
				++i;
				j = 8192-5-i;
				if( j ) {
					sprintf(ScratBuf, ".bsz %o %o", i, j);
					AddLst(&pL, ScratBuf);	// make end of ROM a BSZ
				}
				AddLst(&pL, ".byt 77774 4");
				AddLst(&pL, ".eol 77774 ROM checksums");
				LstBufLen = pL-LstBuf;
			}
		} else if( IsROM==8 ) {	// BPGM
			pL = LstBuf;

			if( mach==MACH_75 ) {
				sprintf(ScratBuf, " :LEXID   EQU %o", *((WORD*)(InBuf+0)));
				AddLst(&pL, ScratBuf);
			}
			if( mach==MACH_75 ) {
				TableBase = 000;
//				AddRec(REC_EQU, FLG_FEQ, "LEXID", *((WORD*)(InBuf+18)), NULL, *((WORD*)(InBuf+18)));
				sprintf(ScratBuf, ".def 0");
				AddLst(&pL, ScratBuf);
				sprintf(ScratBuf, ".feq 0 LEXID");
				AddLst(&pL, ScratBuf);
			} else TableBase = 030;

			sprintf(ScratBuf, ".def %o", TableBase+2);
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".eol %o ptr to table of runtime addresses", TableBase+2);
			AddLst(&pL, ScratBuf);
			dw = *((WORD*)(InBuf+TableBase+02));
			sprintf(ScratBuf, ".lab %o RUNTIM", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".def %o", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".eol %o ptr to table of TOKENS (ASCII keywords)", TableBase+4);
			AddLst(&pL, ScratBuf);
			dw = *((WORD*)(InBuf+TableBase+04));
			sprintf(ScratBuf, ".lab %o TOKENS", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".asp %o", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			dw = *((WORD*)(InBuf+TableBase+06));
			sprintf(ScratBuf, ".lab %o PARSE", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".def %o", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".eol %o ptr to table of parse addresses", TableBase+6);
			AddLst(&pL, ScratBuf);
			dw = *((WORD*)(InBuf+TableBase+010));
			sprintf(ScratBuf, ".lab %o ERMSGS", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".eol %o ptr to table of warning and error messages", TableBase+010);
			AddLst(&pL, ScratBuf);
			dw = *((WORD*)(InBuf+TableBase+012));
			if( mach==MACH_75 ) {
				sprintf(ScratBuf, ".lab %o INTCPT", (unsigned int)dw);
				AddLst(&pL, ScratBuf);
				sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
				AddLst(&pL, ScratBuf);
				sprintf(ScratBuf, ".eol %o Address of LEX intercept routine", TableBase+012);
				AddLst(&pL, ScratBuf);
			} else {
				sprintf(ScratBuf, ".lab %o INIT", (unsigned int)dw);
				AddLst(&pL, ScratBuf);
				sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
				AddLst(&pL, ScratBuf);
				sprintf(ScratBuf, ".eol %o Address of ROM initialization routine", TableBase+012);
				AddLst(&pL, ScratBuf);
			}
			// second look for known code/non-code sequences like calls to ROMJSB, ERROR, ERROR+
			for(i=0; i<InBufLen-3; i++) {
				if( InBuf[i]==0316 && (*((WORD*)(InBuf+i+1))==addrSYSJSB || *((WORD*)(InBuf+i+1))==addrMELJSB || *((WORD*)(InBuf+i+1))==addrWCKJSB) )  {	// if JSB =SYSJSB
					dw = BaseAddress+i;
					sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
					dw += 3;
					g = FLG_SYSJSB_DEF;	// if target ROM is 0, tell .def to use GLOBALS file
					sprintf(ScratBuf, ".def %o %o", (unsigned int)dw, g);
					AddLst(&pL, ScratBuf);
					dw += 2;
					sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
				}
				if( InBuf[i]==0316 && *((WORD*)(InBuf+i+1))==addrROMJSB ) {	// if JSB =ROMJSB
					dw = BaseAddress+i;
					sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
					dw += 3;
					g = (InBuf[i+5]==0)?FLG_ROMJSB_DEF:0;	// if target ROM is 0, tell .def to use GLOBALS file
					sprintf(ScratBuf, ".def %o %d", (unsigned int)dw, g);
					AddLst(&pL, ScratBuf);
					dw += 2;
					sprintf(ScratBuf, ".byt %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
					dw += 1;
					sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
				}
				if( InBuf[i]==0316 && (*((WORD*)(InBuf+i+1))==addrERROR || *((WORD*)(InBuf+i+1))==addrERRORplus) ) {	// if JSB =ERROR or JSB =ERROR+
					dw = BaseAddress+i;
					sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
					dw += 3;
					sprintf(ScratBuf, ".byt %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
					sprintf(ScratBuf, ".dec %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
					dw += 1;
					sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
				}
			}
			// third, handle all RUNTIME routine entry points and primary attribute in front of them
			// Keeping track of processed addresses, since sometimes the same routine will be used for
			// different tokens, and we don't want to duplicate DOT-COMMANDs.
			// Use first 4096 bytes of ScratBuf as bit-flags for 32K possible addresses.

			memset(ScratBuf, 0, 4096);	// set address flags to "not done"
			rp = *((WORD*)(InBuf+TableBase+02))+2-BaseAddress;	// point to RUNTIM table, skip token 0 which is never used
			tp = (*((WORD*)(InBuf+TableBase+04)))-BaseAddress;	// point to TOKENS table
			pp = *((WORD*)(InBuf+TableBase+06))+2;	// point to PARSE table, skip token 0 which is never used
			while( InBuf[tp] != 0377 ) {	// while not end of table
				sprintf(ScratBuf+4096, ".eol %o ", (unsigned int)(rp+BaseAddress));
				j = 4096+strlen(ScratBuf+4096);
				if( (InBuf[tp] & 0x80) && (InBuf[tp] & 0x7F)<' ' ) {	// 87 ASSEMBLER ROM, at least, has some '200' tokens that are bogus
					++tp;	// skip it
					ScratBuf[4096] = 0;	// set buffer to empty
					j = 4096;	// and buffer length
				} else {
					do ScratBuf[j++] = InBuf[tp] & 0x7F;	// copy TOKEN name to RUNTIM entry comment
					while( !(InBuf[tp++] & 0x080) );
					// reached end of this keyword
					ScratBuf[j] = 0;
					AddLst(&pL, ScratBuf+4096);
				}
				// now add DOT-COMMANDs for RUNTIME routine entry point
				dw = *((WORD*)(InBuf+rp));	// get RUNTIME routine entry point address
				if( dw>BaseAddress && dw<BaseAddress+InBufLen ) {
					rp += 2;	// move RUNTIME table ptr to next entry for next loop
					rep = dw-BaseAddress;
					k = 1;	// for primary attribute
					if( (InBuf[rep-1] & 0x0C0)==0 ) k += 1 + ((InBuf[rep-2]>>4)-2+3)/4;	// it's a FUNCTION, include number of SECONDARY ATTRIBUTES!
					else { // statement, add comment to parse table entry
						sprintf(ScratBuf+4096, ".eol %o", (unsigned int)pp);
						ScratBuf[4096+10] = ' ';	// got overwritten by the terminating NUL of the sprintf above
						AddLst(&pL, ScratBuf+4096);
					}
					pp += 2;
					if( !(ScratBuf[rep/8] & (1<<(rep&7))) ) {	// if not processed this one before
						sprintf(ScratBuf+4096, ".cod %o", (unsigned int)dw);	// the entry point has to be code
						AddLst(&pL, ScratBuf+4096);
						sprintf(ScratBuf+4096, ".byt %o %o", (unsigned int)dw-k, k);	// the byte before RUNTIM entry has to be BYT (keyword attribute)
						AddLst(&pL, ScratBuf+4096);
						ScratBuf[rep/8] |= (1<<(rep&7));	// set flag that we've processed this address
					}
				} else {
					rp += 2;
					pp += 2;
				}
			}
			LstBufLen = pL-LstBuf;
		} else {
			LstBufLen = 0;
		}
	} else {	// .LST already exists
		LstBufLen = _lseek(fp, 0, SEEK_END);
		// we allocate 1 more than the size of the file so that there's room for
		// a NUL byte at the end in case the idiot user has edited the file and
		// removed the eol seqquence from the end of the last line
		LstBuf = (char *)malloc(LstBufLen+1);
		if( LstBuf == NULL ) {
			printf("Error allocating LST buffer\r\n");
			exit(7);
		}
		lseek(fp, 0, SEEK_SET);
		read(fp, (void*)LstBuf, LstBufLen);
		close(fp);
		if( BackupMax ) {	// if backing up
			sprintf(ScratBuf, "%s.bak%d", ListFileName, (int)BackupNum);
			if( -1==(OutFile=open(ScratBuf, _O_BINARY | _O_RDWR | _O_CREAT | _O_TRUNC, _S_IREAD | _S_IWRITE)) ) {
				printf("Error creating .BAK file %s", ScratBuf);
			} else {
				write(OutFile, LstBuf, LstBufLen);
				close(OutFile);
				if( ++BackupNum >= BackupMax ) BackupNum = 0;
			}
		}
	}
// process .LST file, create DISASMRECs for comments, labels, data types
	ProcessLST(FALSE);
	free(LstBuf);
	LstBuf = NULL;
	LstBufLen = 0;

	if( ImpStart<ImpEnd ) {
		// Now check for an IMPORT (.imp) filename
		for(i=strlen(ImportFileName)-1; i && ImportFileName[i]!='.'; i--) ;
		if( i>0 && ImportFileName[i]=='.' ) strcpy(ImportFileName+i, ".imp");
		else strcat(ImportFileName, ".imp");

		// open IMP file (if it exists) for reading, allocate buffer, read the file in
		if( -1!=(fp=open(ImportFileName, _O_BINARY | _O_RDONLY)) ) {	// .LST does NOT exist
			LstBufLen = _lseek(fp, 0, SEEK_END);
			// we allocate 1 more than the size of the file so that there's room for
			// a NUL byte at the end in case the idiot user has edited the file and
			// removed the eol seqquence from the end of the last line
			LstBuf = (char *)malloc(LstBufLen+1);
			if( LstBuf == NULL ) {
				printf("Error allocating IMP buffer\r\n");
				exit(7);
			}
			lseek(fp, 0, SEEK_SET);
			read(fp, (void*)LstBuf, LstBufLen);
			close(fp);
	// process .LST file, create DISASMRECs for comments, labels, data types
			ProcessLST(TRUE);
			free(LstBuf);
			LstBuf = NULL;
			LstBufLen = 0;
		}
	}

	//======================================================================
	// CREATE TEMP OUTPUT FILE and WRITE INITIAL LINES
	// create/truncate temp .LST file
	if( -1==(OutFile=open(TmpFileName, _O_BINARY | _O_RDWR | _O_CREAT | _O_TRUNC, _S_IREAD | _S_IWRITE)) ) {
		printf("Error creating temp .LST file %s", TmpFileName);
		exit(9);
	}
	sprintf(ScratBuf, ".opt -%s%c -%c%c", (mach==MACH_75)?"75":((mach==MACH_85)?"85":"87"), HexOut?'x':' ', IsROM==8?'b':'r', (int)(IsROM==8?' ':('0'+IsROM)));
	if( ForceComment>0 ) strcpy(ScratBuf+strlen(ScratBuf), " -c");
	else if( ForceComment==0 ) strcpy(ScratBuf+strlen(ScratBuf), " -C");
	if( DoWarnings==0 ) strcpy(ScratBuf+strlen(ScratBuf), " -q");
	if( DoWarnings>0 ) strcpy(ScratBuf+strlen(ScratBuf), " -Q");
	if( LocalGlobal>0 ) strcpy(ScratBuf+strlen(ScratBuf), " -g");
	if( LocalGlobal==0 ) strcpy(ScratBuf+strlen(ScratBuf), " -G");
	if( BackupNum>=0 ) sprintf(ScratBuf+strlen(ScratBuf), " -B%d:%d", (int)BackupMax, (int)BackupNum);
	WriteLine(ScratBuf);
	sprintf(ScratBuf, ".dis %s %s disassembly - %s", TitleRevStr, BinFileName, TimeString());
	WriteLine(ScratBuf);
	WriteLine(".dis =========================================================================");
	WriteLine(".dis The HP80DIS program assumes that:");
	WriteLine(".dis   1) WhiteSpace at the start of lines is discardable.");
	WriteLine(".dis   2) If the first character on a line is a '.' the line is a DOT COMMAND.");
	WriteLine(".dis   3) If the first character is a ':' is a DAD or EQU.");
	WriteLine(".dis   4) Any line starting with a digit is a line of object/source code.");
	WriteLine(".dis   3) All other lines are full-line comments.  If the first non-WhiteSpace");
	WriteLine(".dis      character is not a '!' then a '!' is added.  Full-line comments are");
	WriteLine(".dis      automatically indented to the normal 'label' column.");
	WriteLine(".dis   4) For HP-85, code LABELS are 1 to 6 characters long and begin in");
	WriteLine(".dis      column 25 (the column immediately after the ':' column).  For HP-87,");
	WriteLine(".dis      code LABELS are 1 to 8 characters long.");
	WriteLine(".dis   5) Actual code starts after WhiteSpace after any LABEL (or non-label).");
	WriteLine(".dis   6) End-of-line comments start after the first '!' character that has a");
	WriteLine(".dis       WhiteSpace character BEFORE it, AFTER the source code.");
	WriteLine(".dis   7) All DADs and EQUs are moved to the start of the file, and there can");
	WriteLine(".dis      be NO full-line comments mixed in with them.  Sorry about that.");
	WriteLine(".dis DOT ('.') COMMANDS can be placed anywhere, each on its own line.");
	WriteLine(".dis ");
	WriteLine(".dis NOTE: All .dis lines are discarded, although this 'help' block is");
	WriteLine(".dis   reinserted each time (so deleting it won't work).  All other");
	WriteLine(".dis   DOT-COMMANDS are moved to the end of the file during processing and are");
	WriteLine(".dis   sorted in increasing address order.");
	WriteLine(".dis =========================================================================");
	WriteLine(".dis");	// to separate above from list of DADs and EQUs

	// Dump all the DADs up front in the same order in which they were originally entered
	j = 0;
	do {
		for(pRec=pDis[0]; pRec!=NULL; pRec = pRec->pNext) {
			if( pRec->type==REC_DAD && (pRec->address & 0x00003FFF)==j ) {
				strcpy(ScratBuf, "                       :");
				strcpy(label, (char*)pRec->label);
				for(i=0; label[i]; i++) ;
				for(; i<8; i++) label[i] = ' ';
				label[i] = 0;
				if( OnlySrc ) sprintf(ScratBuf+strlen(ScratBuf), "%s DAD %o", label, (unsigned int)HP85ADDR(pRec->address));	// option ROM DAD (don't show ROM when outputing SRC)
				else sprintf(ScratBuf+strlen(ScratBuf), "%s DAD %o:%o", label, (unsigned int)HP85ADDR(pRec->address), pRec->datover);	// option ROM DAD
				if( pRec->Comment ) {
					k = strlen(ScratBuf);
					for(bp=ScratBuf+k; k<55; k++) *bp++ = ' ';
					if( *(bp-1)!=' ' ) *bp++ = ' ';
					*bp++ = '!';
					*bp++ = ' ';
					strcpy(bp, pRec->Comment);
				}
				WriteLine(ScratBuf);
				++j;
				break;
			}
		}
	} while( pRec!=NULL );
	WriteLine(".dis");
	// Dump all the EQUs up front
	j = 0;
	do {
		for(pRec=pDis[0]; pRec!=NULL; pRec = pRec->pNext) {
			if( pRec->type==REC_EQU && (pRec->address & 0x00003FFF)==j ) {
				strcpy(ScratBuf, "                       :");
				strcpy(label, (char*)pRec->label);
				for(i=0; label[i]; i++) ;
				for(; i<8; i++) label[i] = ' ';
				label[i] = 0;
				sprintf(ScratBuf+strlen(ScratBuf), "%s EQU %o", label, (unsigned int)HP85ADDR(pRec->address));
				if( pRec->Comment ) {
					k = strlen(ScratBuf);
					for(bp=ScratBuf+k; k<55; k++) *bp++ = ' ';
					if( *(bp-1)!=' ' ) *bp++ = ' ';
					*bp++ = '!';
					*bp++ = ' ';
					strcpy(bp, pRec->Comment);
				}
				WriteLine(ScratBuf);
				++j;
				break;
			}
		}
	} while( pRec!=NULL );
	WriteLine(".dis");

	if( IsROM<8 ) {
		sprintf(ScratBuf, ".dis                             ABS %o", (unsigned int)BaseAddress);
		WriteLine(ScratBuf);
		NxtByt = (0 + BaseAddress)<<16;
	} else if( mach==MACH_85 ) {
		for(i=0; i<6; i++) ScratBuf[1024+i] = InBuf[i];
		while( i && ScratBuf[1024+i-1]==' ' ) --i;
		ScratBuf[1024+i+1] = 0;
		sprintf(ScratBuf, ".dis                             NAM %s", (char*)ScratBuf+1024);
		WriteLine(ScratBuf);
		NxtByt = (26 + BaseAddress)<<16;
	} else if( mach==MACH_87 ) {
		for(i=0; i<4; i++) ScratBuf[1024+i] = InBuf[i];
		while( i && ScratBuf[1024+i-1]==' ' ) --i;
		ScratBuf[1024+i+1] = 0;
		j = (int)(WORD)InBuf[7];
		sprintf(ScratBuf, ".dis                             NAM %o,%s", j, (char*)ScratBuf+1024);
		WriteLine(ScratBuf);
		NxtByt = (26 + BaseAddress)<<16;
	} else if( mach==MACH_75 ) {

	}

	//======================================================================
	// DISASSEMBLE INPUT FILE TO TEMP FILE
	NextRec = pDis[0];
	do {

//if( HP85ADDR(NxtByt)==060720 ) {
//	dw =1;
//}

		DisAsm();
		WriteLine(ScratBuf);
	} while( HP85ADDR(NxtByt)-BaseAddress < InBufLen );

	if( OnlySrc ) WriteLine("                       :         FIN");
	else		  WriteLine(".dis                   :         FIN");

	// dump DOT-COMMANDS
	lastFullLineAddr = -1;
	for(pRec=pDis[0]; pRec!=NULL; pRec = pRec->pNext) {
		switch( pRec->type ) {
		 case REC_DRP:
		 	HOaddr(haddr, (unsigned int)(pRec->address>>16));
		 	sprintf(ScratBuf, ".drp %s", haddr);
			WriteLine(ScratBuf);
		 	break;
		 case REC_BYT:
		 	HOaddr(haddr, (unsigned int)(pRec->address>>16));
		 	sprintf(ScratBuf, ".byt %s %o", haddr, pRec->datover);
			WriteLine(ScratBuf);
		 	break;
		 case REC_VAL:
		 	HOaddr(haddr, (unsigned int)(pRec->address>>16));
		 	sprintf(ScratBuf, ".val %s %s", haddr, pRec->reflabel);
			WriteLine(ScratBuf);
		 	break;
		 case REC_DEF:
		 	HOaddr(haddr, (unsigned int)(pRec->address>>16));
		 	sprintf(ScratBuf, ".def %s %d", haddr, (pRec->flags & FLG_ROMJSB_DEF));
			WriteLine(ScratBuf);
		 	break;
		 case REC_COD:
		 	HOaddr(haddr, (unsigned int)(pRec->address>>16));
		 	sprintf(ScratBuf, ".cod %s %o", haddr, pRec->datover);
			WriteLine(ScratBuf);
		 	break;
		 case REC_ASC:
		 	HOaddr(haddr, (unsigned int)(pRec->address>>16));
		 	sprintf(ScratBuf, ".asc %s %o", haddr, pRec->datover);
			WriteLine(ScratBuf);
		 	break;
		 case REC_ASP:
		 	HOaddr(haddr, (unsigned int)(pRec->address>>16));
		 	sprintf(ScratBuf, ".asp %s", haddr);
			WriteLine(ScratBuf);
		 	break;
		 case REC_BSZ:
		 	HOaddr(haddr, (unsigned int)(pRec->address>>16));
		 	sprintf(ScratBuf, ".bsz %s %o", haddr, pRec->datover);
			WriteLine(ScratBuf);
		 	break;
		 default:
			break;
		}
		if( pRec->type!=REC_DAD && pRec->type!=REC_EQU ) {
			if( pRec->label[0] ) {
				HOaddr(haddr, (unsigned int)(pRec->address>>16));
				sprintf(ScratBuf, ".lab %s %s", haddr, pRec->label);
				WriteLine(ScratBuf);
			}
			if( pRec->Comment!=NULL ) {
				if( pRec->type==REC_REM ) {	// full-line comment
					if( (pRec->address>>16) != lastFullLineAddr ) {
						lastFullLineAddr = (pRec->address>>16);
						fullLineCnt = 0;
					}
					HOaddr(haddr, (unsigned int)(pRec->address>>16));
					sprintf(ScratBuf, ".com %s %o %s", haddr, fullLineCnt, pRec->Comment);
					WriteLine(ScratBuf);
					++fullLineCnt;
				} else {	// eol comment
					HOaddr(haddr, (unsigned int)(pRec->address>>16));
					sprintf(ScratBuf, ".eol %s %s", haddr, pRec->Comment);
					WriteLine(ScratBuf);
				}
			}
		}
/* Added 3.19 to implement ".fco address label" to force an immediate multi-byte load on an unknown drp to be a 2-byte LABEL */
		if( pRec->flags & FLG_LABEL_ADDR ) {
		 	HOaddr(haddr, (unsigned int)HP85ADDR(pRec->address));
			sprintf(ScratBuf, ".fco %s %s", haddr, pRec->reflabel);
			WriteLine(ScratBuf);
		}
		if( pRec->flags & FLG_FEQ ) {
		 	HOaddr(haddr, (unsigned int)HP85ADDR(pRec->address));
			sprintf(ScratBuf, ".feq %s %s", haddr, pRec->reflabel);
			WriteLine(ScratBuf);
		}
		if( pRec->flags & FLG_FDA ) {
		 	HOaddr(haddr, (unsigned int)HP85ADDR(pRec->address));
			sprintf(ScratBuf, ".fda %s %s", haddr, pRec->reflabel);
			WriteLine(ScratBuf);
		}
		if( pRec->flags & FLG_GLO ) {
		 	HOaddr(haddr, (unsigned int)HP85ADDR(pRec->address));
			sprintf(ScratBuf, ".fgl %s %s", haddr, pRec->reflabel);
			WriteLine(ScratBuf);
		}
		if( pRec->flags & FLG_GTO ) {
		 	HOaddr(haddr, (unsigned int)HP85ADDR(pRec->address));
			sprintf(ScratBuf, ".gto %s", haddr);
			WriteLine(ScratBuf);
		}
		if( pRec->flags & FLG_DEC ) {
		 	HOaddr(haddr, (unsigned int)HP85ADDR(pRec->address));
			sprintf(ScratBuf, ".dec %s", haddr);
			WriteLine(ScratBuf);
		}
		if( pRec->flags & FLG_BCD ) {
		 	HOaddr(haddr, (unsigned int)HP85ADDR(pRec->address));
			sprintf(ScratBuf, ".bcd %s", haddr);
			WriteLine(ScratBuf);
		}
	}
	close(OutFile);

	//======================================================================
	// DELETE OLD .LST FILE and RENAME TEMP FILE TO .LST FILE
	if( !BailOut ) {
		if( OnlySrc ) {
			for(i=strlen(ListFileName)-1; i && ListFileName[i]!='.'; i--) ;
			if( i>0 && ListFileName[i]=='.' ) strcpy(ListFileName+i, ".src");
			else strcat(ListFileName, ".src");
		}
		unlink(ListFileName);
		rename(TmpFileName, ListFileName);
	}
	/*
	// For debugging, dump the DISASMRECs list
	if( -1!=(OutFile=open("DEBUG.LST", _O_BINARY | _O_RDWR | _O_CREAT | _O_TRUNC, _S_IREAD | _S_IWRITE)) ) {
		DISASMREC	*psrch;
		char	bbb[128];

		psrch = pDis[0];
		for(; psrch; psrch = psrch->pNext) {
			if(  psrch->label[0] ) {
				sprintf(bbb, "%8.8s %8.8lX\r\n", psrch->label, psrch->address);
				write(OutFile, bbb, strlen(bbb));
			}
		}
		close(OutFile);
	}
	*/
	exit(0);
}

//***************
//* END OF FILE *
//***************
