#include "main.h"
#include "mach85.h"
#include "script.h"

#include <stdio.h>
#include <stdlib.h>
#include <mmsystem.h>
#include <commdlg.h>
#include <fcntl.h>
#include <errno.h>
#include <stdint.h>
#include <winbase.h>
#include <string.h>
#include <malloc.h>
#include <ctype.h>
#include <time.h>
#include <io.h>
#include <sys/stat.h>

#define	NUMAUXFILES	11	// 1-10 for user, 11 for SDSAVE/SDGET/SDEXPORT/SDIMPORT

#define	SLASH		'\\'
#define	SLASHSTR	"\\"
// for EBTKS:
//#define SLASH		'/'
//#define SLASHSTR	"/"
#define	ISSLASH(c)	((c)=='/' || (c)=='\\')

extern int		numROMS;

double cvt_HP85_real_to_IEEE_double(uint8_t number[]);
int32_t cvt_R12_int_to_int32(uint8_t number[]);
void ExecuteMenu(long	item);
/*
NOTE: This emulator loads each plug-in ROM into its own 8K of malloc()'d RAM,
then stores the pointer to that 8K in the pROMS[] array. The Emulator has a
byte-array ROM[65536] that contains all of the ROM and RAM that the Capricorn
CPU can access. To keep things simple for the 'runtime' code, the pROMS[] blocks
are only accessed when 'selecting' a new ROM into the 060000-077777 address range,
at which point the appropriate pROMS[] 8K block is COPIED into ROM[], but when
another ROM is selected, that ROM[] block is NOT copied back into the pROMS[] block.
So, for AUXROM's ROM/RAM window emulation, when something is WRITTEN to that area,
we change BOTH the ROM[] location and the appropriate location in ALL of the
loaded AUXROMs (because those values have to remain the same no matter which AUXROM
is mapped in. A little hokey, but I wanted to minimize the changes to the Emulator
source files as much as possible, so here we are.
*/


#if NOAUXROM
#else

#define	U_WROM		1
#define	U_SDCD		2
#define	U_SDCUR		3
#define	U_SDCAT		4
#define	U_SDFLUSH	5
#define	U_SDOPEN	6
#define	U_SDREAD	7
#define	U_SDCLOS	8
#define	U_SDWRIT	9
#define	U_SDSEEK	10
#define	U_SDDEL		11
#define	U_MKDIR		12
#define	U_RMDIR		13
#define	U_SPF		14
#define	U_MOUNT		15
#define	U_UNMOUNT	16
#define	U_FLAGS		17
#define	U_SDREN		18
#define	U_CLOCK		19
#define	U_HELP		20
#define	U_MEDIA		21
#define	U_MEMCPY	22
#define	U_SETLED	23
#define	U_SDCOPY	24
#define	U_SDEOF		25
#define	U_EXISTS	26
#define	U_EBTKSREV	27
#define	U_RMIDLE	28
#define	U_BATCH		29
#define	U_BOOT		30
#define	U_KEY		31

#define	B_MB0	070000
#define	B_MB6	070006

#define	B_LEN0	070040
#define	B_LEN6	070054

#define	B_USE0	070060
#define B_USE6	070074

#define	A_MBR12 070206
#define	A_FLAGS 070234

#define	B_OPT00	070360
#define	B_OPT01	070361
#define	B_OPT02	070362
#define	B_OPT03 070363
#define	B_OPT04	070364
#define	B_OPT05	070365
#define	B_OPT06	070366
#define	B_OPT07 070367
#define	B_OPT08	070370
#define	B_OPT09	070371
#define	B_OPT10 070372
#define	B_OPT11 070373
#define	B_OPT12	070374
#define	B_OPT13	070375
#define	B_OPT14	070376
#define	B_OPT15	070377

#define	B_BUF0	070400
#define	B_BUF6	073400

extern long	ARnum[14];
extern BYTE	*pROMS[MAXROMS];

//BYTE RMIDLEtest[]="DISP \"Hello, world!\"\\232";
BYTE *pRMIDLEkeys=NULL;//RMIDLEtest;

/* GEEKDO: TEST CODE, REMOVE THIS
int TestKeyCnt=5;
*/

BYTE *Waldo=(BYTE*)"Waldo!!!";
BYTE WaldoNum[8]={0,0,0,0,0377,0x45,0x23,0x01};

/// NOTE: SDcur is initialized at startup in AuxINIT(), called from LoadROMS() in mach85.c
char SDcur[512], SDhome[512], TMPpath[1024];	// TMPpath is for building new absolute path, to make sure it's legal and is <512 bytes once 'normalized'
char Pattern[256];

long ffCAT=-1, ffUTIL=-1;	// start with no findfirst/findnext in progress
struct _finddata_t	CATdat, UTILdat;	// for SDCAT

/// for SDOPEN and related functions
struct {
	int	handle;
} SDfiles[NUMAUXFILES];

DWORD	saveFlags;

/// we cheat here and break isolation by reading/writing ioINTERNAL.c 'internal' variables
extern WORD	IO_CRTSAD, IO_CRTBAD;
extern BYTE IO_CRTCTL;
extern BYTE	*CRT;

BOOL	ScreenSaved=FALSE;
char	SaveScreen[16384];
BYTE	SaveCRTCTL;
WORD	SaveCRTSAD;
WORD	SaveCRTBAD;

///****************************************
/// Write INI for AUX ROMs
void AUXwriteINI() {

	saveFlags = *((DWORD*)&pROMS[ARnum[0]][A_FLAGS-060000]);
	sprintf((char*)ScratBuf, "0x%8.8lX", saveFlags);
	iniSetVar(iniSTR[INIS_AUXFLAGS], ScratBuf);
}

///****************************************
/// Read INI for AUX ROMs
void AUXreadINI() {

	if( iniFindVar(iniSTR[INIS_AUXFLAGS]) ) {
		iniGetNum(&saveFlags);
		saveFlags &= ~256;	// clear "AUX ROMs OK" flag, don't want it preset!
	} else {
		saveFlags =0;
	}
}

/// NOTE: WRITING to the ROM/RAM window is more complex,
/// because we have to write to ROM[] and ALL of the loaded pROMS[] that are AUXROMs.
/// So, we have functions for WRITING to ROM/RAM,
/// but to read, we just do:
///		ROM[addr]
///		*((WORD*)&ROM[addr])
///		*((DWORD*)&ROM[addr])
/// so there are no corresponding RRAMx() functions.

///**********************************************
/// Writes a BYTE to the ROM/RAM window
/// 'addr' is between 070000 and 075777, and address in the 8K AUXROM ROM/RAM window, so pROMS[ARnum[n]][addr-060000] works
/// To access ROM[], we just do ROM[addr]
void WRAMB(int addr, BYTE val) {

	int		i;

	ROM[addr] = val;
	for(i=0; i<14; i++) if( ARnum[i] ) *((BYTE*)&pROMS[ARnum[i]][addr-060000]) = val;
}

///**********************************************
/// Writes a WORD to the ROM/RAM window
/// 'addr' is between 070000 and 075777
void WRAMW(int addr, WORD val) {

	int		i;

	*((WORD*)&ROM[addr]) = val;
	for(i=0; i<14; i++) if( ARnum[i] ) *((WORD*)&pROMS[ARnum[i]][addr-060000]) = val;
}

///**********************************************
/// Writes a DWORD to the ROM/RAM window
/// 'addr' is between 070000 and 075777
void WRAMD(int addr, DWORD val) {

	int		i;

	*((DWORD*)&ROM[addr]) = val;
	for(i=0; i<14; i++) if( ARnum[i] ) *((DWORD*)&pROMS[ARnum[i]][addr-060000]) = val;
}

///********************************
/// no checking for pSRC==NULL. We trust the caller.
void WRAMBUF(int addr, BYTE *pSRC, int len) {

	int		j;

	for(j=0; j<len; j++) {
		WRAMB(addr++, *pSRC++);
	}
}

///*************************************************************************************
/// AuxINIT is called from LoadROMS() in mach85.c, after all of the ROMs have been loaded
void AuxINIT() {

	int	i, j;

	memset((BYTE*)&ARnum[0], 0, sizeof(ARnum));
	for(i=0; i<numROMS; i++) if( pROMS[i][0]==0361 ) break;
	if( i<numROMS ) {	// if AUX ROM 1 is here, log all AUX ROMs in
		for(j=0; j<numROMS; j++) {
			if( pROMS[j][0]==0361 ) ARnum[0] = j;	// set the AR# -> pROMS# map
			else if( pROMS[j][0]==0377 && pROMS[j][1]>0360 && pROMS[j][1]<0376 ) ARnum[pROMS[j][1]-0361] = j;	// set the AR# -> pROMS# map
/*			if( pROMS[j][0]>0360 ) {
				ARnum[pROMS[j][1]-0361] = j;	// set the AR# -> pROMS# map
//				if( pROMS[j][0]!=0361 ) pROMS[j][0] = 0377;	// hide all AUX ROMs except the main one
			}
*/
		}
		// set up default SDcur
		strcpy(SDcur, PgmPath);	// default SDCUR$ is the emulator's location
		for(i=0; SDcur[i] && SDcur[i]!='\\'; i++) ;	// skip leading "C:"
		strcpy(SDcur, SDcur+i);
		if( SDcur[0]!='\\' ) strcpy(SDcur, "\\");
		// now go through and convert all '\\' to '/'
		for(i=0; SDcur[i]; i++) if( SDcur[i]=='\\' ) SDcur[i] = '/';
		strcpy(SDhome, SDcur);
		// set all SDfiles to closed
		for(i=0; i<NUMAUXFILES; i++) SDfiles[i].handle = -1;
		// write the A.FLAGS
		WRAMD(A_FLAGS, saveFlags);
	}
}

///**********************************************************
/// Checks pT against possbily wild-card-containing pP.
/// Returns TRUE if there's a match, otherwise FALSE.
/// pT must NOT have any wildcards, just be a valid filename.
/// pP may or may not have wildcards (* and ?).
/// * matches 0 or more "any characters"
/// ? matches exactly one "any character"
BOOL MatchesPattern(char *pT, char *pP) {

	if( *pT==0 ) {	// if reached the end of the test string
		return (*pP==0 || (*pP=='*' && *(pP+1)==0));	// return TRUE if reached the end of the pattern string, else FALSE
	} else if( *pP=='?' ) {	// pattern matches ANYTHING in the test
		return MatchesPattern(pT+1, pP+1);
	} else if( *pP=='*' ) {
		// start with * each 0 characters, then keep advancing pT so * eats more and more characters
		while( *pT ) {
			if( MatchesPattern(pT, pP+1) ) return TRUE;
			++pT;
		}
		return ( *(pP+1)==0 );	// return TRUE if reached the end of the pattern string, else FALSE
	} else if( toupper(*pT)==toupper(*pP) ) return MatchesPattern(pT+1, pP+1);
	return FALSE;
}

///**********************************************************
/// This function is provided in case EBTKS's (or a port of this emulator to a system different than MS Windows)
/// It's job is to translate the native system's file attributes into MS Windows attribute bits, which is what
/// AUXROM is expecting. Here's it's a do-nothing.
DWORD PatchAttribs(DWORD a) {

	a = (a & ~2) | ((a<<1)&2);	// move READ-ONLY bit to BIT1
	a = (a & ~1) | ((a>>4)&1);	// move SUB-DIR bit to BIT0
	return (a & 3);
}
///**********************************************************
void RetCat(int buf, int blen, int buse) {

	char		*pChar;
	_fsize_t	fsize;
    struct tm	*newtime;
/*
ROM[buf]    = filename nul-terminated
ROM[buf+256]= formatted nul-terminated file DATE$
ROM[blen]   = len of filename
A.BOPT00-63 = size
A.BOPT04-7  = attributes
*/

/// NOTE: the following code assumes CATdat.size is 4 bytes long. If longer or shorter, this needs to change!

	fsize = CATdat.size;
	WRAMD(B_OPT00, fsize);			// save size for AUXROM
	WRAMD(B_OPT04, (DWORD)PatchAttribs(CATdat.attrib));	// save attributes for AUXROM

// format date to ScratBuf+256
    newtime = localtime(&CATdat.time_write);
    if( newtime==NULL ) strcpy((char*)ScratBuf+256, "0000-00-00 00:00");
    else sprintf((char*)ScratBuf+256, "%d-%02d-%02d %02d:%02d", 1900+newtime->tm_year, 1+newtime->tm_mon, newtime->tm_mday, newtime->tm_hour-(newtime->tm_isdst?1:0), newtime->tm_min);

// get just the trailing filename
	for(pChar=CATdat.name+strlen(CATdat.name); pChar>CATdat.name && *(pChar-1)!=SLASH; --pChar) ;	// find the beginning of the last name element
	strcpy((char*)ScratBuf, pChar);
	if( CATdat.attrib & 0x10 ) strcat((char*)ScratBuf, "/");	// SLASH doesn't matter, AUXROM will change all \ and / to the one specified by the user

// write both to ROM[buf]
	WRAMBUF(buf, (BYTE*)ScratBuf, 256+17);	// write filename & date strings to buffer
	WRAMW(blen, strlen((char*)ScratBuf));	// len of filename
	WRAMW(buse, 0);	// OK
}

///*****************************************
/// builds full path in TMPpath, return 0 is OK, -1 if fail
/// converts all '/' and '\\' to SLASH for the native OS.
/// fail if unique==TRUE and wildcards exist, or if wildcards (* ?) are anywhere except in the final element.
int ResolvePath(char *pPath, BOOL unique) {

	char	*pL, *pT, *pN, *pB;

	if( ISSLASH(*pPath) ) {
		strcpy((char*)TMPpath, pPath);	// abs path, ignore cur path
	} else if( strlen(SDcur) + strlen(pPath) > 1022 ) {
		return -1;
	} else {
		strcpy(TMPpath, SDcur);
		strcat(TMPpath, SLASHSTR);
		strcat(TMPpath, pPath);
	}
	// now 'normalize' TMPpath path (removing '.' and '..'; changing any "//" into "/")
	// pT points to the '/' at the start of THIS subdir
	// pL points to the '/' at the start of the LAST subdir
	// pN points to the '/' at the start of the NEXT subdir
	// pB points to the current BYTE we're processing

	// first remove all replicated '\' chars
	for(pB=TMPpath; *pB; ++pB) {
		if( ISSLASH(*pB) && ISSLASH(*(pB+1)) ) {
			strcpy(pB, pB+1);
			*pB-- = SLASH;	// convert to proper SLASH and move back so for() doesn't move us until we retest this SLASH with the NEXT char
		}
	}
	pB = (pT = TMPpath)+1;	// SDcur should ALWAYS be an absolute path, ALWAYS starting with a SLASH (or so THIS code assumes; others may implement otherwise)
	pL = pN = NULL;
	while( *pB ) {	// we're at the beginning of a subdir name with a SLASH one char before us (pointed to by pL)
		for(pN=pB; *pN && !ISSLASH(*pN); pN++) ;	// find NEXT SLASH or end-of-path
		if( pN-pB==1 && *pB=='.' ) {	// THIS subdir is just ".", throw it away
			strcpy(pT, pN);	// just copy the rest of the path down over THIS one
			if( TMPpath[0]==0 ) {	// in case we just blew away starting SLASH (TMPpath was "/."), leave pB->terminating NULL
				strcpy(TMPpath, SLASHSTR);
			}
		} else if( pN-pB==2 && *pB=='.' && *(pB+1)=='.' ) {	// THIS subdir is "..", it cancels out the previous one, so throw them both away IF there's a pL
			if( pT!=TMPpath ) {
				// move pT back one
				pT = pL;
				// move pL back one
				if( pL!=TMPpath ) do --pL; while( !ISSLASH(*pL) );	// if 'older' LAST subdir, seek back for it
				else pL = NULL;	// else there is no older LAST subdir
			}
			/// throw "subdir\.." away
			strcpy(pT, pN);	/// if we're not throwing EVERYTHING away (including the first SLASH), delete the previous subdir and the ".." and the SLASH between them
			/// NOTE: if the path was "/..", the above will set TMPpath="" (with no SLASH). That will abort the WHILE loop, and we'll test for that and error
			pB = pT;
			if( ISSLASH(*pB) ) ++pB;
		} else {	// must be a subdir name
			pL = pT;
			pT = pB = pN;
			if( *pB ) ++pB;	// pN may have been point to the end-of-path
		}
	}
	if( TMPpath[0]==0 ) {	// leading '\' got deleted
		strcpy(TMPpath, SLASHSTR);
	}
	/// now march backward through the path, looking for wildcards
	/// fail if UNIQUE (no wildcards allowed) and there are wildcards
	for(pB=TMPpath+strlen(TMPpath)-1; pB>TMPpath && *pB!='\\'; --pB) if( unique && (*pB=='*' || *pB=='?') ) return -1;
	/// fail if wildcards anywhere but in the last element
	for(; pB>TMPpath; --pB) if( (*pB=='*' || *pB=='?') ) return -1;

	for(pB=TMPpath; *pB; pB++) if( ISSLASH(*pB) ) *pB = '\\';

	return 0;
}

/// For Emulator, file must be in TAPES folder in order to be an emulated TAPE
BOOL IsTAPEfile(char *pS) {

	long	i, j;

	for(i=strlen(pS)-1; i && !ISSLASH(pS[i]); i--) ;	// find beginning of filename
	if( i ) {
		j = i;	// save end of subfolder
		--i;	// get in front of SLASH
		for(; i && !ISSLASH(pS[i]); i--) ;	// find beginning of folder
		++i;	// move back to first char of last sub-folder
		if( j-i==5 && !strnicmp(pS+i, "TAPES", 5) ) {

//GEEKDO: verify that file is appropriate size for tape (802816 bytes)

			return TRUE;
		}
	}
	return FALSE;
}

/// For Emulator, file must be in DISKSn folder in order to be an emulated DISK
BOOL IsDISKfile(char *pS) {

	long	i, j;

	for(i=strlen(pS)-1; i && !ISSLASH(pS[i]); i--) ;	// find beginning of filename
	if( i ) {
		j = i;	// save end of subfolder
		--i;	// get in front of SLASH
		for(; i && !ISSLASH(pS[i]); i--) ;	// find beginning of folder
		++i;	// move back to first char of last sub-folder
		if( j-i==6 && pS[i+5]>='0' && pS[i+5]<='9' && !strnicmp(pS+i, "DISKS", 5) ) {

//GEEKDO: verify that file is appropriate size for disk drive

			return TRUE;
		}
	}
	return FALSE;
}

///*********************
/// NOTE: None of this code is particularly pretty or efficient,
/// but rather is just quickly bashed out for testing purposes.
/// When writing to ROM[], if it's in the ROM/RAM window, then
/// we also have to write to ALL of the AUXROMs that are loaded
/// since that simulated ROM/RAM window must consistently contain
/// the exact same values for ALL of the AUXROMs, regardless of
/// which one is mapped in at any given moment. There's more
/// elegant and efficient ways of doing this, but this required
/// the fewest modifications of other parts of the Emulator, and
/// that took priority for me. So, sort-of-quick and definitely dirty.
void AuxHEYEBTKS(BYTE val) {

	int		i, j=0, romnum, romaddr, romlen, blen, buse, buf;
	long	err, sc, ca, unit;
	WORD	s, t;
	char	*pB;
	DWORD	dw, dn, dd, db;
	BYTE	*pArg, *pFmt;
	double	narg;
	BOOL	seenT, seenD;

	/// val = AUXROM BUF# that's just been mailed to EBTKS

	if( val > 6 ) {	// only MBs 0-6 have buffers, etc., so special case any MBs above that
		if( val < 32 ) {	// only 32 MBs

		} else if( val==0x80 ) {	// enabling ancillary AUX ROMs
			for(j=0; j<numROMS; j++) {
				if( j>0 && ARnum[j] ) {
					pROMS[ARnum[j]][0] = 0361+j;		// unhide all ancillary AUX ROMs
				}
			}
		}
		return;
	}

	/// val is 0-6 at this point
	blen = B_LEN0+2*val;
	buse = B_USE0+2*val;
	buf = B_BUF0+256*val;

	s = *((WORD*)&ROM[buse]);
	WRAMW(buse, 0);	// preset return value to OK
	switch( s ) {
	 case U_WROM:
	 	// B_OPT08 = ROM#
	 	// B_OPT10-11 = ADDRESS
	 	// blen = # bytes
	 	// buf = bytes
		romnum = (int)ROM[B_OPT08] - 0361;					// get rom# from A.BOPT00
		romaddr = (int)(*(((WORD*)&ROM[B_OPT10])));			// get address from A.BOPT02 & A.BOPT03
		romlen = (int)*((WORD*)&ROM[blen]);						// get len of data from A.BLEN0
		if( romnum>=0 && romnum<14 && ARnum[romnum] ) {	// if rom# is in range and that AUXROM is loaded
			if( romaddr+romlen > 0100000 ) romlen = 0100000 - romaddr;	// don't let AUXROM write past end of ROM
			for(i=0; i<romlen; ++i) {	// write as long as we're within the 8K ROM space (trusting AUXROM not to write to ROM/RAM)
				pROMS[ARnum[romnum]][romaddr+i-060000] = ROM[buf+i];
				if( ROM[060000]==romnum+0361 ) ROM[romaddr+i] = ROM[buf+i];
			}
			// now checksum the first 4K of the ROM
			for(i=t=0; i<4096; i+=2) t += *((WORD*)&pROMS[ARnum[romnum]][i]);
			// and the last 1K of the ROM
			for(i=7168; i<8190; i+=2) t += *((WORD*)&pROMS[ARnum[romnum]][i]);
			s = ((t>>16) + (t & 0xffff)) & 0xffff;
// For some reason, this causes a memory fault, as it tries to write past the end of the pROMS[] block, as if it's writing a DWORD, so we write the two bytes individually
//					*((WORD*)&pROMS[ARnum[romnum]][077776]) = (WORD)((WORD)65355-(WORD)s);
			pROMS[ARnum[romnum]][8190] = (255-(s & 0xFF));
			pROMS[ARnum[romnum]][8191] = (255-(s>>8));
			if( ROM[060000]==romnum+0361 ) {
// For some reason, this causes a memory fault, as it tries to write past the end of the pROMS[] block, as if it's writing a DWORD, so we write the two bytes individually
//						*((WORD*)&ROM[077776]) = 65355-s;
				ROM[077776] = (255-(s & 0xFF));
				ROM[077777] = (255-(s>>8));
			}
		}
		break;
	 case U_SDCD:
	 	// NOTE: AUXROM expects returned BUSE to be either 0 (OK) or 214D (Invalid SD path), nothing else!

	 	// we're changing directories, terminate any on-going SDCAT
	 	if( ffUTIL!=-1 ) {
			_findclose(ffUTIL);
			ffUTIL = -1;
	 	}
	 	// AUXROM puts a NUL at the end so we can use strcpy() & such, if needed
	 	if( strlen((char*)(&ROM[buf])) > 1023 ) {
fperr:
			s = errno;
			WRAMW(buse, 217);	// FILE/PATH
			break;
	 	}

		if( ResolvePath((char*)(&ROM[buf]), TRUE) ) goto fperr;	// build full unique path in TMPpath

		i = strlen(TMPpath);
		// now check if the resulting path is a valid path
		if( i>1 ) {	// if not root
			if( ISSLASH(TMPpath[i-1]) ) TMPpath[i-1] = 0;	// get rid of trailing SLASH, but not if it's also the leading SLASH
			ffUTIL = _findfirst(TMPpath, &UTILdat);
			if( ffUTIL==-1 || !(UTILdat.attrib & _A_SUBDIR) ) goto fperr;
			else _findclose(ffUTIL);
			ffUTIL = -1;
			strcat(TMPpath, SLASHSTR);
		}
		strcpy(SDcur, TMPpath);	// it's OK, so copy it as the official current path
		break;
	 case U_SDCUR:
	 	// A.BOPT00 =0 for SDCUR$, =1 for SDHOME$
	 	// can't use strcpy() going this way, as ALL copies of AUXROMs have to be modified
		if( ROM[B_OPT00]==0 ) {
			WRAMBUF(buf, (BYTE*)SDcur, strlen(SDcur));
			WRAMW(blen, strlen(SDcur));
		} else if( ROM[B_OPT00]==1 ) {
			WRAMBUF(buf, (BYTE*)SDhome, strlen(SDhome));
			WRAMW(blen, strlen(SDhome));
		} else goto sderr;
		break;
	 case U_SDCAT:	// used by SDCAT, SDSIZE, SDATTR
	 	// A.BOPT00=0 to "find first," =1 to "find next."
	 	// A.BOPT01=0 if fileSpec$ may have wildcards, =1 if fileSpec$ is UNIQUE (for SDATTR & SDSIZE)
	 	// buf = NUL-terminated fileSpec$ string
	 	// returns BUSE=0 (if OK) or =217D or =213D if error
	 	// if OK, returns buf=string, blen=strlen, buf+256=date$, BUSE=0, BOPT00-BOP63=file size, BOPT04-BOPT07=file flags/attributes (extra big for AUXROM convenience)
	 	i = (int)ROM[B_OPT00];//pROMS[ARnum[0]][B_OPT00-060000];
		if( !i ) {								/// find first
			if( ffCAT>=0 ) {
				_findclose(ffCAT);	// close the previous one
				ffCAT = -1;
			}
			if( ResolvePath((char*)(&ROM[buf]), (BOOL)ROM[B_OPT01]) ) goto fperr;	/// build absolute path in TMPpath, also fail if unique & wildcards exist

			for(pB=TMPpath+strlen(TMPpath)-1; pB>TMPpath && !ISSLASH(*pB); --pB) if( (*pB=='*' || *pB=='?') ) break;	/// look for wild cards
			/// if not UNIQUE, but path ends in SLASH, add "*" to end
			if( !(BOOL)ROM[B_OPT01] && ISSLASH(*pB) && pB==TMPpath+strlen(TMPpath)-1 ) {
				strcat(TMPpath, "*");
				++pB;
			}
			Pattern[0] = 0;
			if( !(BOOL)ROM[B_OPT01] && !ISSLASH(*pB) ) {	/// has WILDCARDS, treat as file name
				while( pB>TMPpath && !ISSLASH(*pB) ) --pB;	// point pB at beginning of name element
				++pB;
				/// save tail end for pattern matching (whether FIRST or NEXT)
				strcpy(Pattern, (char*)pB);
				for(pB=TMPpath; *pB; pB++) if( ISSLASH(*pB) ) *pB = '\\';
				ffCAT = _findfirst((char*)TMPpath, &CATdat);	// first try it as if the last element is a FILE

				WRAMW(blen, 0);			// len of buffer
				WRAMW(buse, 1);			// END
				while( ffCAT!=-1 ) {
					if( MatchesPattern(CATdat.name, Pattern) ) {
						RetCat(buf, blen, buse);
						break;
					}
fnext:
					if( (err=_findnext(ffCAT, &CATdat)) ) {
						_findclose(ffCAT);
						ffCAT = -1;
					}
				}
			} else {	/// NO wildcards
				ffCAT = _findfirst((char*)TMPpath, &CATdat);	// first try it as if the last element is a FILE
				if( ffCAT==-1 ) goto fperr;
				RetCat(buf, blen, buse);
			}
		} else if( ffCAT==-1 ) {	// no SDCAT active
			// return ERROR
sderr:
			i = errno;
			WRAMW(blen, 0);			// len of buffer
			WRAMW(buse, 213);	// SD ERROR
		} else {								/// find next
			WRAMW(blen, 0);			// len of buffer
			WRAMW(buse, 1);	// END
			goto fnext;
		}
		break;
	 case U_SDFLUSH:	// flush a buffer or everything
	 	s = (WORD)ROM[B_OPT00];	// get file#
	 	if( s==0 ) {	// flush everything

	 	} else {	// else flush file# s-1

	 	}
		break;
	 case U_SDOPEN:
	 	// BOPT00 = FILE# (already checked by AUXROM and guaranteed to be 1-NUMAUXFILES
	 	// BOPT01 = MODE (already checked by AUXROM and guaranteed to be 0, 1, or 2
	 	// BUF = name (NUL-terminated, i.e., a C-string)
	 	s = ROM[B_OPT00]-1;	// file# (1 to NUMAUXFILES)->(0 to NUMAUXFILES-1)
	 	if( SDfiles[s].handle!=-1 ) goto sderr;	// make sure file# is not already in use

		if( ResolvePath((char*)(&ROM[buf]), TRUE) ) goto fperr;	// build full unique path in TMPpath

		switch( ROM[B_OPT01] ) {	// mode
		 case 0:	// READ-ONLY, error if the file doesn't exist
		 	if( -1==(SDfiles[s].handle = _open((char*)TMPpath, _O_RDONLY | _O_BINARY, _S_IREAD | _S_IWRITE)) ) goto fperr;
			break;
		 case 1:	// R/W, create it if it doesn't exist, set R/W ptr to end-of-file
		 	if( -1==(SDfiles[s].handle = _open((char*)TMPpath, _O_RDWR | _O_CREAT | _O_BINARY, _S_IREAD | _S_IWRITE)) ) goto fperr;
		 	_lseek(SDfiles[s].handle, 0, SEEK_END);
			break;
		 default:	// R/W, create it if it doesn't exist, truncate it to 0 length if it does exist, set R/W ptr to start-of-file
		 	if( -1==(SDfiles[s].handle = _open((char*)TMPpath, _O_RDWR | _O_CREAT | _O_TRUNC | _O_BINARY, _S_IREAD | _S_IWRITE)) ) goto fperr;
			break;
		}
		break;
	 case U_SDREAD:
		// BOPT00 = file#
		// ROM[blen] = BytesToRead
		// &ROM[buf] = where to read the bytes)

	 	s = ROM[B_OPT00]-1;	// file# (1 to NUMAUXFILES)->(0 to NUMAUXFILES-1)
	 	if( SDfiles[s].handle==-1 ) goto sderr;	// error if file# is not OPEN

	 	i = (int)*((WORD*)&ROM[blen]);	// get BytesToRead
	 	pB = (char*)&ROM[buf];			// pointer for _read()

	 	i = _read(SDfiles[s].handle, pB, i);	 	// read them
	 	WRAMW(blen, (WORD)i);	 			// return the number read
 		// Read into ROM/RAM, spread it to the other ROM images, too
		for(; i--; ) WRAMB(buf+i, ROM[buf+i]);
		break;
	 case U_SDWRIT:
		// BOPT00 = file#
		// BLEN = NumBytesToWrite
		// BUF = BytesToWrite

	 	s = ROM[B_OPT00]-1;	// file# (1 to NUMAUXFILES)->(0 to NUMAUXFILES-1)
	 	if( SDfiles[s].handle==-1 ) goto sderr;	// error if file# is not OPEN

	 	i = (int)*((WORD*)&ROM[blen]);	// get BytesToWrite
	 	pB = (char*)&ROM[buf];			// make it a pointer for _write()

	 	if( i!=_write(SDfiles[s].handle, pB, i) ) goto sderr; 	// write them

		break;
	 case U_SDCLOS:
		// BOPT00 = file#

	 	s = ROM[B_OPT00];	// file# (0-NUMAUXFILES)
	 	if( s ) {
			--s;
			if( SDfiles[s].handle==-1 ) {
				// report error
				WRAMB(buse, 213);
				break;
			} else {
				_close(SDfiles[s].handle);
				SDfiles[s].handle = -1;
			}
	 	} else {
			for(t=0; t<10; t++) {
				if( SDfiles[t].handle!=-1 ) {
					_close(SDfiles[t].handle);
					SDfiles[t].handle = -1;
				}
			}
	 	}
		break;
	 case U_SDSEEK:
	 	// BOPT00 = file#
	 	// BOPT01 = mode (0=absolute position, 1=advance from current position, 2=go to end of file)
	 	// BOPT04-07 = offset (signed value!!!)
	 	s = ROM[B_OPT00]-1;	// file# (1 to NUMAUXFILES)->(0 to NUMAUXFILES-1)
	 	if( SDfiles[s].handle==-1 ) goto sderr;
	 	dw = *((DWORD*)&ROM[B_OPT04]);	// get desired offset
	 	i = ROM[B_OPT01];	// mode

	 	dn = (DWORD)_lseek(SDfiles[s].handle, 0, SEEK_CUR);	// save where we are now
	 	dd = (DWORD)_lseek(SDfiles[s].handle, 0, SEEK_END);	// get end of file
		_lseek(SDfiles[s].handle, dn, SEEK_SET);			// return to the 'current' location

	 	// MODE 0: offset from start of file - must be >= 0
	 	// MODE 1: relative offset from cur position - must be >= -dn (current position) and <= (dd-dn) (distance to end of file)
	 	// MODE 2: offset from end of file - must be <= 0
	 	if( i==0 ) db = dw;
		else if( i==1 ) db = dn+dw;
		else db = dd+dw;

		if( (db & 0x80000000) || db>dd ) goto sderr;
 		db = (DWORD)_lseek(SDfiles[s].handle, db, SEEK_SET);

		WRAMD(B_OPT04, db);
		break;
	 case U_SDDEL:
	 	// buf = NUL-terminated fileSpec$ string
	 	// returns BUSE=0 (if OK) or =217D or =213D if error
		i = 0;	// count of files deleted
		if( ffUTIL>=0 ) {
			_findclose(ffUTIL);	// close the previous one
			ffUTIL = -1;
		}
		if( ResolvePath((char*)(&ROM[buf]), FALSE) ) goto fperr;	/// build absolute path in TMPpath

		for(pB=TMPpath+strlen(TMPpath)-1; pB>TMPpath && !ISSLASH(*pB); --pB) if( (*pB=='*' || *pB=='?') ) break;	/// look for wild cards
		/// if no wildcards, but path ends in SLASH, add "*" to end, to delete all files in a directory via "SDDEL somedir/"
/*		if( ISSLASH(*pB) && pB==TMPpath+strlen(TMPpath)-1 ) {
			strcat(TMPpath, "*");
			++pB;
		}
*/
		Pattern[0] = 0;
		while( pB>TMPpath && !ISSLASH(*pB) ) --pB;	// point pB at beginning of name element
		++pB;
		/// save tail end for pattern matching (whether FIRST or NEXT)
		strcpy(Pattern, (char*)pB);
		for(pB=TMPpath; *pB; pB++) if( ISSLASH(*pB) ) *pB = '\\';
		for(--pB; !ISSLASH(*pB); --pB) ;
		++pB;	// point just past last slash
		ffUTIL = _findfirst((char*)TMPpath, &UTILdat);
		while( ffUTIL!=-1 ) {
			if( stricmp(UTILdat.name,".") && stricmp(UTILdat.name,"..") && MatchesPattern(UTILdat.name, Pattern) ) {
//				ResolvePath(UTILdat.name, FALSE);
				strcpy((char*)pB, UTILdat.name);
				if( -1==_unlink(TMPpath) ) goto sderr;
				++i;
			}
			if( (err=_findnext(ffUTIL, &UTILdat)) ) {
				_findclose(ffUTIL);
				ffUTIL = -1;
			}
		}
		if( !i ) WRAMW(buse, 217);	// return FILE/PATH if no files deleted
		break;
	 case U_MKDIR:
		// blen = len of directoryName in BUF6
		// buf = directoryName (NUL-terminated)

		if( ResolvePath((char*)(&ROM[buf]), TRUE) ) goto fperr;	// build full unique path in TMPpath
		if( -1==_mkdir((char*)TMPpath) ) goto sderr;

		break;
	 case U_RMDIR:
		// blen = len of directoryName in BUF6
		// buf = directoryName (NUL-terminated)

		if( ResolvePath((char*)(&ROM[buf]), TRUE) ) goto fperr;	// build full unique path in TMPpath
		if( -1==_rmdir((char*)TMPpath) ) goto sderr;

		break;
	 case U_SPF:	// SPRINTF
	 	// blen = len of stuff in buf
	 	// buf = sprintf information:
	 	//	BYTE	2			string type
	 	//	WORD	strlen		len of FORMAT$
	 	//	BYTE	[strlen]	bytes of FORMAT$
	 	// The FORMAT$ in buf is then optionally followed by numeric and string arguments until a TYPE of 0xFF is seen.
	 	// The TYPEs are:
	 	//	BYTE	0			REAL type
	 	//	BYTE	[8]			8 bytes of REAL# value
	 	//
	 	//	BYTE	1			INT type
	 	//	BYTE	[3]			3 bytes of INT# value
	 	//
	 	//	BYTE	2			string type
	 	//	WORD	strlen		len of string
	 	//	BYTE	[strlen]	bytes of string value
	 	//
	 	//	BYTE	0xFF		end of argument list

	 	pFmt = &ROM[buf];	// point to buf
	 	if( *pFmt++!=2 ) goto badformat;	// jif no FORMAT$ (should never happen, should be caught by AUXROM, but let's be safe, in case an IDIOT wrote the AUXROM)
	 	t = *((WORD*)pFmt);	// get len of format$
	 	pFmt += 2;			// skip len of format$
		pArg = pFmt + t;	// skip format$ to point to first arg

	 	j = 0;				// how many chars written to ScratBuf (limit of 1024)
	 	// pFmt points to current char in format$
	 	// t = remaining chars in format$ (including the current one). Remember: format$ is NOT NUL-terminated!
	 	// pArg points to the next arg's TYPE byte in buf
	 	// j = index into ScratBuf of where to store next formatted output
		while( j<1024 && t ) {	// process format$ until buffer is full or run out of format$
			int sb, q;

			if( *pFmt!='%' || (t>1 && *(pFmt+1)=='%') ) {
				if( *pFmt=='%' ) {
					--t;
					sb = *pFmt++;	// skip first %, output the second
					++pFmt;
					--t;	// use up second %
				} else if( *pFmt=='\\' ) {
					++pFmt;	// skip the '\'
					--t;
					if( *pFmt=='\\' ) {--t; sb = *pFmt++;}	// skipped the first one, output the second
					else if( toupper(*pFmt)=='r' ) {sb = 13; ++pFmt; --t;}
					else if( toupper(*pFmt)=='n' ) {sb = 10; ++pFmt; --t;}
					else if( toupper(*pFmt)=='t' ) {sb =  9; ++pFmt; --t;}
					else if( toupper(*pFmt)=='x' ) {	// handle \xHH two hex digit char
						++pFmt;	// skip x
						--t;
						if( !isxdigit(*pFmt) || !isxdigit(*(pFmt+1)) ) goto badformat;
						sb = (*pFmt-'0' - ((*pFmt>'9')?7:0))<<4;
						++pFmt;
						--t;
						sb += *pFmt-'0' - ((*pFmt>'9')?7:0);
						++pFmt;
						--t;
					} else if( isdigit(*pFmt) ) {	// handle "\nnn" octal number char
						for(sb=q=0; q<3 && isdigit(*pFmt); q++, t--, pFmt++) sb = (sb<<3) + *pFmt - '0';	// add in the digit
					} else goto badformat;
				} else {
					sb = *pFmt++;
					--t;
				}
				ScratBuf[j++] = sb;
			} else {	/// handle %
				BYTE	*pCopy;
				int		slen;

				pCopy = ScratBuf+1024;
				/// copy '%' and any FLAGS
				do *pCopy++ = *pFmt++;
				while( --t && (*pFmt=='-' || *pFmt=='+' || *pFmt==' ' || *pFmt=='0' || *pFmt=='\'' || *pFmt=='#') );
				/// copy WIDTH field
				if( t && *pFmt=='*' && *pArg!=0xFF ) {	// if '*' and there's args left on the list
					// substitute a value from the ARG LIST
subarg:
					++pFmt;	// skip '*'
					--t;	// dec remaining char count
					if( *pArg==0 ) {
						narg = cvt_HP85_real_to_IEEE_double(pArg+1);	// get REAL arg in the usable format
						pArg += 9;	// skip tag and 8-byte value
					} else if( *pArg==1 ) {
						*pArg = 0xFF;	// change tag to tagged-INT tag
						narg = cvt_HP85_real_to_IEEE_double(pArg-4);	// get tagged-INT arg in the usable format
						pArg += 4;	// skip tag and 3-byte value
					} else {
argmismatch:
						WRAMW(buse, 220);	// FORMAT/ARG MISMATCH
						break;
					}
					sprintf((char*)pCopy, "%d", (int)narg);	// hope and pray that narg IS an integer
					pCopy += strlen((char*)pCopy);	// advance pCopy by however many chars we output to the temp format string
				} else {
					// copy any WIDTH number that's there
					while( t && isdigit(*pFmt) ) {*pCopy++ = *pFmt++; --t;}
				}
				/// copy possible WIDTH/PRECISION separating '.'
				if( t && *pFmt=='.' ) {*pCopy++ = *pFmt++; --t;}
				/// copy PRECISION field
				if( t && *pFmt=='*' ) {	// if the user does something stupid like "%*.***d", they'll go to subarg multiple times and get what they deserve
					// substitute a value from the ARG LIST
					goto subarg;
				} else {
					// copy any PRECISION number that's there
					while( t && isdigit(*pFmt) ) {*pCopy++ = *pFmt++; --t;}
				}
				/// NOTE: LENGTH field NOT supported

				/// get TYPE field
				if( t-- ) {
					if( *pFmt=='X' || *pFmt=='E' || *pFmt=='G' ) *pCopy++ = *pFmt;	// copy the TYPE char (maintain uppercase)
					else *pCopy++ = tolower(*pFmt);	// copy the TYPE char (force uppercase to lowercase)
					*pCopy = 0;			// NUL-term the copied format$ piece
					switch( tolower(*pFmt++) ) {
					 case 'd':
					 case 'i':
					 case 'u':
					 case 'x':
					 case 'o':
					 	// get next (numeric!) arg
						if( *pArg==0 ) {
							narg = cvt_HP85_real_to_IEEE_double(pArg+1);	// get REAL arg in the usable format
							pArg += 9;	// skip tag and 8-byte value
						} else if( *pArg==1 ) {
							*pArg = 0xFF;	// change tag to tagged-INT tag
							narg = cvt_HP85_real_to_IEEE_double(pArg-4);	// get tagged-INT arg in the usable format
							pArg += 4;	// skip tag and 3-byte value
						} else goto argmismatch;
						sprintf((char*)ScratBuf+j, (char*)ScratBuf+1024, (long)narg);
						break;
					 case 'f':
					 case 'e':
					 case 'g':
					 	// get next (numeric!) arg
						if( *pArg==0 ) {
							narg = cvt_HP85_real_to_IEEE_double(pArg+1);	// get REAL arg in the usable format
							pArg += 9;	// skip tag and 8-byte value
						} else if( *pArg==1 ) {
							*pArg = 0xFF;	// change tag to tagged-INT tag
							narg = cvt_HP85_real_to_IEEE_double(pArg-4);	// get tagged-INT arg in the usable format
							pArg += 4;	// skip tag and 3-byte value
						} else goto argmismatch;
						sprintf((char*)ScratBuf+j, (char*)ScratBuf+1024, narg);
						break;
					 case 's':
					 	// get the next (string!) arg
						if( *pArg!=2 ) goto argmismatch;	// fail if next arg is not STRING!
						++pArg;	// skip to len
						slen = *((WORD*)pArg);	// get $ len
						pArg += 2;	// skip len
						if( slen+j>1024 ) goto badformat;
						memcpy(ScratBuf+2048, pArg, slen);	// copy string so we can NUL-term it
						pArg += slen;			// skip string arg
						ScratBuf[2048+slen] = 0;	// NUL-term
						sprintf((char*)ScratBuf+j, (char*)ScratBuf+1024, (char*)ScratBuf+2048);	// copy the string to the output according to the %s formatting
						break;
					 case 'c':
					 	// get next numeric arg and make it a char
					 	// or get the next string arg and use the first char
						if( *pArg==0 ) {
							narg = cvt_HP85_real_to_IEEE_double(pArg+1);	// get REAL arg in the usable format
							pArg += 9;	// skip tag and 8-byte value
						} else if( *pArg==1 ) {
							*pArg = 0xFF;	// change tag to tagged-INT tag
							narg = cvt_HP85_real_to_IEEE_double(pArg-4);	// get tagged-INT arg in the usable format
							pArg += 4;	// skip tag and 3-byte value
						} else if( *pArg==2 ) {
							++pArg;	// point to len
							slen = *((WORD*)pArg);	// get $ len
							if( slen==0 ) goto badformat;
							pArg += 2;
							narg = (double)(long)(*pArg);	// get first char of string as numeric value
							pArg += slen;		// skip string to next arg
						} else goto badformat;
						ScratBuf[j++] = (BYTE)(long)narg;
						ScratBuf[j] = 0;	// so advance of j via strlen() below works
						break;
					 case 'p':
					 case 'a':
					 case 'n':
					 default:
badformat:
						WRAMW(buse, 219);	// BAD FORMAT
						goto spfexit;
					}
					j += strlen((char*)ScratBuf+j);
				}
			}
		}
spfexit:
		// copy formatted output to buf
		WRAMBUF(buf, ScratBuf, j);
		// set returned length
		WRAMW(blen, (WORD)j);
		break;
	 case U_MOUNT:
		/// buf = filePath$ (NUL-terminated)
		/// buf+256 = msus$ (NUL-terminated)
		/// BOPT00 = 0=mount, error if not there; 1=mount, create if not there; 2=error if there, create & mount

		/// return OK or error code in B_USE0
		/// Possible errors:
		///		14D INVALID MODE
		///		15D INVALID MSUS (MSUS is bad, or points to REAL HARDWARE rather than EMULATED tape/disk)
		///		16D BAD SD PATH (ResolvePath() error)
		///		17D FILE/PATH (file open error)

		if( ResolvePath((char*)(&ROM[buf]), TRUE) ) {
			WRAMW(buse, 216);	// BAD SD PATH
			break;
		}
		switch( GetDevType((char*)(&ROM[buf+256])) ) {
		 case DEVTYPE_TAPE:	// tape
		 case DEVTYPE_5_25:	// 5.25" floppy
		 case DEVTYPE_3_5:	// 3.5" floppy
		 case DEVTYPE_8:	// 8" floppy
		 case DEVTYPE_5MB:	// 5MB hard disk
			seenT = IsTAPEfile(TMPpath);	// see if the file is in the TAPES subfolder
			seenD = IsDISKfile(TMPpath);	// see if the file is in one of the DISKSn subfolders

			/// needs an EmulatedMSUS(char*) function that returns TRUE if a valid emulated (SD) tape/disk, else returns FALSE
			/// Or, do something like...

			/// chew up MSUS$. If looks ugly or isn't an emulated tape or disk drive, error
			/// to be valid disk file, sc<8, (sc,ca,unit) specifies an emulated drive, and the folder containing the file is .../DISKSn/...
			/// to be valid tape file, sc==8, and the folder containing the file is .../TAPES/...
			if( !ProcessMSUS((char*)(&ROM[buf+256]), &sc, &ca, &unit) || (sc<8 && (!IsUnit(sc, ca) || !seenD)) || (sc==8 && !seenT) ) {
				WRAMW(buse, 215);	// INVALID MSUS
				break;
			}
			if( sc>7 ) {	// MOUNT ":T:
				EjectTape();	// remove any tape in the drive
				/// need to use ONLY the filename, not the path, as the tape files HAVE to be in the TAPES folder for the Emulator
				for(i=strlen(TMPpath)-1; i && !ISSLASH(TMPpath[i]); i--);
				i++;
				if( strlen(TMPpath)-i >= 63 ) {
	filepatherr:
					WRAMW(buse, 217);	// FILE/PATH
					break;
				}
				/// Try to load the tape
				strcpy(CurTape, TMPpath+i);	// name it
				switch( ROM[B_OPT00] ) {
				 case 0:	// mount, error if fail
					if( !LoadTape() ) goto filepatherr;
					break;
				 case 1:	// mount, create if not there
					if( !LoadTape() ) {
	newtap:
						NewTape();				// create and load a new tape
						WriteTape();			// write it to disk
						EjectTape();			// eject it
						if( !LoadTape() ) goto filepatherr;
					}
					break;
				 case 2:	// error if exists, else create & mount
					if( LoadTape() ) {	// if file exists (and loads), error
						EjectTape();
						WRAMW(buse, 214);	// INVALID MODE
					} goto newtap;	// file doesn't exist, create it and load it
					break;
				}
				DrawAll();
			} else {		// MOUNT ":D..."
				/// UNMOUNT any tape/disk in the device, then try to OPEN/MOUNT the TMPpath[] SD file as a proper device media
				EjectDisk(sc, ca, unit);	// eject it
				/// need to use ONLY the filename and diskfolder portion of the path, as the disk files HAVE to be in the DISKSn folderf for the Emulator
				for(i=strlen(TMPpath)-1; i && !ISSLASH(TMPpath[i]); i--);
				TMPpath[i] = 0;	// turn '\' into NUL for DISKn terminator
				for(j=i-1; j>0 && !ISSLASH(TMPpath[j]); j--);
				++j;	// point to start of (hopefully) DISKSn
				++i;	// point to start of disk filename
				if( strnicmp(TMPpath+j, "DISKS",5) ) goto filepatherr;
				j = TMPpath[j+5]-'0';	// get (hopefully) folder#
				if( j<0 || j>9 ) goto filepatherr;
				SetDiskFolder(sc, ca, unit, j);
				if( strlen(TMPpath)-1 >= 127 ) goto filepatherr;
				switch( ROM[B_OPT00] ) {
				 case 0:	// mount, error if not there
					if( !LoadDisk(sc, ca, unit, (BYTE*)TMPpath+i) ) goto filepatherr;
					break;
				 case 1:	// mount, create if not there
					if( !LoadDisk(sc, ca, unit, (BYTE*)TMPpath+i) ) {
newdis:
						NewDisk(sc, ca, unit, (BYTE*)TMPpath+i);
						EjectDisk(sc, ca, unit);			// eject it (to write it)
						if( !LoadDisk(sc, ca, unit, (BYTE*)TMPpath+i) ) goto filepatherr;
					}
					break;
				 case 2:	// eror if there, create & mount
					if( LoadDisk(sc, ca, unit, (BYTE*)TMPpath+i) ) {	// if file exists (and loads), error
						EjectDisk(sc, ca, unit);
						WRAMW(buse, 214);	// INVALID MODE
					} goto newdis;	// file doesn't exist, create it and load it
					break;
				}
				DrawDisks(sc, ca);
			}


			break;
		 case DEVTYPE_FILEPRT:
		 	if( GetDevSCTLA((char*)(&ROM[buse+256]), &sc, &ca) ) {	// chew up the MSUS$ string and return S.C. and T.L.A
				if( !OpenFilePRT(sc, ca, TMPpath, ROM[B_OPT00]) ) {	// open/create the desired print-to-file filename using the desired mode
					WRAMW(buse, 217);	// FILE/PATH
				}
				break;
		 	}
		 case DEVTYPE_HARDPRT:	// error for HARD printer, can't MOUNT via software
		 default:	// error
			WRAMW(buse, 227);	// BAD DEVICE
			break;
		}
		break;
	 case U_UNMOUNT:
		/// blen = len of msus$ in buf
		/// buf = msus$ (NUL-terminated)

		/// return 0 or error code in buse
		/// Possible errors:
		///		any non-zero value = 15D INVALID MSUS (MSUS is bad, or points to REAL HARDWARE rather than EMULATED tape/disk)

		/// needs an EmulatedMSUS(char*) function that returns TRUE if a valid emulated (SD) tape/disk, else returns FALSE
		/// Or, do something like...

		/// chew up MSUS$. If looks ugly or isn't an emulated tape or disk drive, error
		/// to be valid disk file, sc<8, (sc,ca,unit) specifies an emulated drive, and the folder containing the file is .../DISKSn/...
		/// to be valid tape file, sc==8, and the folder containing the file is .../TAPES/...
		switch( GetDevType((char*)(&ROM[buf])) ) {
		 case DEVTYPE_TAPE:	// tape
		 case DEVTYPE_5_25:	// 5.25" floppy
		 case DEVTYPE_3_5:	// 3.5" floppy
		 case DEVTYPE_8:	// 8" floppy
		 case DEVTYPE_5MB:	// 5MB hard disk
			if( !ProcessMSUS((char*)(&ROM[buf]), &sc, &ca, &unit) || (sc<8 && !IsUnit(sc, ca)) ) {
				WRAMW(buse, 215);	// INVALID MSUS
			} else {
				if( sc>7 ) {	// MOUNT ":T:
					EjectTape();	// remove any tape in the drive
					DrawAll();
				} else {		// MOUNT ":D..."
					EjectDisk(sc, ca, unit);	// eject it
					DrawDisks(sc, ca);
				}
			}
			break;
		 case DEVTYPE_FILEPRT:
		 	if( GetDevSCTLA((char*)(&ROM[buse]), &sc, &ca) ) {
				HPIBdevFlush(sc, ca, 0);
				break;
		 	}
		 case DEVTYPE_HARDPRT:	// error for HARD printer, can't UNMOUNT via software
		 default:	// error
			WRAMW(buse, 227);	// BAD DEVICE
			break;
		}
		break;
	 case U_FLAGS:
		/// A.FLAGS (70224-70227) have changed. Write them to the config file so they can be reloaded at next power-on
		/// The Emulator saves them when everything else is written to the .ini file, but EBTKS needs to write them
		/// RIGHT AWAY, in case the user POWERS-DOWN the 85.
		break;
	 case U_SDREN:
		/// buf+256 (BUF1) = newName$ (NUL-terminated)
		/// buf = oldName$ (NUL-terminated)

		/// return OK or error code in B_USE0
		/// Possible errors:
		///		14D INVALID MODE
		///		15D INVALID MSUS (MSUS is bad, or points to REAL HARDWARE rather than EMULATED tape/disk)
		///		16D BAD SD PATH (ResolvePath() error)
		///		17D FILE/PATH (file open error)

	 	// preset return BUSE0 = 0 for OK
	 	WRAMB(buse, 0);
		// turn newName into a fully resolved pathname
		if( ResolvePath((char*)(&ROM[buf+256]), TRUE) ) {
			WRAMW(buse, 216);	// BAD SD PATH
			break;
		}
		// save the newName
		strcpy((char*)ScratBuf, TMPpath);
		// turn oldName into a fully resolved pathname
		if( ResolvePath((char*)(&ROM[buf]), TRUE) ) {
			WRAMW(buse, 216);	// BAD SD PATH
			break;
		}

//GEEKDO: check if open or mounted

		if( rename(TMPpath, (char*)ScratBuf) ) {
			i = errno;
			WRAMW(buse, 213);	// SD ERROR
		}
		break;
	 case U_CLOCK:	// return Real Time Clock
	 	/// WARNING! Ignore the following. Work in progress... :-)

		/// On return:
		///		A.BOPT00-03 = binary count of seconds since midnight
		///		A.BOPT04-05 = day (1-31)
		///		A.BOPT06-07 = month (1-12)
		///		A.BOPT08-09 = year (1970, 1971, ..., 2020, 2021, 2022, etc.)
		{
			KTIME	kT;

			KGetTime(&kT);
			WRAMD(B_OPT00, kT.s + kT.mi*60 + kT.h*3600);
			WRAMW(B_OPT04, kT.d);
			WRAMW(B_OPT06, kT.mo);
			WRAMW(B_OPT08, kT.y);
		}
	 	WRAMB(buse, 0);	// return 0 if okay
		break;
	 case U_HELP:	// throws up a dummy HELP screen
		///	blen = len of requested HELP$
		///	buf = text of requested HELP$

		/// NOTE: For this dummy, graphics is ignored, we assume we're in alpha mode.
		/// Currently, HELP is programmable, so it COULD execute while in graphics mode,
		/// which is a good argument, I suppose, for making it NON-programmable, EXCEPT
		/// the user could type:
		///		GRAPHICS @ HELP ME
		/// and you'd still be in GRAPHICS when it runs, and I can imagine scenarios
		/// where folks might WANT to use HELP in a program, so best to deal with that eventually.

		/// for now, just save the current alpha screen, write the dummy help screen, then
		/// return WITHOUT releasing buffer 6. This will leave the AUXROM cooling its heels
		/// until a key is pressed, which we'll somehow trap and then reverse this process
		/// before releasing buffer 6.

		/// A.BOPT00 =1 to save the CRT state, =0 to restore the CRT state
		if( Model<2 ) {
			if( ROM[B_OPT00] ) {	// save CRT
				ScreenSaved = TRUE;	// we don't check if already saved, in case someone gets out of sync, so this will allow them to get back into sync
				memcpy(SaveScreen, CRT, 8192);
				SaveCRTBAD = IO_CRTBAD;
				SaveCRTSAD = IO_CRTSAD;
				SaveCRTCTL = IO_CRTCTL;
			} else if( ScreenSaved ) {			// restore CRT, don't restore if hasn't been saved, to avoid garbage on display
				ScreenSaved = FALSE;
				memcpy(CRT, SaveScreen, 8192);
				IO_CRTBAD = SaveCRTBAD;
				IO_CRTSAD = SaveCRTSAD;
				IO_CRTCTL = SaveCRTCTL;
				HP85InvalidateCRT();
			}
		} else {
			if( ROM[B_OPT00] ) {	// save CRT
				ScreenSaved = TRUE;	// we don't check if already saved, in case someone gets out of sync, so this will allow them to get back into sync
				memcpy(SaveScreen, CRT, 16384);
				SaveCRTBAD = IO_CRTBAD;
				SaveCRTSAD = IO_CRTSAD;
				SaveCRTCTL = IO_CRTCTL;
			} else if( ScreenSaved ) {			// restore CRT, don't restore if hasn't been saved, to avoid garbage on display
				ScreenSaved = FALSE;
				memcpy(CRT, SaveScreen, 16384);
				IO_CRTBAD = SaveCRTBAD;
				IO_CRTSAD = SaveCRTSAD;
				IO_CRTCTL = SaveCRTCTL;
				HP85InvalidateCRT();
			}
		}
/*
		pB = SaveScreen;
		/// for purposes of this demo, we ignore the case where some nimnuts set CRTBAD to an ODD address
		j = IO_CRTSAD;
		for(i=0; i<16*32; i++) {	// save screen
			*pB++ = CRT[j/2];
			j = (j+2) & 0x0FFF;	// bump CRT address and keep within ALPHA
		}
		/// now write the demo help screen
		pB = HelpScreen;
		j = IO_CRTSAD;
		for(i=0; i<16*32; i++) {	// write screen
			CRT[j/2] = *pB++;
			j = (j+2) & 0x0FFF;	// bump CRT address and keep within ALPHA
		}
		HP85InvalidateCRT();

		sendHelpKey = TRUE;	// set flag so when Emulator receives a key, it will send it to HelpKey() in this file
*/
	 	WRAMB(buse, 0);	// return 0 if okay, return 224D if help subject not found/recognized
		break;	// return without releasing MB6
	 case U_MEDIA:
		/// blen = len of msus$ in BUF6
		/// buf = msus$ (NUL-terminated)

		/// return 0 or error code in B_USE0
		/// Possible errors:
		///		any non-zero value = 15D INVALID MSUS (MSUS is bad, or points to REAL HARDWARE rather than EMULATED tape/disk)

		/// needs an EmulatedMSUS(char*) function that returns TRUE if a valid emulated (SD) tape/disk, else returns FALSE
		/// Or, do something like...

		/// chew up MSUS$. If looks ugly or isn't an emulated tape or disk drive, error
		/// to be valid disk file, sc<8, (sc,ca,unit) specifies an emulated drive, and the folder containing the file is .../DISKSn/...
		/// to be valid tape file, sc==8, and the folder containing the file is .../TAPES/...
		if( !ProcessMSUS((char*)(&ROM[buf]), &sc, &ca, &unit) || (sc<8 && !IsUnit(sc, ca)) ) {
			WRAMW(buse, 215);	// INVALID MSUS
		} else {
			if( sc>7 ) {
				if( CurTape[0]!=0 ) sprintf((char*)ScratBuf, "TAPES/%s", CurTape);
				else ScratBuf[0] = 0;
				pB = (char*)ScratBuf;
			} else {
				pB = GetDiskName(sc, ca, unit);
				if( pB[0] ) pB = GetDiskPathName(sc, ca, unit);
			}
			if( pB[0] ) ResolvePath(pB, TRUE);
			else TMPpath[0] = 0;
			i = strlen(TMPpath);
			WRAMBUF(buf, (BYTE*)TMPpath, i);
			WRAMW(blen, i);
		}
		break;
	 case U_MEMCPY:
	 	// BOPT00-62 = 3-byte SRC address
	 	// BOPT03 =0 if SRC address is lowest byte of block, =1 if SRC address is (highest byte + 1) of block (ie, fetch bytes in reverse order)
	 	// BOPT04-66 = 3-byte DST address
	 	// BOPT07 =0 if DST address is lowest byte of block, =1 if DST address is (highest byte + 1) of block (ie, store bytes in reverse order)
        // COPT0-2 = # of bytes to copy (3 byte value)
		{
			BYTE	*pS, *pD;
			BOOL	dstROM;

			db = (*((DWORD*)&ROM[B_OPT00]))&0x00FFFFFF;
			pS = &ROM[db];
			dn = (*((DWORD*)&ROM[B_OPT04]))&0x00FFFFFF;
			dstROM = dn>=070000 && dn<=076000;	// set dstROM TRUE if the DST is a ROM/RAM buffer
			if( dn<0100000 && !dstROM ) goto sderr;	// fail if writing to ROM and not in ROM/RAM window
			pD = &ROM[dn];
			dw = (*((DWORD*)&ROM[B_OPT12]))&0x00FFFFFF;
			if( ROM[B_OPT03] ) {	// SRC is reversed
				if( ROM[B_OPT07] ) {	// DST is reversed
					if( dstROM ) while( dw-- ) WRAMB(--dn, *--pS);
					else while( dw-- ) *--pD = *--pS;
				} else {				// DST is normal
					if( dstROM ) while( dw-- ) WRAMB(dn++, *--pS);
					else while( dw-- ) *pD++ = *--pS;
				}
			} else {				// SRC is normal
				if( ROM[B_OPT07] ) {	// DST is reversed
					if( dstROM ) while( dw-- ) WRAMB(--dn, *pS++);
					else while( dw-- ) *--pD = *pS++;
				} else {				// DST is normal
					if( dstROM ) while( dw-- ) WRAMB(dn++, *pS++);
					else while( dw-- ) *pD++ = *pS++;
				}
			}
		}
	 	break;
	 case U_SETLED:
	 	/// BOPT00 = 1,2,3 (whether to set #1, #2 or BOTH LEDs)
	 	/// BOPT01 = RED value
	 	/// BOPT02 = GREEN value
	 	/// BOPT03 = BLUE value
		dw = (((DWORD)ROM[B_OPT03])<<16) | (((DWORD)ROM[B_OPT02])<<8) | ((DWORD)ROM[B_OPT01]);
		i = ROM[B_OPT00];
		if( i & 1 ) LED1color = dw;
		if( i & 2 ) LED2color = dw;
	 	DrawLEDs();
		break;
	 case U_SDCOPY:
		/// buf = src file name
		/// buf+256 = dst file name
		/// BOPT00 = 0 for no overwrite, 1 or 2 for overwrite
		if( ResolvePath((char*)(&ROM[buf+256]), TRUE) ) goto fperr;	// build full unique path in TMPpath
		strcpy((char*)ScratBuf, TMPpath);
		if( ResolvePath((char*)(&ROM[buf]), TRUE) ) goto fperr;	// build full unique path in TMPpath
		if( !CopyFileA(TMPpath, (char*)ScratBuf, !ROM[B_OPT00]) ) goto sderr;

		WRAMW(buse, 0);	// OK
		break;
	 case U_SDEOF:
	 	/// ENTRY: BOPT00 = file# of open SD file
	 	/// EXIT: BOPT04-07 = DWORD # of bytes from current position to EOF

	 	// BOPT00 = file#
	 	// BOPT01 = mode (0=absolute position, 1=advance from current position, 2=go to end of file)
	 	// BOPT04-07 = offset (signed value!!!)
	 	s = ROM[B_OPT00]-1;	// file# (1 to NUMAUXFILES)->(0 to NUMAUXFILES-1)
	 	if( SDfiles[s].handle==-1 ) goto sderr;
	 	dn = (DWORD)_lseek(SDfiles[s].handle, 0, SEEK_CUR);	// save where we are now
	 	dd = (DWORD)_lseek(SDfiles[s].handle, 0, SEEK_END);	// get end of file
		_lseek(SDfiles[s].handle, dn, SEEK_SET);			// return to the 'current' location

		WRAMD(B_OPT04, dd-dn);
		break;
	 case U_EXISTS:
	 	/// ENTRY: buf = SD filename (no wildcards)
	 	/// EXIT: BOPT00 = 0 if doesn't exist, = 1 if does exist
	 	if( ffUTIL!=-1 ) {
			_findclose(ffUTIL);
			ffUTIL = -1;
	 	}
	 	// AUXROM puts a NUL at the end so we can use strcpy() & such, if needed
	 	if( strlen((char*)(&ROM[buf])) > 1023 ) goto fperr;
		if( ResolvePath((char*)(&ROM[buf]), TRUE) ) goto fperr;	// build full unique path in TMPpath
		i = strlen(TMPpath);
		// now check if the resulting path is a valid path
		if( i>1 ) {	// if not root
			if( ISSLASH(TMPpath[i-1]) ) TMPpath[i-1] = 0;	// get rid of trailing SLASH, but not if it's also the leading SLASH
			ffUTIL = _findfirst(TMPpath, &UTILdat);
			if( ffUTIL==-1 ) i = 0;
			else {
				i = 1;
				_findclose(ffUTIL);
				ffUTIL = -1;
			}
		}
		WRAMB(B_OPT00, i);
		break;
	 case U_EBTKSREV:
	 	/// return revision string of your choice in buf
	 	i = strlen(szAppName);
		WRAMBUF(buf, (BYTE*)szAppName, i);
		WRAMB(blen, i);
		break;
	 case U_RMIDLE:
	 	/// alert EBTKS that we're idle in RMIDLE (R17==0, R16==0 or 4)
	 	/// A.BOPT00=R16 at entry
	 	/// Returns A.BUSEx:
	 	///		0 = do nothing
	 	///		1 = A.BOPT00 = keycode to 'type' (place in KEYHIT)
	 	///		2 = take over KYIDLE and send all keys except RESET to EBTKS via U_KEY
	 	///		3 = release KYIDLE
	 	/// NOTE: The Emulator ONLY returns 0 or 1, it doesn't support 2 or 3.
	 	///
	 	if( pRMIDLEkeys && *pRMIDLEkeys ) {
			j = *pRMIDLEkeys++;	// get next char/keycode (most normal keys)
			if( j=='\\' ) {
				if( *pRMIDLEkeys=='\\' ) ++pRMIDLEkeys;	// handle "\\" for a single '\'
				else for(i=j=0; i<3 && *pRMIDLEkeys; i++) j = (j<<3) + (*pRMIDLEkeys++ - '0');	// handle \ooo for a 3-octal-digit keycode
			} else if( j==10 || (j==13 && *pRMIDLEkeys==10 )) {
				if( j==13 ) ++pRMIDLEkeys;
				j = 0232;
			}
			ROM[B_OPT00] = j;	// return the keycode
			WRAMB(buse, 1);		// signal we're returning a keycode to AUXROM
	 	}
/* GEEKDO: TEST CODE, REMOVE THIS
if( TestKeyCnt==5 ) WRAMB(buse, 2);	// take keyboard
else if( TestKeyCnt>0 ) WRAMB(buse, 0);	// do nothing
else if( TestKeyCnt==0 ) {
	--TestKeyCnt;
	WRAMB(buse, 3);
} else WRAMB(buse, 0);
*/
		break;
	 case U_BATCH:
	 	/// ENTRY: buf = SD filename (no wildcards)
	 	/// EXIT: B_USE=error code (if missing file) or 0
		if( ResolvePath((char*)(&ROM[buf]), TRUE) ) goto fperr;	// build full unique path in TMPpath
	 	if( -1==(i = _open((char*)TMPpath, _O_RDONLY | _O_BINARY, _S_IREAD | _S_IWRITE)) ) goto fperr;
		j = _lseek(i, 0, SEEK_END);	// see how long the file is
		_lseek(i, 0, SEEK_SET);		// move back to start
		pRMIDLEkeys = (BYTE*)malloc(j+1);	// alloc buffer, +1 for NUL at end
		if( pRMIDLEkeys!=NULL ) {
			_read(i, pRMIDLEkeys, j);// read file into buffer
			pRMIDLEkeys[j] = 0;		// put NUL on end
		}
		_close(i);
		break;
	 case U_BOOT:
	 	/// 'reboot' Series 80 and EBTKS
		ExecuteMenu(MENU_RESET);
		break;
	 case U_KEY:
	 	/// Sent when U_RMIDLE has returned 2 (take over keyboard) and a key has been pressed
	 	/// A.BOPT00 = KEYCOD value
	 	/// A.BOPT01 = KEYSTS value:
	 	///	BIT	0 KBD interrupts enabled
	 	///		1 a key is down
	 	///		2
	 	///		3 SHIFT is down
	 	///		4
	 	///		5
	 	///		6
	 	///		7 GLOBAL interrupts are enabled

		/// EBTKS can do whatever it wants. The Emulator never enters this mode, so does nothing.
		/// A.BUSEx return value MUST be 0.
/* GEEKDO: TEST CODE, REMOVE THIS!!!
--TestKeyCnt;
sprintf((char*)ScratBuf, "%3.3o %3.3o", ROM[B_OPT00], ROM[B_OPT01]);
SystemMessage((char*)ScratBuf);
*/

		break;

/*
	 case :
		break;
*/
	 default:
		sprintf((char*)ScratBuf, "Bad AUXROM cmd %d", (int)(*((WORD*)&ROM[buse])));
		SystemMessage((char*)ScratBuf);
		break;
	}
	// release the ownership of the MAILBOX
	WRAMB(B_MB0+val, 0);
}

//*********************************
/*
void HelpKey(DWORD key) {

	char	*pB;
	int		i, j;
	/// now restore the screen
	pB = SaveScreen;
	j = IO_CRTSAD;
	for(i=0; i<16*32; i++) {	// write screen
		CRT[j/2] = *pB++;
		j = (j+2) & 0x0FFF;	// bump CRT address and keep within ALPHA
	}
	HP85InvalidateCRT();

	sendHelpKey = FALSE;	// release keyboard to the main emulator

	WRAMB(B_MB6, 0);		// release MB6
}
*/
/// The following code is from Philip Freidin. Thanks, Philip!
//
// Process a Real from the R12 stack. 12 digit sign magnitude mantissa, the exponent is 3 digits
// in 10's complement format exponent is +499 to -499.
//
// IEEE 754 Double precision starts to run out of steam at 10E308 and 10E-308 . Some de-normalized
// numbers extend the range out to about E320 and E-320 but you are no longer playing with 12 digits precision
// Decided to just clip it at E+307/E-307
//

double cvt_HP85_real_to_IEEE_double(uint8_t number[])
{
	int		exponent;	// 3 digits, 10's complement
	double	result;
	BOOL	negative_mantisa;
	char	numtext[25];

	if( number[4] == 0377 ) {	//  We have a Tagged Integer
		return (double)cvt_R12_int_to_int32(number);
	}
	//
	//  Handle a 12 digit Real, with an exponent between -499 and +499
	//  IEEE Double only supports 10E308 and 10E-308 , so we can't represent all numbers.
	//  Do our best
	//
	exponent =  (((number[0] & 0x0F)     ) * 1.0);
	exponent += (((number[0] & 0xF0) >> 4) * 10.0);
	exponent += (((number[1] & 0xF0) >> 4) * 100.0);

  //  Serial.printf("10's com exp is %5d ", exponent);      //  Diagnostic

  if(exponent > 499)
  {
    exponent = exponent - 1000;
  }

  //
  //  Limit the exponent, even though this makes the conversion less accurate
  //

  if(exponent > 307) exponent = 307;
  if(exponent < -307) exponent = -307;

  //  Serial.printf("limit exp is %5d ", exponent);      //  Diagnostic

  //
  //  Now deal with the mantissa, which is Sign Magnitude, not 10's complement
  //  Not the prettiest solution, but easy to understand. Convert the BCD nibbles
  //  to text and then let sscanf() sort things out
  //

  negative_mantisa = ((number[1] & 0x0F) == 9);
  numtext[ 0] = negative_mantisa ? '-' : ' ';

  numtext[ 1] = ((number[7] & 0xF0) >> 4) + '0';
  numtext[ 2] = '.';
  numtext[ 3] = ((number[7] & 0x0F)     ) + '0';

  numtext[ 4] = ((number[6] & 0xF0) >> 4) + '0';
  numtext[ 5] = ((number[6] & 0x0F)     ) + '0';

  numtext[ 6] = ((number[5] & 0xF0) >> 4) + '0';
  numtext[ 7] = ((number[5] & 0x0F)     ) + '0';

  numtext[ 8] = ((number[4] & 0xF0) >> 4) + '0';
  numtext[ 9] = ((number[4] & 0x0F)     ) + '0';

  numtext[10] = ((number[3] & 0xF0) >> 4) + '0';
  numtext[11] = ((number[3] & 0x0F)     ) + '0';

  numtext[12] = ((number[2] & 0xF0) >> 4) + '0';
  numtext[13] = ((number[2] & 0x0F)     ) + '0';

  snprintf(&numtext[14], 6, "E%04d", exponent);    //  fills 14..19, including a trailing 0x00
  //  Serial.printf("numtext is [%s] ", numtext);      //  Diagnostic
  sscanf(numtext , "%lf", &result);
  return result;
}

//
//  Process a Tagged Integer from the R12 stack -99999 to 99999 in 10's complement format
//

int32_t cvt_R12_int_to_int32(uint8_t number[])
{
  uint32_t    result;

  result =  (((number[5] & 0x0F)     ) * 1);
  result += (((number[5] & 0xF0) >> 4) * 10);

  result += (((number[6] & 0x0F)     ) * 100);
  result += (((number[6] & 0xF0) >> 4) * 1000);

  result += (((number[7] & 0x0F)     ) * 10000);
  result += (((number[7] & 0xF0) >> 4) * 100000);   //  0 for positive, 9 for 10's complement negative

  if(result > 99999)
  {
    result = result - 1000000;  //  Minus one million because the sign indicator is a 9, so neg numbers will be 9xxxxx
  }
  return result;
}

#if 0
//
//  This function converts a 64 bit IEEE Double to a HP Real.  While some results could be represented
//  with a Tagged Integer, I don't think that it has to do that. Confirmed, no problem returning integers
//  between -99999 and +99999 as 8 byte Reals (non tagged), HP85 is fine with the results.
//
//  Dest points to an 8 byte area that can hold an 8 byte real
//
//  The conversion was helped by a code example here:
//    https://stackoverflow.com/questions/31331723/how-to-control-the-number-of-exponent-digits-after-e-in-c-printf-e
//
//  Tested with all the above values, and they all converted correctly
//

//                      1 . yyyyyyyyyyy e 0 EEE \0
//                    - 1 . xxxxxxxxxxx e - EEE \0
#define ExpectedSize (1+1+1       +11  +1+1+ 3 + 1)

void cvt_IEEE_double_to_HP85_number(uint8_t * dest, double val)
{
  bool  negative;
  char  buf[ExpectedSize + 10];

  if((negative = val < 0))
  {
    val = -val;       //  We only want to format positive numbers
  }

  snprintf(buf, sizeof buf, "%.11e", val);
  char *e = strchr(buf, 'e');                           // lucky 'e' not in "Infinity" nor "NaN"
  if (e)
  {
    e++;
    int expo = atoi(e);
    //
    //    But nothing is ever easy. The HP85 DECIMAL Exponent is in 10's complement format
    //
    if(expo < 0)
    {
      expo = 1000 + expo;
    }
    snprintf(e, sizeof buf - (e - buf), "%04d", expo);
  }

//
//  The number is now formatted in buf. See page 3-11 of the HP85 Assembler manual for the Nibble codes
//                      Nibble
//    buf[0 ]   1         M0
//    buf[1 ]   .
//    buf[2 ]   2         M1
//    buf[3 ]   3         M2
//    buf[4 ]   4         M3
//    buf[5 ]   5         M4
//    buf[6 ]   6         M5
//    buf[7 ]   7         M6
//    buf[8 ]   8         M7
//    buf[9 ]   9         M8
//    buf[10]   10        M9
//    buf[11]   11        M10
//    buf[12]   12        M11
//    buf[13]   E
//    buf[14]   - or 0    MS  Mantissa Sign due to above code, always 0 in buf[14]
//    buf[15]   1         E0  most sig digit of exponent, 10's complement format
//    buf[16]   2         E1
//    buf[17]   3         E2  least sig digit of exponent
//    buf[18]
//    buf[19]
//    buf[20]
//
  *dest++ = ((buf[16] & 0x0F) << 4) | (buf[17]        & 0x0F);     //  E1  E2
  *dest++ = ((buf[15] & 0x0F) << 4) | (negative? 0x09 : 0x00);     //  E0  MS
  *dest++ = ((buf[11] & 0x0F) << 4) | (buf[12]        & 0x0F);     //  M10 M11
  *dest++ = ((buf[9]  & 0x0F) << 4) | (buf[10]        & 0x0F);     //  M8  M9
  *dest++ = ((buf[7]  & 0x0F) << 4) | (buf[8]         & 0x0F);     //  M6  M7
  *dest++ = ((buf[5]  & 0x0F) << 4) | (buf[6]         & 0x0F);     //  M4  M5
  *dest++ = ((buf[3]  & 0x0F) << 4) | (buf[4]         & 0x0F);     //  M2  M3
  *dest++ = ((buf[0]  & 0x0F) << 4) | (buf[2]         & 0x0F);     //  M0  M1
}
#endif

#endif // NOAUXROM
