

// need to do something special to handle HOMECURS on HP87


/*******************************************
 SCRIPT.C - command-line-specified script-control of the emulator
	Copyright 2020 Everett Kaser
	All rights reserved.
 See HP85EM.TXT for legal usage.
 *******************************************/

#include "main.h"
#include "mach85.h"
#include "script.h"

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <ctype.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>

/// Script window housekeeping variables
long	scriptWndX, scriptWndY, scriptWndW, scriptWndH, scriptWndMisc;	// for intialization ONLY!  NOT updated! (read at iR_READ_INI, use at iR_STARTUP)
long	scriptWnd;

OVERRIDES	Override85;
int			ORline;
BYTE		*ORbuf=NULL, *ORptr;
BOOL		KeepIni=FALSE, DoExit=FALSE;

BYTE		*RSkeybuf=NULL;		// buffer for storing keys pressed while Recording Script
int			RSkeycnt;			// # of keys currently in buffer
BOOL		RSkeyNOW;			// whether keys in RSkeybuf are during IDLE (FALSE) or RUN/INPUT (TRUE)
int			RSfile;				// file handle for RecordScript output file

long		ORwait=WAITUNTIL_DONT;
BOOL		ORclick;			// whether CLICK on script has happened or not during WAITUNTIL_CLICK
DWORD		ORseconds, ORlastTime;
BYTE		*pKeysNow=NULL, *pNowPtr;
BOOL		inKEYSNOW;
BYTE		Sname1[PATHLEN], Sname2[PATHLEN], *PSargs=NULL, *pArgs[10];

int			WaitKey, WaitSC, WaitCA, WaitUNIT;

typedef struct {
	char	*KeyName;
	BYTE	Code85, Code87;
} KEYNAMCOD;

/// NOTE: Any SHORTER names that identically match the beginning of a LONGER name
/// *must* come AFTER the longer name, so that the longer name will be searched
/// and found (if matches) BEFORE the shorter name is searched and found. The
/// only current example of this is K1 and K10-K14.
KEYNAMCOD	Keys[]={
	{"K14",-1,147},
	{"K13",-1,172},
	{"K12",-1,165},
	{"K11",-1,207},
	{"K10",-1,206},
	{"K9",-1,205},
	{"K8",135,204},

	{"K7",134,156},
	{"K6",133,162},
	{"K5",132,161},
	{"K4",131,131},
	{"K3",130,130},
	{"K2",129,129},
	{"K1",128,128},

	{"REW",136,-1},
	{"COPY",137,-1},
	{"PAPADV",138,-1},
	{"RESET",139,139},
	{"INIT",140,140},
	{"RUN",141,141},
	{"PAUSE",142,142},

	{"CONT",143,143},
	{"STEP",144,144},
	{"TEST",145,146},
	{"-SCREEN",146,137},
	{"-LINE",160,157},
	{"-CHR",164,136},
	{"KEYLABEL",150,150},

	{"GRAPH",147,-1},
	{"A/G",-1,168},
	{"LIST",148,148},
	{"PLIST",149,149},
	{"BKSPACE",153,153},
	{"FASTBKSPACE",155,155},
	{"HOMECURS",165,163},

	{"LFCURS",156,159},
	{"RTCURS",157,170},
	{"UPCURS",161,163},
	{"DNCURS",162,164},
	{"ENDLINE",154,154},
	{"ROLLUP",158,145},
	{"ROLLDN",159,169},

	{"I/R",163,158},
	{"RESULT",166,166},
	{"DELETE",168,-1},
	{"STORE",169,-1},
	{"LOAD",170,-1},
	{"AUTO",172,-1},
	{"SCRATCH",173,-1},

	{"TR/NORM",-1,173},
	{"NUME",-1,160},
	{"S-NUMDIV",123,123},
	{"S-NUMMIN",125,125},
	{"S-NUMMUL",126,126},
	{"S-NUMPLUS",127,127},
};
int	KeysCnt=sizeof(Keys)/sizeof(Keys[0]);

///*******************************************************
/// This is the same as Is85Idle() in mach85.c, except that it doesn't check Int85 or IsKeyPressed()
BOOL RecordIs85Idle() {

	WORD	RMidle, Exec, EndExec, w;

	if( Halt85==FALSE ) {
		Exec = 072;
		EndExec = 0117;
		if( Model<2 ) RMidle = 0102434;
		else RMidle = 0103706;
		w = *((WORD *)(Reg+4));
		if( Reg[016]==0 && ((w>=RMidle && w<RMidle+7) || (w>=Exec && w<EndExec)) ) return TRUE;
	}
	return FALSE;
}

///************************************
void RecordFlushKeybuf() {

	int		i, ib, ob;
	char	*pC;

	if( RSkeycnt ) {
		if( RSkeyNOW==0 ) strcpy((char*)ScratBuf, "KEYSNOW");
		else strcpy((char*)ScratBuf, "KEYSIDLE");
		ib = 0;
		ob = strlen((char*)ScratBuf);
		while( ib<RSkeycnt ) {
			if( RSkeybuf[ib]<123 ) {	// ASCII chars
				ScratBuf[ob++] = ' ';
				ScratBuf[ob++] = '"';	// opening quote
				for(; ib<RSkeycnt && RSkeybuf[ib]<123; ib++) {
					if( RSkeybuf[ib]=='"' || RSkeybuf[ib]=='\\' ) ScratBuf[ob++] = '\\';
					ScratBuf[ob++] = RSkeybuf[ib];
				}
				ScratBuf[ob++] = '"';	// closing quote
			} else {					// NON-ASCII key
				ScratBuf[ob++] = ' ';
				for(i=0; i<KeysCnt; i++) if( RSkeybuf[ib]==((Model<2) ? Keys[i].Code85 : Keys[i].Code87) ) break;
				for(pC=Keys[i].KeyName; *pC; ) ScratBuf[ob++] = *pC++;
				++ib;
			}
		}
		ob += GetSystemEOL(ScratBuf+ob);
		write(RSfile, ScratBuf, ob);
		RSkeycnt = 0;
	}
}

///************************************
void RecordScriptKey(BYTE k) {

	int		i;
	BOOL	RSidle;

	if( RSkeybuf!=NULL && !InTypeFile ) {
		RSidle = RecordIs85Idle();
		if( !RSkeycnt ) RSkeyNOW = RSidle;	// if just starting a string of keys, set whether KEYSIDLE or KEYSNOW
		else if( RSkeyNOW!=RSidle ) {		// else if IDLE/RUN state has changed
			RecordFlushKeybuf();			// flush the buffer
			RSkeyNOW = RSidle;				// and reset which state we're recording
		}
		if( k<123 ) RSkeybuf[RSkeycnt++] = k;	// ASCII key
		else {								// non-ASCII key
			for(i=0; i<KeysCnt; i++) if( k==((Model<2) ? Keys[i].Code85 : Keys[i].Code87) ) break;
			if( i< KeysCnt ) RSkeybuf[RSkeycnt++] = k;
			else SystemMessage("RECORD SCRIPT: unknown/unexpected key code");
		}
		if( RSkeycnt > 159 ) RecordFlushKeybuf();
	}
}

///*************************************
void RecordBoot() {

	BYTE	eolbuf[16];
	int		i;

	if( RSkeybuf!=NULL ) {
		RecordFlushKeybuf();
		GetSystemEOL(eolbuf);
		sprintf((char*)ScratBuf,"BOOT %s {%s", ((Model==0) ? "85A" : ((Model==1) ? "85B" : "87")), eolbuf);
		write(RSfile, ScratBuf, strlen((char*)ScratBuf));

		strcpy((char*)ScratBuf, "\tROMS");
		for(i=0; i<RomListCnt; i++) {
			if( (RomLoad & (1<<i)) && !(RomMUST[Model] & (1<<i)) && !(RomCANT[Model] & (1<<i)) ) {
				sprintf((char*)ScratBuf+strlen((char*)ScratBuf), " %3.3o", (unsigned int)RomList[i]);
			}
		}
		strcat((char*)ScratBuf, (char*)eolbuf);
		write(RSfile, ScratBuf, strlen((char*)ScratBuf));

		sprintf((char*)ScratBuf, "\tRAM %d %d%s", (int)RamSize, (RamSizeExt[Model] ? (16<<RamSizeExt[Model]) : 0), eolbuf);
		write(RSfile, ScratBuf, strlen((char*)ScratBuf));

		sprintf((char*)ScratBuf, "\tDISPLAYx2 %d%s", (CfgFlags & CFG_BIGCRT) ? 1 : 0, eolbuf);
		write(RSfile, ScratBuf, strlen((char*)ScratBuf));

		sprintf((char*)ScratBuf, "\tSLOWSPEED %d%s", (CfgFlags & CFG_NATURAL_SPEED) ? 1 : 0, eolbuf);
		write(RSfile, ScratBuf, strlen((char*)ScratBuf));

		sprintf((char*)ScratBuf, "}%s", eolbuf);
		write(RSfile, ScratBuf, strlen((char*)ScratBuf));
	}
}

///**************************************************************
void RecordFile(int action, char* name, char* name2) {

	BYTE	eolbuf[16];

	if( RSkeybuf!=NULL ) {
		RecordFlushKeybuf();
		GetSystemEOL(eolbuf);

		switch( action ) {
		 case RECFILE_TYPE:
			sprintf((char*)ScratBuf, "TYPEFILE \"%s\"", name);
			break;
		 case RECFILE_IMPORT:
			sprintf((char*)ScratBuf, "IMPORTFILE \"%s\" \"%s\"", name, name2);
			break;
		 case RECFILE_EXPORT:
			sprintf((char*)ScratBuf, "EXPORTFILE \"%s\" \"%s\"", name, name2);
			break;
		}
		strcat((char*)ScratBuf, (char*)eolbuf);
		write(RSfile, ScratBuf, strlen((char*)ScratBuf));
	}
}

///**************************************************************
void RecordCatAll(char *fname) {

	BYTE	eolbuf[16];

	if( RSkeybuf!=NULL ) {
		RecordFlushKeybuf();
		GetSystemEOL(eolbuf);

		sprintf((char*)ScratBuf, "CATALL \"%s\"%s", fname, (char*)eolbuf);
		write(RSfile, ScratBuf, strlen((char*)ScratBuf));
	}
}

///**************************************************************
void RecordDisk(int action, int msus, int folder, char* name, char* name2) {

	BYTE	eolbuf[16];

	if( RSkeybuf!=NULL ) {
		RecordFlushKeybuf();
		GetSystemEOL(eolbuf);

		switch( action ) {
		 case RECDISK_NONE:
			sprintf((char*)ScratBuf, "DISK %d NONE", msus);
			break;
		 case RECDISK_NEW:
			sprintf((char*)ScratBuf, "DISK %d NEW %d \"%s\"", msus, folder, name);
			break;
		 case RECDISK_LOAD:
			sprintf((char*)ScratBuf, "DISK %d LOAD %d \"%s\"", msus, folder, name);
			break;
		 case RECDISK_DELETE:
			sprintf((char*)ScratBuf, "DISK %d DELETE %d \"%s\"", msus, folder, name);
			break;
		 case RECDISK_RENAME:
			sprintf((char*)ScratBuf, "DISK %d RENAME %d \"%s\" \"%s\"", msus, folder, name, name2);
			break;
		}
		strcat((char*)ScratBuf, (char*)eolbuf);
		write(RSfile, ScratBuf, strlen((char*)ScratBuf));
	}
}

///****************************************
void RecordOptions(DWORD oldflags) {

	DWORD	f;
	BYTE	eolbuf[16];


	f = (oldflags ^ CfgFlags) & (CFG_BIGCRT | CFG_NATURAL_SPEED);
	if( f ) {
		RecordFlushKeybuf();
		GetSystemEOL(eolbuf);

		if( f & CFG_BIGCRT ) {
			sprintf((char*)ScratBuf, "DISPLAYx2 %d", (CfgFlags & CFG_BIGCRT) ? 1 : 0);
			strcat((char*)ScratBuf, (char*)eolbuf);
			write(RSfile, ScratBuf, strlen((char*)ScratBuf));
		}
		if( f & CFG_NATURAL_SPEED ) {
			sprintf((char*)ScratBuf, "SLOWSPEED %d", (CfgFlags & CFG_NATURAL_SPEED) ? 1 : 0);
			strcat((char*)ScratBuf, (char*)eolbuf);
			write(RSfile, ScratBuf, strlen((char*)ScratBuf));
		}
	}
}

///*************************************
void RecordScriptStart(char *pCmd) {

	int		i;

	memset((void*)&Override85, 0xFF, sizeof(Override85));
	RSkeybuf = (BYTE *)malloc(256);			// allocate RecordScript key buffer (which, not being NULL, also indicates we're recording a script)
	RSkeycnt = 0;
	if( *pCmd=='"' ) {	// quoted name
		++pCmd;
		for(i=0; *pCmd && *pCmd!='"'; i++) Sname1[i] = *pCmd++;
		if( *pCmd=='"' ) ++pCmd;
	} else {	// unquoted name
		for(i=0; *pCmd && *pCmd!=' '; i++) Sname1[i] = *pCmd++;
	}
	Sname1[i] = 0;
	if( -1==(RSfile=open((char*)Sname1, _O_BINARY | _O_RDWR | _O_CREAT | _O_TRUNC, _S_IREAD | _S_IWRITE)) ) {
		SystemMessage("Error creating SCRIPT output file");
		free(RSkeybuf);
		RSkeybuf = NULL;
	}
}
///*************************************
void RecordScriptExit() {

	if( RSkeybuf!=NULL ) {
		RecordFlushKeybuf();	// in case we're recording a script and there's buffered keys
		strcpy((char*)ScratBuf, "END EXIT");
		GetSystemEOL(ScratBuf+strlen((char*)ScratBuf));
		write(RSfile, ScratBuf, strlen((char*)ScratBuf));
		close(RSfile);
		free(RSkeybuf);
		RSkeybuf = NULL;
	}
}

///******************************************************
BYTE *POskipWhite(BYTE *p) {

	while( *p=='\r' || *p=='\n' || *p==' ' || *p=='\t' || (*p=='/' && (*(p+1)=='*' || *(p+1)=='/')) ) {
		if( *p=='/' ) {	// comment
			++p;
			if( *p++=='*' ) {	// skip /* comment
				while( *p && (*p!='*' || *(p+1)!='/') ) {
					if( *p=='\n' ) ++ORline;
					++p;
				}
				++p;	// skip the '*', the '/' will get skipped below
			} else {	// skip // comment
				while( *p && *p!='\r' && *p!='\n' ) {	// skip everything to EOL
					++p;
				}
			}
		}
		if( *p=='\n' ) ++ORline;
		++p;
	}
	return p;
}

///******************************************************
int POgetNum() {

	int	retval;

	ORptr = POskipWhite(ORptr);
	if( !isdigit(*ORptr) ) return -1;
	retval = atoi((const char*)ORptr);
	while( isdigit(*ORptr) ) ++ORptr;
	ORptr = POskipWhite(ORptr);
	return retval;
}

///******************************************************
BOOL POgetName(BYTE *pName) {

	ORptr = POskipWhite(ORptr);
	if( *ORptr!='"' ) return FALSE;
	++ORptr;
	while( *ORptr && *ORptr!='"' && *ORptr!='\r' && *ORptr!='\n' ) *pName++ = *ORptr++;
	if( *ORptr!='"' ) return FALSE;
	++ORptr;
	*pName = 0;	// nul-term filename
	return TRUE;
}

///******************************************************
void GetMSUS(BYTE *pName, int *pSC, int *pCA, int *pUNIT) {

	int		i, j;

	// scan for MSUS (move 'i' to ':' or end-of-string)
	for(i=0; pName[i] && pName[i]!=':'; i++) ;
	if( pName[i]==':' ) {	// if MSUS there
		j = i+1;
		if( pName[j] ) ++j; // skip 'D'
		if( pName[j]=='1' && pName[j+1]=='0' ) {	// special-case SC=10
			*pSC = 7;
			j += 2;
		} else if( pName[j]>='3' && pName[j]<='9' ) {	// handle 300-900
			*pSC = pName[j]-'3';
			++j;
		} else *pSC = -1;
		if( *pSC>=0 ) {
			if( pName[j]>='0' && pName[j]<='7' ) {
				*pCA = pName[j]-'0';
				++j;
				if( pName[j]>='0' && pName[j]<='3' ) *pUNIT = pName[j]-'0';
				else *pUNIT = 0;
			} else {
				*pCA = 0;
				*pUNIT = 0;
			}
		} else {	// default to :D700
			*pSC = 4;
			*pCA = 0;
			*pUNIT = 0;
		}
		pName[i] = 0;	// erase the MSUS from the file name
	} else {
		*pSC = MenuDiskSC;
		*pCA = MenuDiskTLA;
		*pUNIT = MenuDiskDrive;
	}
}

///******************************************************
BOOL POinDiffDrive(BYTE *pFname, int sc, int ca, int unit) {

	int		sci, tli, di;
	char	*pName;

	for(sci=0; sci<8; sci++) {
		for(tli=0; tli<8; tli++) {
			for(di=0; di<2; di++) {
				pName = GetDiskName(sci, tli, di);
				if( pName!=NULL && *pName!=0 && (sci!=sc || tli!=ca || di!=unit) ) {	// if disk in drive and not THIS drive
					if( !strcmp(pName, (char*)pFname) ) {	// if disk is in another drive, don't allow it in THIS drive
						sprintf((char *)ScratBuf, "SCRIPT DISK is in another drive) on line %d", ORline);
						SystemMessage((char *)ScratBuf);
						return TRUE;
					}
				}
			}
		}
	}
	return FALSE;
}

///******************************************************
int ORparseKeys(BYTE *pS, BYTE *pD) {

	int		outcnt=0, kc, i;
	char	kerrbuf[16];

	while( *pS==' ' || *pS=='\t' ) ++pS;	// skip non-EOL whitespace
	// process line
	while( *pS && *pS!='\r' && *pS!='\n' ) {
		if( *pS=='"' ) {			// start of quoted ASCII string
			++pS;				// skip opening quote
			while( *pS && *pS!='"' && *pS!='\r' && *pS!='\n' ) {
				if( *pS=='\\' && *(pS+1)!='\r' && *(pS+1)!='\n' ) ++pS;
				if( pD!=NULL ) *pD++ = *pS;
				++pS;
				++outcnt;
			}
			if( *pS=='"' ) ++pS;// skip closing quote
		} else if( *pS=='0' ) {		// octal keycode
			kc = 0;
			while( *pS && *pS>='0' && *pS<='7' ) kc = (kc<<3) | ((*pS++)-'0');
octdeckc:
			if( kc>255 ) {
				sprintf((char *)ScratBuf, "SCRIPT invalid keycode %d on line %d", kc, ORline);
				SystemMessage((char *)ScratBuf);
				return 0;
			}
			if( pD!=NULL ) *pD++ = kc;
			++outcnt;
		} else if( isdigit(*pS) ) {	// decimal keycode
			kc = 0;
			while( *pS && *pS>='0' && *pS<='9' ) kc = kc*10 + ((*pS++)-'0');
			goto octdeckc;
		} else {					// key NAME
			for(i=0; i<KeysCnt; i++) {
				kc = strlen(Keys[i].KeyName);
				if( !strnicmp((const char *)Keys[i].KeyName, (const char *)pS, kc) ) {
					pS += kc;
					if( Keys[i].Code85==154 ) kc = '\n';	// ENDLINE
					else kc = (Model==2) ? Keys[i].Code87 : Keys[i].Code85;
					if( pD!=NULL ) *pD++ = kc;
					++outcnt;
					break;
				}
			}
			if( i>=KeysCnt ) {	// didn't find a matching key name
				for(i=0; i<15 && *pS && *pS!=' ' && *pS!='\t' && *pS!='\r' && *pS!='\n'; i++) kerrbuf[i] = *pS++;
				kerrbuf[i] = 0;
				sprintf((char *)ScratBuf, "SCRIPT invalid key name %s on line %d", kerrbuf, ORline);
				SystemMessage((char *)ScratBuf);
				return 0;
			}
		}
		while( *pS==' ' || *pS=='\t' ) ++pS;	// skip non-EOL whitespace
	}
	return outcnt;
}

/*
			 case VK_HOME:
			 	if( Model<2 ) k = 0245;
			 	else {
					k = 0243;
					SetHomeKey(TRUE);	// fake SHIFT key down
			 	}
				break;
			}
			SetKeyDown(k);
*/

///***********************************************************
/// Figures out where everything in the SCRIPT window should be drawn.
void RelocateScript() {

}

///***********************************************************
/// assumes that RelocateDisks() has been called previously by KWnd[] create/resize code, thus the pD->cat XYWH variables have already been set.
void GetScriptListRect(RECT *r) {

	r->top = 0;
	r->bottom = KWnds[scriptWnd].vh;
	r->left = 0;
	r->right = KWnds[scriptWnd].vw;
}

///**************************************************************
void ScriptStartup(char *pCmd) {

	int		OR85, i, j, k, tcnt;
	char	*p, *s, c;
	BYTE	*pA, *pB, *ORbufA, *pLine;

	memset((void*)&Override85, 0xFF, sizeof(Override85));
	if( *pCmd ) {	// if "script" file provided
		// get SCRIPT NAME
		if( *pCmd=='"' ) {	// quoted name
			++pCmd;
			for(i=0; *pCmd && *pCmd!='"'; i++) Sname1[i] = *pCmd++;
			if( *pCmd=='"' ) ++pCmd;
		} else {	// unquoted name
			for(i=0; *pCmd && *pCmd!=' '; i++) Sname1[i] = *pCmd++;
		}
		Sname1[i] = 0;
		/// get SCRIPT args
		/// if there is no arg[x], pArgs[x]==NULL
		memset(pArgs, 0, sizeof(pArgs));	// all arg ptrs = NULL
		while( *pCmd==' ' ) ++pCmd;	// skip spaces after script name
		if( *pCmd ) {	// we've got an arg
			PSargs = (BYTE*)malloc(strlen(pCmd)+1);	// args can't take more than the remaining command line itself (because spaces get thrown away)
			if( PSargs!=NULL ) {
				pA = PSargs;
				for(i=0; i<10 && *pCmd; i++) {	// get no more than 10 args
					if( *pCmd ) {	// we've got an arg
						pArgs[i] = pA;
						// get SCRIPT arg
						if( *pCmd=='"' ) {	// quoted name
							++pCmd;
							while( *pCmd && *pCmd!='"' ) {
								if( *pCmd=='\\' && (*(pCmd+1)=='\\' || *(pCmd+1)=='"') ) *pA++ = *pCmd++;	// copy '\', following '\' or '"' copied on next line
								*pA++ = *pCmd++;
							}
							if( *pCmd=='"' ) ++pCmd;
						} else {	// unquoted name
							while( *pCmd && *pCmd!=' ' ) {
								if( *pCmd=='\\' && (*(pCmd+1)=='\\' || *(pCmd+1)=='"') ) *pA++ = *pCmd++;	// copy '\', following '\' or '"' copied on next line
								*pA++ = *pCmd++;
							}
						}
						*pA++ = 0;
					}
					while( *pCmd==' ' ) ++pCmd;	// skip spaces
				}
			}
		} else PSargs = NULL;

		/// read script, replacing \args and TABS
		if( -1!=(OR85=open((char*)Sname1, _O_BINARY | _O_RDONLY)) ) {	// try to open the script file for reading
			i = lseek(OR85, 0, SEEK_END);	// see how big it is
			lseek(OR85, 0, SEEK_SET);		// move back to the start of the file
			ORbuf = malloc(i+1);			// allocate a buffer for the script + NUL
			if( ORbuf!=NULL ) {				// if allocate went okay
				read(OR85, ORbuf, i);		// read the script into buffer
				close(OR85);
				ORbuf[i] = 0;				// NUL-terminate it
				ORptr = ORbuf;				// set up the script processing pointer

				/// do \arg and TAB substitution
				// first scan through counting additional length needed for all arg subs and tab expansion
				// j is the "running length"
				pLine = ORbuf;
				for(j=i+1, pA=ORbuf; *pA; ) {
					// possibly set "start of line" pointer for tab expansion
					if( pA!=ORbuf && (*(pA-1)=='\r' || *(pA-1)=='\n') ) pLine = pA;
					// handle comments (no cmd-line arg replacement within, but IS tab handling)
					if( *pA=='/' && *(pA+1)=='*' ) {	// block comment, do no cmd-line arg replacement
						pA += 2;
						while( *pA && (*pA!='*' || *(pA+1)!='/') ) {
							if( (*(pA-1)=='\r' || *(pA-1)=='\n') ) pLine = pA;
							if( *pA=='\t' ) {
								tcnt = 4 - ((pA-pLine+4) & 3);	// # of spaces needed for this tab
								j += tcnt-1;
							}
							++pA;
						}
						continue;
					}
					if( *pA=='/' && *(pA+1)=='/' ) {	// EOL comment, do no cmd-line arg replacement
						pA += 2;
						while( *pA && (*pA!='\r' && *pA!='\n') ) {
							if( *pA=='\t' ) {
								tcnt = 4 - ((pA-pLine+4) & 3);	// # of spaces needed for this tab
								j += tcnt-1;
							}
							++pA;
						}
						continue;
					}
					// no comment, handle normal script commands
					// if TAB
					if( *pA=='\t' ) {
						tcnt = 4 - ((pA-pLine+4) & 3);	// # of spaces needed for this tab
						j += tcnt-1;
						++pA;
					} else if( *pA=='\\' ) {
						++pA;
						if( *pA>='0' && *pA<='9' ) {
							k = *pA-'0';
							++pA;
							if( PSargs==NULL || pArgs[k]==NULL ) {
								sprintf((char*)ScratBuf, "SCRIPT missing command-line argument #%d. Script aborted.", k);
								SystemMessage((char*)ScratBuf);
								free(ORbuf);
								return;
							}
							j += strlen((char*)pArgs[k])-2;	// -2 for the /N that will take 2 bytes from the ARG, so we could actually subtract 1 or 2 for a 1-char or empty ("") arg
						} else if( *pA ) ++pA;
					} else ++pA;
				}
				// now allocate a (possibly) bigger buffer
				ORbufA = pB = (BYTE*)malloc(j);
				if( pB!=NULL ) {
					// now copy the old buffer to the new buffer, substituting args as we go
					pLine = pA = ORbuf;
					while( *pA ) {
						if( pA!=ORbuf && (*(pA-1)=='\r' || *(pA-1)=='\n') ) pLine = pA;
						// handle comments (no cmd-line arg replacement within, but IS tab handling)
						if( *pA=='/' && *(pA+1)=='*' ) {	// block comment, do no cmd-line arg replacement
							*pB++ = *pA++;
							*pB++ = *pA++;
							while( *pA && (*pA!='*' || *(pA+1)!='/') ) {
								if( (*(pA-1)=='\r' || *(pA-1)=='\n') ) pLine = pA;
								if( *pA=='\t' ) {
									*pA = ' ';	// for final char copy below
									tcnt = 3 - ((pA-pLine+4) & 3);	// # of spaces needed for this tab (in addition to final space copy below)
									while( tcnt-- ) *pB++ = ' ';
								}
								*pB++ = *pA++;
							}
							if( *pA ) *pB++ = *pA++;	// skip closing "*/"
							if( *pA ) *pB++ = *pA++;
							continue;
						}
						if( *pA=='/' && *(pA+1)=='/' ) {	// EOL comment, do no cmd-line arg replacement
							*pB++ = *pA++;
							*pB++ = *pA++;
							while( *pA && (*pA!='\r' && *pA!='\n') ) {
								if( *pA=='\t' ) {
									*pA = ' ';	// for final char copy below
									tcnt = 3 - ((pA-pLine+4) & 3);	// # of spaces needed for this tab (in addition to final space copy below)
									while( tcnt-- ) *pB++ = ' ';
								}
								*pB++ = *pA++;
							}
							continue;
						}
						if( *pA=='\t' ) {
							tcnt = 4 - ((pA-pLine+4) & 3);	// # of spaces needed for this tab
							while( tcnt-- ) *pB++ = ' ';	// convert TAB to SPACES
							++pA;
						} else if( *pA=='\\' && *(pA+1)>='0' && *(pA+1)<='9' ) {
							++pA;
							k = *pA++ - '0';
							for(j=0; pArgs[k][j]; j++) *pB++ = pArgs[k][j];
						} else *pB++ = *pA++;
					}
					*pB = 0;
					free(ORbuf);
					ORbuf = ORptr = ORbufA;
				}
				/// Create script WINDOW and add script lines to it
				sprintf((char*)ScratBuf+6000, "Series 80 SCRIPT: %s", Sname1);
				scriptWndX = scriptWndY = 0;
				scriptWndW = 400;
				scriptWndH = 600;
				scriptWnd = KCreateWnd(scriptWndX, scriptWndY, scriptWndW, scriptWndH, (char*)ScratBuf+6000, &ScriptWndProc, FALSE);
				KGetWindowPos(KWnds[scriptWnd].hWnd, &KWnds[scriptWnd].x, &KWnds[scriptWnd].y, &KWnds[scriptWnd].w, &KWnds[scriptWnd].h, &KWnds[scriptWnd].misc);
				// add all the script lines to the scriptList
				for(p=(char*)ORbuf; *p; ) {
					s = p;	// save start of this line
					while( *p && *p!='\r' && *p!='\n' ) ++p;	// find end of line
					c = *p;	// save next char at end-of-line
					*p = 0;	// terminate the line with a NUL
					AddScriptList(s);	// add the line to the scriptList

					// now check if it's an END EXIT KEEPINI line, and if so, set the flag
					s = (char*)POskipWhite((BYTE*)s);
					if( !strnicmp(s, "END", 3) ) {	// it's the END line
						s += 3;
						s = (char*)POskipWhite((BYTE*)s);
						if( !strnicmp(s, "EXIT", 4) ) {
							s += 4;
							s = (char*)POskipWhite((BYTE*)s);
							if( !strnicmp((const char*)s, "KEEPINI", 7) ) KeepIni = TRUE;
						}
					}

					*p = c;	// restore the char at end-of-line
					if( *p && *p=='\n' ) ++p;	// skip the EOL sequence
					else if( *p && *p=='\r' ) {
						++p;
						if( *p && *p=='\n' ) ++p;
					}
				}
				ORline = 1;					// and keep track of which line we're on for error messages
				ShowWindow(KWnds[scriptWnd].hWnd, SW_SHOW);

				ProcessScript(0);				// start processing it
			} else close(OR85);					// close the script file
		}
	}
}

///******************************************************
BOOL IsScriptWaitingOnKey(int k) {

	if( ORbuf==NULL || ORwait!=WAITUNTIL_KEY ) return FALSE;
	if( WaitKey==-1 || k==WaitKey ) ORwait = WAITUNTIL_DONT;
	return TRUE;
}

///******************************************************
// 'from' =
//		0 program startup, just process first BOOT block
//		1 HP85Idle(), process whatever you want
//		2 Exec85(), don't process:
void ProcessScript(int from) {

	int		i, j, c, sc, ca, unit, df, keycnt;
	char	poerrbuf[16], *pDN;
	DWORD	tc;
	SCRIPTETC	*e;

	/// Make sure we have a buffer and something to do
	if( ORbuf==NULL ) return;	// bail if nothing to process

	/// Handle any pending KEYSIDLE or KEYSNOW
	if( pKeysNow!=NULL ) {	// have keys to press RIGHT NOW!
		if( inKEYSNOW || (!inKEYSNOW && from==1) ) {
			if( *pNowPtr ) {
				WORD	svcwrd;
				svcwrd = (Model<2) ? 0100151 : 0100162;
				// if no key pressed and no key pending in the Series 80 machine
				if( !IsKeyPressed() && !(LoadMem(svcwrd) & 1) ) {
					if( *pNowPtr=='\n' ) *pNowPtr = 154;	// translate ENDLINE back to actual keycode
					// The following three lines SLOW DOWN the typing when the "Program Development ROM" is present
					// AND the next keycode is the same as the one before it, because the PD ROM was causing the
					// second key to be skipped (de-bounced).
					while( (RomLoad & 2) && *pNowPtr==pKeyLastKEYCOD && pKeyLastTime+100>KGetTicks() ) ;
					pKeyLastKEYCOD = *pNowPtr;
					pKeyLastTime = KGetTicks();

					SetKeyDown(*pNowPtr++);
				}
			}
			if( *pNowPtr==0 ) {
				free(pKeysNow);
				pKeysNow = NULL;
			}
		}
		return;	/// BLOCK as long as there are pending KEYSIDLE or KEYSNOW keys
	}
	if( pKeyMem!=NULL ) return;	/// BLOCK as long as there are TYPEFILE keys

	ORptr = POskipWhite(ORptr);	/// make sure we're pointing at SOMETHING (or NUL)

	if( *ORptr==0 ) {	/// end of script, free the memory and bail
		if( !DoExit ) SystemMessage("SCRIPT completed");
ORfin:
		free(ORbuf);
		ORbuf = NULL;
		if( PSargs!=NULL ) free(PSargs);
		memset((void*)&Override85, 0xFF, sizeof(Override85));
		KDestroyWnd(scriptWnd);
//		if( DoExit ) KillProgram();
		return;
	}

	/// Check if we're WAITING
	switch( ORwait ) {
	 case WAITUNTIL_DONT:
		break;
	 case WAITUNTIL_SECONDS:
		tc = KGetTicks();
		i = (abs(tc-ORlastTime))/1000;
		if( i ) {
			ORseconds -= (i>ORseconds) ? ORseconds : i;
			ORlastTime = tc;
		}
		if( !ORseconds ) ORwait = WAITUNTIL_DONT;	// set to DON'T wait
		return;
	 case WAITUNTIL_IDLE:
		if( pKeyMem==NULL && pKeysNow==NULL && Is85Idle(0) && from==1 ) ORwait = WAITUNTIL_DONT;
		return;
	 case WAITUNTIL_INPUT:
		if( Is85Idle(4) ) ORwait = WAITUNTIL_DONT;
		return;
	 case WAITUNTIL_ALPHA:
	 	if( !IsGraphics() ) ORwait = WAITUNTIL_DONT;
		return;
	 case WAITUNTIL_GRAPH:
	 	if( IsGraphics() ) ORwait = WAITUNTIL_DONT;
		return;
	 case WAITUNTIL_CRTUP:
		if( !(IO_CRTCTL & 6) ) ORwait = WAITUNTIL_DONT;
		return;
	 case WAITUNTIL_KEY:
		return;
	 case WAITUNTIL_CLICK:
		if( ORclick ) ORwait = WAITUNTIL_DONT;
		return;
	 case WAITUNTIL_DISK:
		if( IsDisk(WaitSC, WaitCA, WaitUNIT) ) ORwait = WAITUNTIL_DONT;
		return;
	 default:
	 	sprintf((char *)ScratBuf, "SCRIPT unknown WAIT state: %ld on line %d", ORwait, ORline);
	 	SystemMessage((char *)ScratBuf);
ORbail:
			sprintf((char *)ScratBuf, "SCRIPT ORbail on line %d", ORline);
			SystemMessage((char *)ScratBuf);

		MainRetVal = 1;	// indicate error
	 	goto ORfin;
	}

	e = (SCRIPTETC*)KWnds[scriptWnd].etc;
	ListBox_SetCurSel(e->hScriptList, ORline-1);

	/// Handle a BOOT if it's next
	if( !strnicmp((const char*)ORptr, "BOOT", 4) ) {	// if next thing is a BOOT block
		ORptr += 4;	// skip BOOT
		ORptr = POskipWhite(ORptr);
		if( *ORptr!='8' ) {
			sprintf((char *)ScratBuf, "SCRIPT invalid model (85A/85B/87) on line %d", ORline);
			SystemMessage((char *)ScratBuf);
			goto ORbail;
		}
		++ORptr;
		if( *ORptr=='7' ) {
			Override85.Model = 2;
		} else if( *ORptr=='5' ) {
			++ORptr;
			if( toupper(*ORptr)=='A' ) Override85.Model = 0;
			else if( toupper(*ORptr)=='B' ) Override85.Model = 1;
			else --ORptr;
		}
		++ORptr;	// skip last letter of model
		ORptr = POskipWhite(ORptr);
		if( *ORptr!='{' || Override85.Model==0xFFFFFFFF ) {
			sprintf((char *)ScratBuf, "SCRIPT expected '{' after BOOT model# on line %d", ORline);
			SystemMessage((char *)ScratBuf);
			goto ORbail;
		}
		++ORptr;	// skip {
		ORptr = POskipWhite(ORptr);
		while( *ORptr && *ORptr!='}' ) {
			if( !strnicmp((const char *)ORptr, "ROMS", 4) ) {
				ORptr += 4;
				ORptr = POskipWhite(ORptr);
				Override85.RomLoad = 0;
				while( *ORptr>='0' && *ORptr<='7' ) {	// while another ROM#
					c = 0;	// init this ROM#
					while( *ORptr=='0' ) ++ORptr;	// skip leading 0s
					while( *ORptr>='0' && *ORptr<='7' ) {	// chew up the rest of the ROM#
						c = (c<<3) | ((*ORptr++)-'0');
					}
					for(i=0; i<RomListCnt; i++) {
						if( RomList[i]==c ) {
							Override85.RomLoad |= (1<<i);
							break;
						}
					}
					if( i>=RomListCnt ) {	// unsupported ROM#
						sprintf((char *)ScratBuf, "OVERRIDE ROM %3.3o invalid on line %d", (unsigned int)c, ORline);
						SystemMessage((char *)ScratBuf);
						goto ORbail;
					}
					ORptr = POskipWhite(ORptr);
				}
			} else if( !strnicmp((const char *)ORptr, "RAM", 3) ) {
				ORptr += 3;
				ORptr = POskipWhite(ORptr);
				if( -1==(Override85.RamSize=POgetNum()) || -1==(Override85.RamSizeExt=POgetNum()) ) {
ramerr:
					sprintf((char *)ScratBuf, "SCRIPT RAM specification error on line %d", ORline);
					SystemMessage((char *)ScratBuf);
					goto ORbail;
				}
				if( Override85.RamSize!=16 && Override85.RamSize!=32 ) goto ramerr;
				if( Override85.RamSize==16 && Override85.Model ) Override85.RamSize = 32;
				if( Override85.RamSizeExt!=0 ) {
					if( Override85.RamSizeExt==32 ) Override85.RamSizeExt = 1;
					else if( Override85.RamSizeExt==64 ) Override85.RamSizeExt = 2;
					else if( Override85.RamSizeExt==128 ) Override85.RamSizeExt = 3;
					else if( Override85.RamSizeExt==256 ) Override85.RamSizeExt = 4;
					else if( Override85.RamSizeExt==512 ) Override85.RamSizeExt = 5;
					else if( Override85.RamSizeExt==1024 ) Override85.RamSizeExt = 6;
					else goto ramerr;
				}
			} else if( !strnicmp((const char *)ORptr, "DISPLAYx2", 9) ) {
				ORptr += 9;
				ORptr = POskipWhite(ORptr);
				if( Override85.CfgFlags==0xFFFFFFFF ) Override85.CfgFlags = 0;
				if( *ORptr=='0' ) ;
				else if( *ORptr=='1' ) Override85.CfgFlags |= CFG_BIGCRT;
				else {
					sprintf((char *)ScratBuf, "SCRIPT invalid DISPLAYx2 value (0/1) on line %d", ORline);
					SystemMessage((char *)ScratBuf);
					goto ORbail;
				}
				++ORptr;	// skip 0 or 1 char
			} else if( !strnicmp((const char *)ORptr, "SLOWSPEED", 9) ) {
				ORptr += 9;
				ORptr = POskipWhite(ORptr);
				if( Override85.CfgFlags==0xFFFFFFFF ) Override85.CfgFlags = 0;
				if( *ORptr=='0' ) ;
				else if( *ORptr=='1' ) Override85.CfgFlags |= CFG_NATURAL_SPEED;
				else {
					sprintf((char *)ScratBuf, "SCRIPT invalid SLOWSPEED value (0/1) on line %d", ORline);
					SystemMessage((char *)ScratBuf);
					goto ORbail;
				}
				++ORptr;	// skip 0 or 1 char
			}
			ORptr = POskipWhite(ORptr);
		}
		if( *ORptr ) ++ORptr;	// skip }

		// shutdown if was already running
		if( HP85UP ) {
			if( (IO_TAPSTS | IO_TAPCART) & 1 ) WriteTape();	// make sure any tape changes are saved
			for(c=0; c<16; c++) {	// flush out any changes to disks
				if( IOcards[c].initProc != NULL && IOcards[c].type==IOTYPE_HPIB ) IOcards[c].initProc(iR_FLUSH, Model, c, 0);
			}
			HP85PWO(FALSE);
			HP85OnShutdown();
		}
		HP85OnStartup();
		// automatically block until IDLE after BOOT
		ORwait = WAITUNTIL_IDLE;
		return;
	}

	/// If Emulator is just starting, don't do anything BUT the BOOT
	if( from==0 ) return;	// bail if they ONLY wanted a BOOT block (initial startup)

	/// Now process all other 'normal' COMMANDS

	///$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
	if( !strnicmp((const char*)ORptr, "WAITUNTIL", 9) ) {	// if next thing is a WAITUNTIL line
		if( pKeyMem!=NULL || pKeysNow!=NULL ) return;	// block on WAIT until all keys have been processed

		ORptr += 9;
		ORptr = POskipWhite(ORptr);
		if( !strnicmp((const char*)ORptr, "SECONDS", 7) ) {
			ORptr += 7;
			ORptr = POskipWhite(ORptr);
			if( (i = POgetNum()) >= 0 ) {
				ORwait = WAITUNTIL_SECONDS;
				ORseconds = i;
				ORlastTime = KGetTicks();
			}
		} else if( !strnicmp((const char*)ORptr, "IDLE", 4) ) {
			ORptr += 4;
			ORwait = WAITUNTIL_IDLE;
		} else if( !strnicmp((const char*)ORptr, "INPUT", 5) ) {
			ORptr += 5;
			ORwait = WAITUNTIL_INPUT;
		} else if( !strnicmp((const char*)ORptr, "ALPHA", 5) ) {
			ORptr += 5;
			ORwait = WAITUNTIL_ALPHA;
		} else if( !strnicmp((const char*)ORptr, "GRAPH", 5) ) {
			ORptr += 5;
			ORwait = WAITUNTIL_GRAPH;
		} else if( !strnicmp((const char*)ORptr, "CRTUP", 5) ) {
			ORptr += 5;
			ORwait = WAITUNTIL_CRTUP;
		} else if( !strnicmp((const char*)ORptr, "KEY", 3) ) {
			ORptr += 3;

			keycnt = ORparseKeys(ORptr, NULL);
			if( keycnt==0 ) {	// no key specified, take ANY key
				WaitKey = -1;	// accept ANY key
			} else {	// specific key required
				// no error (have keys), so allocate a buffer and parse it again into the buffer this time
				if( NULL != (pDN = (char *)malloc(keycnt+1)) ) {	// one extra for NUL
					keycnt = ORparseKeys(ORptr, (BYTE*)pDN);
					WaitKey = (int)(unsigned int)(BYTE)pDN[0];	// specific key to wait for
					free(pDN);
					while( *ORptr && *ORptr!='\r' && *ORptr!='\n' ) ++ORptr;	// skip over the line
				} else {
					SystemMessage("SCRIPT failed to allocate KEYSIDLE buffer");
					goto ORbail;
				}
			}
			ORwait = WAITUNTIL_KEY;
		} else if( !strnicmp((const char*)ORptr, "CLICK", 5) ) {
			ORptr += 5;
			ORclick = FALSE;
			ORwait = WAITUNTIL_CLICK;
		} else if( !strnicmp((const char*)ORptr, "DISK", 4) ) {
			ORptr += 4;
			ORptr = POskipWhite(ORptr);

			if( *ORptr=='1' && *(ORptr+1)=='0' ) {WaitSC = 7; ORptr += 2;}
			else if( *ORptr>='3' && *ORptr<='9' ) WaitSC = *ORptr++ - '3';
			else goto diskmsuserr;

			if( *ORptr>='0' && *ORptr<='7' ) WaitCA = *ORptr++ - '0';
			else goto diskmsuserr;

			if( *ORptr>='0' && *ORptr<='3' ) WaitUNIT = *ORptr++ - '0';
			else goto diskmsuserr;

			ORwait = WAITUNTIL_DISK;
		} else {
			for(i=0; i<15 && ORptr[i]; i++) poerrbuf[i] = ORptr[i];
			ORptr[i] = 0;
			sprintf((char *)ScratBuf, "SCRIPT unknown WAITUNTIL condition [%s] on line %d", poerrbuf, ORline);
			SystemMessage((char *)ScratBuf);
			goto ORbail;	// terminate script
		}
	///$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
	} else if( !strnicmp((const char*)ORptr, "DISK", 4) ) {	// if next thing is a DISK line
		if( from!=1 ) return;	// only swap disks from HP85OnIdle()

		ORptr += 4;
		ORptr = POskipWhite(ORptr);

		if( *ORptr=='1' && *(ORptr+1)=='0' ) {sc = 7; ORptr += 2;}
		else if( *ORptr>='3' && *ORptr<='9' ) {sc = *ORptr-'3'; ORptr++;}
		else {
diskmsuserr:
			sprintf((char *)ScratBuf, "SCRIPT invalid MSUS on line %d", ORline);
			SystemMessage((char *)ScratBuf);
			goto ORbail;
		}
		if( *ORptr>='0' && *ORptr<='7' ) ca = *ORptr++ - '0';
		else goto diskmsuserr;

		if( *ORptr>='0' && *ORptr<='3' ) unit = *ORptr++ - '0';
		else goto diskmsuserr;

		if( *ORptr!=' ' && *ORptr!='\t' ) {
diskerr:
			sprintf((char *)ScratBuf, "SCRIPT invalid DISK syntax on line %d", ORline);
			SystemMessage((char *)ScratBuf);
			goto ORbail;
		}
		ORptr = POskipWhite(ORptr);
		if( !strnicmp((const char*)ORptr, "NONE", 4) ) {	// EJECT disk
			ORptr += 4;
			EjectDisk(sc, ca, unit);
			DrawDisks(sc, ca);
		} else if( !strnicmp((const char*)ORptr, "NEW", 3) ) {	// NEW disk
			ORptr += 3;
			ORptr = POskipWhite(ORptr);

			if( *ORptr<'0' || *ORptr>'9' ) goto diskerr;
			i = *ORptr++ - '0';	// i = FOLDER#

			if( !POgetName(Sname1) ) goto diskerr;

			EjectDisk(sc, ca, unit);	// eject any disk in drive

			SetDiskFolder(sc, ca, unit, i);
			NewDisk(sc, ca, unit, Sname1);	// create a disk in memory
			EjectDisk(sc, ca, unit);			// write it to hardware disk file
			LoadDisk(sc, ca, unit, Sname1);	// load it back as real disk file (uninitialized)

			DrawDisks(sc, ca);
		} else if( !strnicmp((const char*)ORptr, "DELETE", 6) ) {	// DELETE disk
			ORptr += 6;
			ORptr = POskipWhite(ORptr);

			if( *ORptr<'0' || *ORptr>'9' ) goto diskerr;
			df = *ORptr++ - '0';	// FOLDER#

			if( !POgetName(Sname1) ) goto diskerr;
			if( POinDiffDrive(Sname1, sc, ca, unit) ) goto ORbail;

			pDN = GetDiskName(sc, ca, unit);
			if( !strcmp(pDN, (char*)Sname1) ) EjectDisk(sc, ca, unit);
			sprintf((char*)ScratBuf, "DISKS%d\\%s", df, Sname1);
			Kdelete((char*)ScratBuf);
			return;
		} else if( !strnicmp((const char*)ORptr, "RENAME", 6) ) {	// RENAME disk
			ORptr += 6;
			ORptr = POskipWhite(ORptr);

			if( *ORptr<'0' || *ORptr>'9' ) goto diskerr;
			df = *ORptr++ - '0';	// FOLDER#

			if( !POgetName(Sname1) ) goto diskerr;
			if( !POgetName(Sname2) ) goto diskerr;
			if( POinDiffDrive(Sname1, sc, ca, unit) ) goto ORbail;

			pDN = GetDiskName(sc, ca, unit);
			if( !strcmp(pDN, (char*)Sname1) ) {
				EjectDisk(sc, ca, unit);
				sprintf((char*)ScratBuf, "DISKS%d\\%s", df, Sname1);
				sprintf((char*)ScratBuf+256, "DISKS%d\\%s", df, Sname2);
				Krename((char*)ScratBuf, (char*)ScratBuf+256);
				LoadDisk(sc, ca, unit, Sname2);
			} else {
				sprintf((char*)ScratBuf, "DISKS%d\\%s", df, Sname1);
				sprintf((char*)ScratBuf+256, "DISKS%d\\%s", df, Sname2);
				Krename((char*)ScratBuf, (char*)ScratBuf+256);
			}
			DrawDisks(sc, ca);
			return;
		} else if( !strnicmp((const char*)ORptr, "LOAD", 4) ) {
			ORptr += 4;
			ORptr = POskipWhite(ORptr);

			if( *ORptr<'0' || *ORptr>'9' ) goto diskerr;
			df = *ORptr++ - '0';

			if( !POgetName(Sname1) ) goto diskerr;

			if( POinDiffDrive(Sname1, sc, ca, unit) ) goto ORbail;

			SetDiskFolder(sc, ca, unit, df);
			EjectDisk(sc, ca, unit);
			LoadDisk(sc, ca, unit, Sname1);
			DrawDisks(sc, ca);
		} else goto diskerr;
	///$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
	} else if( !strnicmp((const char*)ORptr, "SCREENCAPTURE", 13) ) {	// if next thing is a SCREENCAPTURE line
		PBMB	hBM;

		ORptr += 13;
		ORptr = POskipWhite(ORptr);

		if( !strnicmp((const char*)ORptr, "ALPHA", 5) ) {i = 0; ORptr += 5;}
		else if( !strnicmp((const char*)ORptr, "GRAPH", 5) ) {i = 1; ORptr += 5;}
		else if( !strnicmp((const char*)ORptr, "ALL", 3) ) {i = 2; ORptr += 3;}
		else {
			sprintf((char *)ScratBuf, "SCRIPT expected SCREENCAPTURE ALPHA, GRAPH, or ALL on line %d", ORline);
			SystemMessage((char *)ScratBuf);
			goto ORbail;	// terminate script
		}

		if( !POgetName(Sname1) ) {
			sprintf((char *)ScratBuf, "SCRIPT expected SCREENCAPTURE filename on line %d", ORline);
			SystemMessage((char *)ScratBuf);
			goto ORbail;	// terminate script
		}
		// if not .BMP on end of filename, add it
		if( stricmp((char *)Sname1+strlen((char *)Sname1)-4, ".bmp") ) strcat((char *)Sname1, ".bmp");

		if( i==2 ) i = SaveBMB(hp85BM, (char *)Sname1);
		else {
			if( Model<2 ) {
				hBM = CreateBMB(256, 192, NULL);
				BltBMB(hBM, 0, 0, hp85BM, 0, i ? 192 : 0, 256, 192, 0);
			} else {
				hBM = CreateBMB(640, 240, NULL);
				BltBMB(hBM, 0, 0, hp85BM, 0, i ? 240 : 0, 640, 240, 0);
			}
			i = SaveBMB(hBM, (char *)Sname1);
			DeleteBMB(hBM);
		}
		if( !i ) {	// error during SaveBMB()
			sprintf((char *)ScratBuf, "SCRIPT failed to save SCREENCAPTURE on line %d", ORline);
			SystemMessage((char *)ScratBuf);
			goto ORbail;	// terminate script
		}
	///$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
	} else if( !strnicmp((const char*)ORptr, "KEYSIDLE", 8) ) {	// if next thing is a KEYS line
		if( pKeyMem!=NULL || pKeysNow!=NULL || !Is85Idle(0) || from!=1 ) return;	// block until called from HP85Idle(), while machine is IDLE, with no TYPEFILE or prior KEYSIDLE/KEYSNOW running

		ORptr += 8;
		keycnt = ORparseKeys(ORptr, NULL);
		if( keycnt==0 ) goto ORbail;	// error occurred, terminate SCRIPT
		// no error (have keys), so allocate a buffer and parse it again into the buffer this time
		if( NULL != (pKeysNow = (BYTE *)malloc(keycnt+1)) ) {	// one extra for NUL on last line
			keycnt = ORparseKeys(ORptr, pKeysNow);
			pKeysNow[keycnt] = 0;
			pNowPtr = pKeysNow;
			inKEYSNOW = FALSE;
			while( *ORptr && *ORptr!='\r' && *ORptr!='\n' ) ++ORptr;	// skip over the line
			ORwait = WAITUNTIL_IDLE;
		} else {
			SystemMessage("SCRIPT failed to allocate KEYSIDLE buffer");
			goto ORbail;
		}
	///$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
	} else if( !strnicmp((const char*)ORptr, "KEYSNOW", 7) ) {	// if next thing is a KEYS line
		if( pKeyMem!=NULL || pKeysNow!=NULL ) return;	// block until called with no TYPEFILE or prior KEYSIDLE/KEYSNOW running

		keycnt = ORparseKeys(ORptr+7, NULL);
		if( keycnt==0 ) goto ORbail;	// error occurred, terminate SCRIPT
		// no error (have keys), so allocate a buffer and parse it again into the buffer this time
		if( NULL != (pKeysNow = (BYTE *)malloc(keycnt+1)) ) {	// one extra for NUL on last line
			keycnt = ORparseKeys(ORptr+7, pKeysNow);
			pKeysNow[keycnt] = 0;
			pNowPtr = pKeysNow;
			inKEYSNOW = TRUE;
			while( *ORptr && *ORptr!='\r' && *ORptr!='\n' ) ++ORptr;	// skip over the line
		} else {
			SystemMessage("SCRIPT failed to allocate KEYSNOW buffer");
			goto ORbail;
		}
	///$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
	} else if( !strnicmp((const char *)ORptr, "SLOWSPEED", 9) ) {
		if( from!=1 ) return;	// block until coming from the IDLE loop call, so the SLOWSPEED change takes effect IMMEDIATELY in the script
		ORptr += 9;
		ORptr = POskipWhite(ORptr);
		if( *ORptr=='0' ) CfgFlags &= ~CFG_NATURAL_SPEED;
		else if( *ORptr=='1' ) CfgFlags |= CFG_NATURAL_SPEED;
		else {
			sprintf((char *)ScratBuf, "SCRIPT invalid SLOWSPEED value (0/1) on line %d", ORline);
			SystemMessage((char *)ScratBuf);
			goto ORbail;
		}
		++ORptr;	// skip 0 or 1 char
	///$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
	} else if( !strnicmp((const char *)ORptr, "DISPLAYx2", 9) ) {
		DWORD			oldflags;

		if( from!=1 ) return;	// block until coming from the IDLE loop call

	 	oldflags = CfgFlags;
		ORptr += 9;
		ORptr = POskipWhite(ORptr);
		if( *ORptr=='0' ) CfgFlags &= ~CFG_BIGCRT;
		else if( *ORptr=='1' ) CfgFlags |= CFG_BIGCRT;
		else {
			sprintf((char *)ScratBuf, "SCRIPT invalid DISPLAYx2 value (0/1) on line %d", ORline);
			SystemMessage((char *)ScratBuf);
			goto ORbail;
		}
		++ORptr;	// skip 0 or 1 char
		if( !SetupDispVars() ) KillProgram();
		else DrawAll();
		if( (oldflags ^ CfgFlags) & CFG_BIGCRT ) {
			for(j=0; j<8; j++) {
				if( IOcards[j].initProc != NULL && IOcards[j].type==IOTYPE_HPIB )  {	// if card is present and is HPIB card
					for(i=0; i<31; i++) {
						AdjustDisks(j, i);
						DrawDisks(j, i);
					}
				}
			}
		}
	///$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
	} else if( !strnicmp((const char*)ORptr, "TYPEFILE", 8) ) {	// if next thing is a TYPEFILE line
		long	f;

		if( pKeyMem!=NULL || pKeysNow!=NULL || !Is85Idle(0) || from!=1 ) return;	// block until called from HP85OnIdle and no pending keys

		ORptr += 8;

		if( !POgetName(Sname1) ) {
			sprintf((char *)ScratBuf, "SCRIPT expected TYPEFILE filename on line %d", ORline);
			SystemMessage((char *)ScratBuf);
			goto ORbail;	// terminate script
		}
		if( -1!=(f=KopenAbs((char*)Sname1, O_RDBIN)) ) {
			KeyMemLen = Kseek(f, 0, SEEK_END);
			if( NULL != (pKeyMem = (BYTE *)malloc(KeyMemLen+1)) ) {	// one extra for NUL on last line in case no CR/LF on last line
				Kseek(f, 0, SEEK_SET);
				Kread(f, pKeyMem, KeyMemLen);
				pKeyMem[KeyMemLen] = 0;
				pKeyPtr = pKeyMem;
				InTypeFile = TRUE;
			}
			Kclose(f);
		} else {
			sprintf((char *)ScratBuf, "SCRIPT TYPEFILE filename error on line %d", ORline);
			SystemMessage((char *)ScratBuf);
			goto ORbail;	// terminate script
		}
	///$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
	} else if( !strnicmp((const char*)ORptr, "IMPORTFILE", 10) ) {	// if next thing is a IMPORTFILE line
		if( pKeyMem!=NULL || pKeysNow!=NULL || !Is85Idle(0) || from!=1 ) return;

		ORptr += 10;

		if( !POgetName(Sname1) ) {
imperr:
			sprintf((char *)ScratBuf, "SCRIPT expected IMPORTFILE <OSname> <LIFname> on line %d", ORline);
			SystemMessage((char *)ScratBuf);
			goto ORbail;	// terminate script
		}
		if( !POgetName(Sname2) ) goto imperr;
		if( !Sname1[0] || !Sname2[0] ) goto imperr;		// don't have a name

		GetMSUS(Sname2, &sc, &ca, &unit);

		if( IsDisk(sc, ca, unit) ) {
			if( IsDiskInitialized(sc, ca, unit) ) {
				if( !IsDiskProtected(sc, ca, unit) ) {
					ImportFile(sc, ca, unit, (char*)Sname1, (char*)Sname2);
				} else {
					sprintf((char *)ScratBuf, "SCRIPT Disk is write protected on line %d",ORline);
					SystemMessage((char *)ScratBuf);
					goto ORbail;
				}
			} else {
				sprintf((char *)ScratBuf, "SCRIPT Disk is not initialized on line %d",ORline);
				SystemMessage((char *)ScratBuf);
				goto ORbail;
			}
		} else {
			sprintf((char *)ScratBuf, "SCRIPT Disk is not loaded on line %d",ORline);
			SystemMessage((char *)ScratBuf);
			goto ORbail;
		}
	///$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
	} else if( !strnicmp((const char*)ORptr, "EXPORTFILE", 10) ) {	// if next thing is a EXPORTFILE line
		if( pKeyMem!=NULL || pKeysNow!=NULL || !Is85Idle(0) || from!=1 ) return;

		ORptr += 10;

		if( !POgetName(Sname1) ) {
experr:
			sprintf((char *)ScratBuf, "SCRIPT expected EXPORTFILE <LIFname> <OSname> on line %d", ORline);
			SystemMessage((char *)ScratBuf);
			goto ORbail;	// terminate script
		}
		if( !POgetName(Sname2) ) goto experr;
		if( !Sname1[0] || !Sname2[0] ) goto imperr;		// don't have a name

		GetMSUS(Sname1, &sc, &ca, &unit);

		for(i=0; i<10 && Sname1[i]; i++) ;
		for(; i<10; i++) Sname1[i] = ' ';
		Sname1[i] = 0;

		ExportLIF(sc, ca, unit, (char*)Sname1, (char*)Sname2);
	///$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
	} else if( !strnicmp((const char*)ORptr, "MAKEROM", 7) ) {	// if next thing is a MAKEROM line
		unsigned int	s, t;
		long	hFile;

		if( pKeyMem!=NULL || pKeysNow!=NULL || !Is85Idle(0) || from!=1 ) return;

		ORptr += 7;

		if( !POgetName(Sname1) ) {
mkromerr:
			sprintf((char *)ScratBuf, "SCRIPT expected MAKEROM <ROM DATA file name[:DMSUS]> <OS ROM path/name> on line %d", ORline);
			SystemMessage((char *)ScratBuf);
			goto ORbail;	// terminate script
		}

		if( !POgetName(Sname2) ) goto mkromerr;
		if( !Sname1[0] || !Sname2[0] ) goto mkromerr;		// don't have a name

		GetMSUS(Sname1, &sc, &ca, &unit);


		for(i=0; i<10 && Sname1[i]; i++) ;
		for(; i<10; i++) Sname1[i] = ' ';
		Sname1[i] = 0;

		if( !ReadROMfile(sc, ca, unit, (char*)Sname1) ) {
			sprintf((char *)ScratBuf, "SCRIPT error reading MAKEROM LIF file on line %d", ORline);
			SystemMessage((char *)ScratBuf);
			goto ORbail;	// terminate script
		}
		// ROM bytes are now in ScratBuf, 0-filled to end
		// now generate secondary checksum in [8188] and [8189]
		{
			unsigned int	r26, r27, r45, w;

			for(i=t=0; i<8188; i+=2) {
				w = (unsigned int)ScratBuf[i] + (((unsigned int)ScratBuf[i+1])<<8);
				for(j=0; j<16; j++) {
					r26 = t & 0xff;
					r27 = t>>8;
					r45 = r27;
					r27 = (r27<<4) & 0xff;
					r26 = ((r26<<1) & 0xff) | (w & 1);
					w >>= 1;
					r45 ^= r26;
					r45 ^= r27;
					if( r45 & 1 ) r45 = (r45 + 0x80) & 0xff;
					t = ((t<<1) & 0xffff) | (r45 >>7);
				}
			}
			ScratBuf[8188] = t & 0xFF;
			ScratBuf[8189] = t >> 8;
		}

		// now generate primary checksum in [8190] and [8191]
		for(i=t=0; i<8190; i+=2) t += (unsigned int)ScratBuf[i] + (((unsigned int)ScratBuf[i+1])<<8);
		s = ((t>>16) + (t & 0xffff)) & 0xffff;
		ScratBuf[8190] = (char)(255-(s & 0xFF));
		ScratBuf[8191] = (char)(255-(s>>8));

		if( -1 == (hFile=KopenAbs((char*)Sname2, O_WRBINNEW)) ) {
			sprintf((char *)ScratBuf, "SCRIPT error writing MAKEROM OS file on line %d", ORline);
			SystemMessage((char *)ScratBuf);
			goto ORbail;	// terminate script
		}
		Kwrite(hFile, ScratBuf, 8192);
		Kclose(hFile);
	///$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
	} else if( !strnicmp((const char*)ORptr, "END", 3) ) {	// if next thing is an END line
		if( from!=1 ) return;
		ORptr += 3;
		ORptr = POskipWhite(ORptr);
		if( !strnicmp((const char*)ORptr, "EXIT", 4) ) {
			ORptr += 4;
			ORptr = POskipWhite(ORptr);
			if( !strnicmp((const char*)ORptr, "KEEPINI", 7) ) {
				ORptr += 7;
				KeepIni = TRUE;
			}
			DoExit = TRUE;
		}
		*ORptr = 0;	// stop processing script
	///$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
	} else if( !strnicmp((const char*)ORptr, "CATALL", 6) ) {	// if next thing is a CATALL line
		if( from!=1 ) return;
		ORptr += 6;

		if( !POgetName(Sname1) ) {
			sprintf((char *)ScratBuf, "SCRIPT expected CATALL filename on line %d", ORline);
			SystemMessage((char *)ScratBuf);
			goto ORbail;	// terminate script
		}
		CatAll((char*)Sname1);
	///$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
	} else {
		ORptr = POskipWhite(ORptr);
		if( *ORptr ) {
			for(i=0; i<15 && ORptr[i]; i++) poerrbuf[i] = ORptr[i];
			ORptr[i] = 0;
			sprintf((char *)ScratBuf, "SCRIPT unexpected command [%s] on line %d", poerrbuf, ORline);
			SystemMessage((char *)ScratBuf);
		}
		goto ORbail;	// terminate script
	}
}

